@@ -15,7 +15,7 @@ import OptParser
 
 import Cryptol.REPL.Command (loadCmd,loadPrelude,CommandExitCode(..))
 import Cryptol.REPL.Monad (REPL,updateREPLTitle,setUpdateREPLTitle,
-                   io,prependSearchPath,setSearchPath)
+                   io,prependSearchPath,setSearchPath,parseSearchPath)
 import qualified Cryptol.REPL.Monad as REPL
 import Cryptol.ModuleSystem.Env(ModulePath(..))
 
@@ -23,19 +23,16 @@ import REPL.Haskeline
 import REPL.Logo
 
 import Cryptol.Utils.PP
-import Cryptol.Version (commitHash, commitBranch, commitDirty)
-import Paths_cryptol (version)
+import Cryptol.Version (displayVersion)
 
 import Control.Monad (when)
-import Data.Maybe (isJust)
-import Data.Version (showVersion)
 import GHC.IO.Encoding (setLocaleEncoding, utf8)
 import System.Console.GetOpt
     (OptDescr(..),ArgOrder(..),ArgDescr(..),getOpt,usageInfo)
 import System.Directory (getTemporaryDirectory, removeFile)
 import System.Environment (getArgs, getProgName, lookupEnv)
 import System.Exit (exitFailure,exitSuccess)
-import System.FilePath (searchPathSeparator, splitSearchPath, takeDirectory)
+import System.FilePath (searchPathSeparator, takeDirectory)
 import System.IO (hClose, hPutStr, openTempFile)
 
 
@@ -49,12 +46,14 @@ data Options = Options
   { optLoad            :: [FilePath]
   , optVersion         :: Bool
   , optHelp            :: Bool
-  , optBatch           :: Maybe FilePath
+  , optBatch           :: ReplMode
+  , optCallStacks      :: Bool
   , optCommands        :: [String]
   , optColorMode       :: ColorMode
   , optCryptolrc       :: Cryptolrc
   , optCryptolPathOnly :: Bool
   , optStopOnError     :: Bool
+  , optNoUnicodeLogo   :: Bool
   } deriving (Show)
 
 defaultOptions :: Options
@@ -62,12 +61,14 @@ defaultOptions  = Options
   { optLoad            = []
   , optVersion         = False
   , optHelp            = False
-  , optBatch           = Nothing
+  , optBatch           = InteractiveRepl
+  , optCallStacks      = True
   , optCommands        = []
   , optColorMode       = AutoColor
   , optCryptolrc       = CryrcDefault
   , optCryptolPathOnly = False
   , optStopOnError     = False
+  , optNoUnicodeLogo   = False
   }
 
 options :: [OptDescr (OptParser Options)]
@@ -75,6 +76,9 @@ options  =
   [ Option "b" ["batch"] (ReqArg setBatchScript "FILE")
     "run the script provided and exit"
 
+  , Option "" ["interactive-batch"] (ReqArg setInteractiveBatchScript "FILE")
+    "run the script provided and exit, but behave as if running an interactive session"
+
   , Option "e" ["stop-on-error"] (NoArg setStopOnError)
     "stop script execution as soon as an error occurs."
 
@@ -95,6 +99,12 @@ options  =
   , Option "h" ["help"] (NoArg setHelp)
     "display this message"
 
+  , Option "" ["no-call-stacks"] (NoArg setNoCallStacks)
+    "Disable tracking of call stack information, which reduces interpreter overhead"
+
+  , Option "" ["no-unicode-logo"] (NoArg setNoUnicodeLogo)
+    "Don't use unicode characters in the REPL logo"
+
   , Option ""  ["ignore-cryptolrc"] (NoArg setCryrcDisabled)
     "disable reading of .cryptolrc files"
 
@@ -121,7 +131,11 @@ setStopOnError = modify $ \opts -> opts { optStopOnError = True }
 
 -- | Set a batch script to be run.
 setBatchScript :: String -> OptParser Options
-setBatchScript path = modify $ \ opts -> opts { optBatch = Just path }
+setBatchScript path = modify $ \ opts -> opts { optBatch = Batch path }
+
+-- | Set an interactive batch script
+setInteractiveBatchScript :: String -> OptParser Options
+setInteractiveBatchScript path = modify $ \ opts -> opts { optBatch = InteractiveBatch path }
 
 -- | Set the color mode of the terminal output.
 setColorMode :: String -> OptParser Options
@@ -130,6 +144,10 @@ setColorMode "none"   = modify $ \ opts -> opts { optColorMode = NoColor }
 setColorMode "always" = modify $ \ opts -> opts { optColorMode = AlwaysColor }
 setColorMode x        = OptFailure ["invalid color mode: " ++ x ++ "\n"]
 
+-- | Disable unicde characters in the REPL logo
+setNoUnicodeLogo :: OptParser Options
+setNoUnicodeLogo = modify $ \opts -> opts { optNoUnicodeLogo = True }
+
 -- | Signal that version should be displayed.
 setVersion :: OptParser Options
 setVersion  = modify $ \ opts -> opts { optVersion = True }
@@ -142,6 +160,10 @@ setHelp  = modify $ \ opts -> opts { optHelp = True }
 setCryrcDisabled :: OptParser Options
 setCryrcDisabled  = modify $ \ opts -> opts { optCryptolrc = CryrcDisabled }
 
+-- | Disable call stack tracking
+setNoCallStacks :: OptParser Options
+setNoCallStacks = modify $ \opts -> opts { optCallStacks = False }
+
 -- | Add another file to read as a @.cryptolrc@ file, unless @.cryptolrc@
 -- files have been disabled
 addCryrc :: String -> OptParser Options
@@ -160,16 +182,6 @@ parseArgs args = case getOpt (ReturnInOrder addFile) options args of
   (ps,[],[]) -> runOptParser defaultOptions (mconcat ps)
   (_,_,errs) -> Left errs
 
-displayVersion :: IO ()
-displayVersion = do
-    let ver = showVersion version
-    putStrLn ("Cryptol " ++ ver)
-    putStrLn ("Git commit " ++ commitHash)
-    putStrLn ("    branch " ++ commitBranch ++ dirtyLab)
-      where
-      dirtyLab | commitDirty = " (non-committed files present during build)"
-               | otherwise   = ""
-
 displayHelp :: [String] -> IO ()
 displayHelp errs = do
   prog <- getProgName
@@ -183,9 +195,15 @@ displayHelp errs = do
             , "addition to the default locations"
             ]
           )
-        , ( "SBV_{ABC,BOOLECTOR,CVC4,MATHSAT,YICES,Z3}_OPTIONS"
+        , ( "EDITOR"
+          , [ "Sets the editor executable to use when opening an editor"
+            , "via the `:edit` command"
+            ]
+          )
+        , ( "SBV_{ABC,BOOLECTOR,CVC4,CVC5,MATHSAT,YICES,Z3}_OPTIONS"
           , [ "A string of command-line arguments to be passed to the"
             , "corresponding solver invoked for `:sat` and `:prove`"
+            , "when using a prover via SBV"
             ]
           )
         ]
@@ -204,11 +222,12 @@ main  = do
 
     Right opts
       | optHelp opts    -> displayHelp []
-      | optVersion opts -> displayVersion
+      | optVersion opts -> displayVersion putStrLn
       | otherwise       -> do
           (opts', mCleanup) <- setupCmdScript opts
           status <- repl (optCryptolrc opts')
                          (optBatch opts')
+                         (optCallStacks opts')
                          (optStopOnError opts')
                          (setupREPL opts')
           case mCleanup of
@@ -228,23 +247,17 @@ setupCmdScript opts =
       (path, h) <- openTempFile tmpdir "cmds.icry"
       hPutStr h (unlines cmds)
       hClose h
-      when (isJust (optBatch opts)) $
+      when (optBatch opts /= InteractiveRepl) $
         putStrLn "[warning] --command argument specified; ignoring batch file"
-      return (opts { optBatch = Just path }, Just path)
+      return (opts { optBatch = InteractiveBatch path }, Just path)
 
 setupREPL :: Options -> REPL ()
 setupREPL opts = do
   mCryptolPath <- io $ lookupEnv "CRYPTOLPATH"
   case mCryptolPath of
     Nothing -> return ()
-    Just path | optCryptolPathOnly opts -> setSearchPath path'
-              | otherwise               -> prependSearchPath path'
-#if defined(mingw32_HOST_OS) || defined(__MINGW32__)
-      -- Windows paths search from end to beginning
-      where path' = reverse (splitSearchPath path)
-#else
-      where path' = splitSearchPath path
-#endif
+    Just path | optCryptolPathOnly opts -> setSearchPath (parseSearchPath path)
+              | otherwise               -> prependSearchPath (parseSearchPath path)
   smoke <- REPL.smokeTest
   case smoke of
     [] -> return ()
@@ -257,14 +270,17 @@ setupREPL opts = do
     AlwaysColor -> return True
     NoColor     -> return False
     AutoColor   -> canDisplayColor
-  displayLogo color
+
+  let useUnicode = not (optNoUnicodeLogo opts)
+  displayLogo color useUnicode
 
   setUpdateREPLTitle (shouldSetREPLTitle >>= \b -> when b setREPLTitle)
   updateREPLTitle
   case optBatch opts of
-    Nothing -> return ()
     -- add the directory containing the batch file to the module search path
-    Just file -> prependSearchPath [ takeDirectory file ]
+    Batch file -> prependSearchPath [ takeDirectory file ]
+    _ -> return ()
+
   case optLoad opts of
     []  -> loadPrelude `REPL.catch` \x -> io $ print $ pp x
     [l] -> loadCmd l `REPL.catch` \x -> do
