@@ -6,6 +6,7 @@
 -- Stability   :  provisional
 -- Portability :  portable
 
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE MultiWayIf #-}
 {-# LANGUAGE PatternGuards #-}
 {-# OPTIONS_GHC -fno-warn-orphans #-}
@@ -15,80 +16,96 @@ module REPL.Haskeline where
 import           Cryptol.REPL.Command
 import           Cryptol.REPL.Monad
 import           Cryptol.REPL.Trie
-import           Cryptol.Utils.PP
+import           Cryptol.Utils.PP hiding ((</>))
 import           Cryptol.Utils.Logger(stdoutLogger)
 import           Cryptol.Utils.Ident(modNameToText, interactiveName)
 
 import qualified Control.Exception as X
 import           Control.Monad (guard, join)
 import qualified Control.Monad.Trans.Class as MTL
+#if !MIN_VERSION_haskeline(0,8,0)
 import           Control.Monad.Trans.Control
+#endif
 import           Data.Char (isAlphaNum, isSpace)
-import           Data.Maybe(isJust)
 import           Data.Function (on)
 import           Data.List (isPrefixOf,nub,sortBy,sort)
 import qualified Data.Set as Set
 import qualified Data.Text as T (unpack)
-import           System.IO (stdout)
 import           System.Console.ANSI (setTitle, hSupportsANSI)
 import           System.Console.Haskeline
 import           System.Directory ( doesFileExist
                                   , getHomeDirectory
                                   , getCurrentDirectory)
 import           System.FilePath ((</>))
+import           System.IO (stdout)
 
 import           Prelude ()
 import           Prelude.Compat
 
 
+data ReplMode
+  = InteractiveRepl -- ^ Interactive terminal session
+  | Batch FilePath  -- ^ Execute from a batch file
+  | InteractiveBatch FilePath
+     -- ^ Execute from a batch file, but behave as though
+     --   lines are entered in an interactive session.
+ deriving (Show, Eq)
+
 -- | One REPL invocation, either from a file or from the terminal.
-crySession :: Maybe FilePath -> Bool -> REPL CommandExitCode
-crySession mbBatch stopOnError =
+crySession :: ReplMode -> Bool -> REPL CommandExitCode
+crySession replMode stopOnError =
   do settings <- io (setHistoryFile (replSettings isBatch))
-     let act = runInputTBehavior behavior settings (withInterrupt loop)
+     let act = runInputTBehavior behavior settings (withInterrupt (loop 1))
      if isBatch then asBatch act else act
   where
-  (isBatch,behavior) = case mbBatch of
-    Nothing   -> (False,defaultBehavior)
-    Just path -> (True,useFile path)
+  (isBatch,behavior) = case replMode of
+    InteractiveRepl       -> (False, defaultBehavior)
+    Batch path            -> (True,  useFile path)
+    InteractiveBatch path -> (False, useFile path)
 
-  loop :: InputT REPL CommandExitCode
-  loop =
+  loop :: Int -> InputT REPL CommandExitCode
+  loop lineNum =
     do ln <- getInputLines =<< MTL.lift getPrompt
        case ln of
          NoMoreLines -> return CommandOk
          Interrupted
            | isBatch && stopOnError -> return CommandError
-           | otherwise -> loop
-         NextLine line
-           | all isSpace line -> loop
-           | otherwise        -> doCommand line
+           | otherwise -> loop lineNum
+         NextLine ls
+           | all (all isSpace) ls -> loop (lineNum + length ls)
+           | otherwise            -> doCommand lineNum ls
 
-  doCommand txt =
-    case parseCommand findCommandExact txt of
-      Nothing | isBatch && stopOnError  -> return CommandError
-              | otherwise -> loop -- say somtething?
+  run lineNum cmd =
+    case replMode of
+      InteractiveRepl    -> runCommand lineNum Nothing cmd
+      InteractiveBatch _ -> runCommand lineNum Nothing cmd
+      Batch path         -> runCommand lineNum (Just path) cmd
+
+  doCommand lineNum txt =
+    case parseCommand findCommandExact (unlines txt) of
+      Nothing | isBatch && stopOnError -> return CommandError
+              | otherwise -> loop (lineNum + length txt)  -- say somtething?
       Just cmd -> join $ MTL.lift $
-        do status <- handleInterrupt (handleCtrlC CommandError) (runCommand cmd)
+        do status <- handleInterrupt (handleCtrlC CommandError) (run lineNum cmd)
            case status of
              CommandError | isBatch && stopOnError -> return (return status)
              _ -> do goOn <- shouldContinue
-                     return (if goOn then loop else return status)
+                     return (if goOn then loop (lineNum + length txt) else return status)
 
 
-data NextLine = NextLine String | NoMoreLines | Interrupted
+data NextLine = NextLine [String] | NoMoreLines | Interrupted
 
 getInputLines :: String -> InputT REPL NextLine
 getInputLines = handleInterrupt (MTL.lift (handleCtrlC Interrupted)) . loop []
   where
   loop ls prompt =
-    do mb <- getInputLine prompt
+    do mb <- fmap (filter (/= '\r')) <$> getInputLine prompt
        let newPropmpt = map (\_ -> ' ') prompt
        case mb of
          Nothing -> return NoMoreLines
          Just l
            | not (null l) && last l == '\\' -> loop (init l : ls) newPropmpt
-           | otherwise -> return $ NextLine $ unlines $ reverse $ l : ls
+           | otherwise -> return $ NextLine $ reverse $ l : ls
 
 loadCryRC :: Cryptolrc -> REPL CommandExitCode
 loadCryRC cryrc =
@@ -103,32 +120,39 @@ loadCryRC cryrc =
        let file = dir </> ".cryptolrc"
        present <- io (doesFileExist file)
        if present
-         then crySession (Just file) True
+         then crySession (Batch file) True
          else check others
 
   loadMany []       = return CommandOk
-  loadMany (f : fs) = do status <- crySession (Just f) True
+  loadMany (f : fs) = do status <- crySession (Batch f) True
                          case status of
                            CommandOk -> loadMany fs
                            _         -> return status
 
 -- | Haskeline-specific repl implementation.
-repl :: Cryptolrc -> Maybe FilePath -> Bool -> REPL () -> IO CommandExitCode
-repl cryrc mbBatch stopOnError begin =
-  runREPL (isJust mbBatch) stdoutLogger $
-  do status <- loadCryRC cryrc
-     case status of
-       CommandOk -> begin >> crySession mbBatch stopOnError
-       _         -> return status
+repl :: Cryptolrc -> ReplMode -> Bool -> Bool -> REPL () -> IO CommandExitCode
+repl cryrc replMode callStacks stopOnError begin =
+  runREPL isBatch callStacks stdoutLogger replAction
 
+ where
+  -- this flag is used to suppress the logo and prompts
+  isBatch = case replMode of
+              InteractiveRepl -> False
+              Batch _ -> True
+              InteractiveBatch _ -> True
 
+  replAction =
+    do status <- loadCryRC cryrc
+       case status of
+         CommandOk -> begin >> crySession replMode stopOnError
+         _         -> return status
 
 -- | Try to set the history file.
 setHistoryFile :: Settings REPL -> IO (Settings REPL)
 setHistoryFile ss =
   do dir <- getHomeDirectory
      return ss { historyFile = Just (dir </> ".cryptol_history") }
-   `X.catch` \(SomeException {}) -> return ss
+   `X.catch` \(X.SomeException {}) -> return ss
 
 -- | Haskeline settings for the REPL.
 replSettings :: Bool -> Settings REPL
@@ -159,9 +183,11 @@ data Cryptolrc =
 
 -- Utilities -------------------------------------------------------------------
 
+#if !MIN_VERSION_haskeline(0,8,0)
 instance MonadException REPL where
   controlIO f = join $ liftBaseWith $ \f' ->
     f $ RunIO $ \m -> restoreM <$> (f' m)
+#endif
 
 -- Titles ----------------------------------------------------------------------
 
@@ -196,7 +222,7 @@ canDisplayColor = io (hSupportsANSI stdout)
 cryptolCommand :: CompletionFunc REPL
 cryptolCommand cursor@(l,r)
   | ":" `isPrefixOf` l'
-  , Just (cmd,rest) <- splitCommand l' = case nub (findCommand cmd) of
+  , Just (_,cmd,rest) <- splitCommand l' = case nub (findCommand cmd) of
 
       [c] | null rest && not (any isSpace l') -> do
             return (l, cmdComp cmd c)
