@@ -11,32 +11,39 @@
 {-# LANGUAGE DeriveGeneric #-}
 module Cryptol.Eval.Type where
 
-import Cryptol.Eval.Monad
+import Cryptol.Backend.Monad (evalPanic)
 import Cryptol.TypeCheck.AST
 import Cryptol.TypeCheck.PP(pp)
 import Cryptol.TypeCheck.Solver.InfNat
 import Cryptol.Utils.Panic (panic)
 import Cryptol.Utils.Ident (Ident)
+import Cryptol.Utils.RecordMap
+import Cryptol.Utils.Types
 
 import Data.Maybe(fromMaybe)
-import qualified Data.Map.Strict as Map
+import qualified Data.IntMap.Strict as IntMap
 import GHC.Generics (Generic)
-import GHC.Stack(HasCallStack)
 import Control.DeepSeq
 
 -- | An evaluated type of kind *.
 -- These types do not contain type variables, type synonyms, or type functions.
 data TValue
-  = TVBit                     -- ^ @ Bit @
+  = TVBit                     -- ^ @ Bit @  
   | TVInteger                 -- ^ @ Integer @
+  | TVFloat Integer Integer   -- ^ @ Float e p @
   | TVIntMod Integer          -- ^ @ Z n @
+  | TVRational                -- ^ @Rational@
+  | TVArray TValue TValue     -- ^ @ Array a b @
   | TVSeq Integer TValue      -- ^ @ [n]a @
   | TVStream TValue           -- ^ @ [inf]t @
   | TVTuple [TValue]          -- ^ @ (a, b, c )@
-  | TVRec [(Ident, TValue)]   -- ^ @ { x : a, y : b, z : c } @
+  | TVRec (RecordMap Ident TValue) -- ^ @ { x : a, y : b, z : c } @
   | TVFun TValue TValue       -- ^ @ a -> b @
+  | TVNewtype Newtype
+              [Either Nat' TValue]
+              (RecordMap Ident TValue)     -- ^ a named newtype
   | TVAbstract UserTC [Either Nat' TValue] -- ^ an abstract type
-    deriving (Generic, NFData)
+    deriving (Generic, NFData, Eq)
 
 -- | Convert a type value back into a regular type
 tValTy :: TValue -> Type
@@ -44,17 +51,24 @@ tValTy tv =
   case tv of
     TVBit       -> tBit
     TVInteger   -> tInteger
+    TVFloat e p -> tFloat (tNum e) (tNum p)
     TVIntMod n  -> tIntMod (tNum n)
+    TVRational  -> tRational
+    TVArray a b -> tArray (tValTy a) (tValTy b)
     TVSeq n t   -> tSeq (tNum n) (tValTy t)
     TVStream t  -> tSeq tInf (tValTy t)
     TVTuple ts  -> tTuple (map tValTy ts)
-    TVRec fs    -> tRec [ (f, tValTy t) | (f, t) <- fs ]
+    TVRec fs    -> tRec (fmap tValTy fs)
     TVFun t1 t2 -> tFun (tValTy t1) (tValTy t2)
-    TVAbstract u vs -> tAbstract u (map arg vs)
-      where arg x = case x of
-                      Left Inf     -> tInf
-                      Left (Nat n) -> tNum n
-                      Right v      -> tValTy v
+    TVNewtype nt vs _ -> tNewtype nt (map tNumValTy vs)
+    TVAbstract u vs -> tAbstract u (map tNumValTy vs)
+
+tNumTy :: Nat' -> Type
+tNumTy Inf     = tInf
+tNumTy (Nat n) = tNum n
+
+tNumValTy :: Either Nat' TValue -> Type
+tNumValTy = either tNumTy tValTy
 
 
 instance Show TValue where
@@ -73,6 +87,10 @@ tvSeq :: Nat' -> TValue -> TValue
 tvSeq (Nat n) t = TVSeq n t
 tvSeq Inf     t = TVStream t
 
+-- | The Cryptol @Float64@ type.
+tvFloat64 :: TValue
+tvFloat64 = uncurry TVFloat float64ExpPrec
+
 -- | Coerce an extended natural into an integer,
 --   for values known to be finite
 finNat' :: Nat' -> Integer
@@ -84,27 +102,49 @@ finNat' n' =
 
 -- Type Evaluation -------------------------------------------------------------
 
-type TypeEnv = Map.Map TVar (Either Nat' TValue)
+newtype TypeEnv =
+  TypeEnv
+  { envTypeMap  :: IntMap.IntMap (Either Nat' TValue) }
+  deriving (Show)
 
+instance Monoid TypeEnv where
+  mempty = TypeEnv mempty
+
+instance Semigroup TypeEnv where
+  l <> r = TypeEnv
+    { envTypeMap  = IntMap.union (envTypeMap l) (envTypeMap r) }
+
+lookupTypeVar :: TVar -> TypeEnv -> Maybe (Either Nat' TValue)
+lookupTypeVar tv env = IntMap.lookup (tvUnique tv) (envTypeMap env)
+
+bindTypeVar :: TVar -> Either Nat' TValue -> TypeEnv -> TypeEnv
+bindTypeVar tv ty env = env{ envTypeMap = IntMap.insert (tvUnique tv) ty (envTypeMap env) }
 
 -- | Evaluation for types (kind * or #).
-evalType :: HasCallStack => TypeEnv -> Type -> Either Nat' TValue
+evalType :: TypeEnv -> Type -> Either Nat' TValue
 evalType env ty =
   case ty of
     TVar tv ->
-      case Map.lookup tv env of
+      case lookupTypeVar tv env of
         Just v -> v
         Nothing -> evalPanic "evalType" ["type variable not bound", show tv]
 
     TUser _ _ ty'  -> evalType env ty'
-    TRec fields    -> Right $ TVRec [ (f, val t) | (f, t) <- fields ]
+    TRec fields    -> Right $ TVRec (fmap val fields)
+
+    TNewtype nt ts -> Right $ TVNewtype nt tvs $ evalNewtypeBody env nt tvs
+        where tvs = map (evalType env) ts
+
     TCon (TC c) ts ->
       case (c, ts) of
         (TCBit, [])     -> Right $ TVBit
         (TCInteger, []) -> Right $ TVInteger
+        (TCRational, []) -> Right $ TVRational
+        (TCFloat, [e,p])-> Right $ TVFloat (inum e) (inum p)
         (TCIntMod, [n]) -> case num n of
                              Inf   -> evalPanic "evalType" ["invalid type Z inf"]
                              Nat m -> Right $ TVIntMod m
+        (TCArray, [a, b]) -> Right $ TVArray (val a) (val b)
         (TCSeq, [n, t]) -> Right $ tvSeq (num n) (val t)
         (TCFun, [a, b]) -> Right $ TVFun (val a) (val b)
         (TCTuple _, _)  -> Right $ TVTuple (map val ts)
@@ -119,25 +159,38 @@ evalType env ty =
                 , "*** Name: " ++ show (pp u)
                 ]
 
-        -- FIXME: What about TCNewtype?
         _ -> evalPanic "evalType" ["not a value type", show ty]
     TCon (TF f) ts      -> Left $ evalTF f (map num ts)
     TCon (PC p) _       -> evalPanic "evalType" ["invalid predicate symbol", show p]
-    TCon (TError _ x) _ -> evalPanic "evalType"
-                                ["Lingering typer error", show (pp x)]
+    TCon (TError _) ts -> evalPanic "evalType"
+                             $ "Lingering invalid type" : map (show . pp) ts
   where
     val = evalValType env
     num = evalNumType env
+    inum x = case num x of
+               Nat i -> i
+               Inf   -> evalPanic "evalType"
+                                  ["Expecting a finite size, but got `inf`"]
+
+-- | Evaluate the body of a newtype, given evaluated arguments
+evalNewtypeBody :: TypeEnv -> Newtype -> [Either Nat' TValue] -> RecordMap Ident TValue
+evalNewtypeBody env0 nt args = fmap (evalValType env') (ntFields nt)
+  where
+  env' = loop env0 (ntParams nt) args
+
+  loop env [] [] = env
+  loop env (p:ps) (a:as) = loop (bindTypeVar (TVBound p) a env) ps as
+  loop _ _ _ = evalPanic "evalNewtype" ["type parameter/argument mismatch"]
 
 -- | Evaluation for value types (kind *).
-evalValType :: HasCallStack => TypeEnv -> Type -> TValue
+evalValType :: TypeEnv -> Type -> TValue
 evalValType env ty =
   case evalType env ty of
     Left _ -> evalPanic "evalValType" ["expected value type, found numeric type"]
     Right t -> t
 
 -- | Evaluation for number types (kind #).
-evalNumType :: HasCallStack => TypeEnv -> Type -> Nat'
+evalNumType :: TypeEnv -> Type -> Nat'
 evalNumType env ty =
   case evalType env ty of
     Left n -> n
@@ -145,7 +198,7 @@ evalNumType env ty =
 
 
 -- | Reduce type functions, raising an exception for undefined values.
-evalTF :: HasCallStack => TFun -> [Nat'] -> Nat'
+evalTF :: TFun -> [Nat'] -> Nat'
 evalTF f vs
   | TCAdd           <- f, [x,y]   <- vs  =      nAdd x y
   | TCSub           <- f, [x,y]   <- vs  = mb $ nSub x y
@@ -162,5 +215,5 @@ evalTF f vs
   | otherwise  = evalPanic "evalTF"
                         ["Unexpected type function:", show ty]
 
-  where mb = fromMaybe (typeCannotBeDemoted ty)
+  where mb = fromMaybe (evalPanic "evalTF" ["type cannot be demoted", show (pp ty)])
         ty = TCon (TF f) (map tNat' vs)
