@@ -9,9 +9,9 @@ import           Data.Set ( Set )
 import qualified Data.Set as Set
 import           Data.Map ( Map )
 import qualified Data.Map as Map
-import           Data.Semigroup (Semigroup(..))
 
 import Cryptol.TypeCheck.AST
+import Cryptol.Utils.RecordMap
 
 data Deps = Deps { valDeps  :: Set Name
                    -- ^ Undefined value names
@@ -32,7 +32,7 @@ instance Monoid Deps where
                 , tyParams  = Set.empty
                 }
 
-  mappend d1 d2 = d1 <> d2
+  mappend = (<>)
 
   mconcat ds = Deps { valDeps   = Set.unions (map valDeps ds)
                     , tyDeps    = Set.unions (map tyDeps  ds)
@@ -93,24 +93,22 @@ instance FreeVars Decl where
 instance FreeVars DeclDef where
   freeVars d = case d of
                  DPrim -> mempty
+                 DForeign _ -> mempty
                  DExpr e -> freeVars e
 
 
 instance FreeVars Expr where
   freeVars expr =
     case expr of
+      ELocated _r t     -> freeVars t
       EList es t        -> freeVars es <> freeVars t
       ETuple es         -> freeVars es
-      ERec fs           -> freeVars (map snd fs)
+      ERec fs           -> freeVars (recordElements fs)
       ESel e _          -> freeVars e
-      ESet e _ v        -> freeVars [e,v]
+      ESet ty e _ v     -> freeVars ty <> freeVars [e,v]
       EIf e1 e2 e3      -> freeVars [e1,e2,e3]
       EComp t1 t2 e mss -> freeVars [t1,t2] <> rmVals (defs mss) (freeVars e)
-                                            <> mconcat (map fvsArm mss)
-        where
-        fvsArm     = foldr mat mempty
-        mat x rest = freeVars x <> rmVals (defs x) rest
-
+                                            <> mconcat (map foldFree mss)
       EVar x            -> mempty { valDeps = Set.singleton x }
       ETAbs a e         -> rmTParam a (freeVars e)
       ETApp e t         -> freeVars e <> freeVars t
@@ -118,8 +116,12 @@ instance FreeVars Expr where
       EAbs x t e        -> freeVars t <> rmVal x (freeVars e)
       EProofAbs p e     -> freeVars p <> freeVars e
       EProofApp e       -> freeVars e
-      EWhere e ds       -> freeVars ds <> rmVals (defs ds) (freeVars e)
-
+      EWhere e ds       -> foldFree ds <> rmVals (defs ds) (freeVars e)
+      EPropGuards guards _ -> mconcat [ freeVars e | (_, e) <- guards ]
+    where
+      foldFree :: (FreeVars a, Defs a) => [a] -> Deps
+      foldFree = foldr updateFree mempty
+      updateFree x rest = freeVars x <> rmVals (defs x) rest
 
 instance FreeVars Match where
   freeVars m = case m of
@@ -139,8 +141,8 @@ instance FreeVars Type where
       TVar tv -> freeVars tv
 
       TUser _ _ t -> freeVars t
-      TRec fs     -> freeVars (map snd fs)
-
+      TRec fs     -> freeVars (recordElements fs)
+      TNewtype nt ts -> freeVars nt <> freeVars ts
 
 instance FreeVars TVar where
   freeVars tv = case tv of
@@ -148,14 +150,11 @@ instance FreeVars TVar where
                   _         -> mempty
 
 instance FreeVars TCon where
-  freeVars tc =
-    case tc of
-     TC (TCNewtype (UserTC n _)) -> mempty { tyDeps = Set.singleton n }
-     _ -> mempty
+  freeVars _tc = mempty
 
 instance FreeVars Newtype where
   freeVars nt = foldr rmTParam base (ntParams nt)
-    where base = freeVars (ntConstraints nt) <> freeVars (map snd (ntFields nt))
+    where base = freeVars (ntConstraints nt) <> freeVars (recordElements (ntFields nt))
 
 
 --------------------------------------------------------------------------------
@@ -178,6 +177,3 @@ instance Defs Match where
   defs m = case m of
              From x _ _ _ -> Set.singleton x
              Let d -> defs d
-
-
-
