@@ -10,14 +10,15 @@ module Cryptol.ModuleSystem.Fingerprint
   ( Fingerprint
   , fingerprint
   , fingerprintFile
+  , fingerprintHexString
   ) where
 
 import Control.DeepSeq          (NFData (rnf))
 import Crypto.Hash.SHA1         (hash)
 import Data.ByteString          (ByteString)
-import System.IO.Error          (IOError)
 import Control.Exception        (try)
 import qualified Data.ByteString as B
+import qualified Data.Vector as Vector
 
 newtype Fingerprint = Fingerprint ByteString
   deriving (Eq, Show)
@@ -38,3 +39,13 @@ fingerprintFile path =
        case res :: Either IOError ByteString of
          Left{}  -> Nothing
          Right b -> Just $! fingerprint b
+
+fingerprintHexString :: Fingerprint -> String
+fingerprintHexString (Fingerprint bs) = B.foldr hex "" bs
+  where
+  digits   = Vector.fromList "0123456789ABCDEF"
+  digit x  = digits Vector.! fromIntegral x
+  hex b cs = let (x,y) = divMod b 16
+             in digit x : digit y : cs
+
+
