@@ -10,35 +10,46 @@
 {-# LANGUAGE DeriveAnyClass #-}
 {-# LANGUAGE DeriveGeneric #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE BlockArguments #-}
 module Cryptol.ModuleSystem.Monad where
 
 import           Cryptol.Eval (EvalEnv,EvalOpts(..))
 
-import qualified Cryptol.Eval.Monad           as E
+import           Cryptol.Backend.FFI (ForeignSrc)
+import           Cryptol.Backend.FFI.Error
+import qualified Cryptol.Backend.Monad           as E
+
 import           Cryptol.ModuleSystem.Env
-import           Cryptol.ModuleSystem.Fingerprint
+import qualified Cryptol.ModuleSystem.Env as MEnv
 import           Cryptol.ModuleSystem.Interface
 import           Cryptol.ModuleSystem.Name (FreshM(..),Supply)
-import           Cryptol.ModuleSystem.Renamer
-                     (RenamerError(),RenamerWarning(),NamingEnv)
+import           Cryptol.ModuleSystem.Renamer (RenamerError(),RenamerWarning())
+import           Cryptol.ModuleSystem.NamingEnv(NamingEnv)
 import qualified Cryptol.Parser     as Parser
 import qualified Cryptol.Parser.AST as P
-import           Cryptol.Parser.Position (Located)
 import           Cryptol.Utils.Panic (panic)
 import qualified Cryptol.Parser.NoPat as NoPat
+import qualified Cryptol.Parser.ExpandPropGuards as ExpandPropGuards
 import qualified Cryptol.Parser.NoInclude as NoInc
 import qualified Cryptol.TypeCheck as T
 import qualified Cryptol.TypeCheck.AST as T
-import           Cryptol.Parser.Position (Range)
+import qualified Cryptol.TypeCheck.Solver.SMT as SMT
+
+import           Cryptol.Parser.Position (Range, Located)
 import           Cryptol.Utils.Ident (interactiveName, noModuleName)
 import           Cryptol.Utils.PP
 import           Cryptol.Utils.Logger(Logger)
 
+import qualified Control.Monad.Fail as Fail
 import Control.Monad.IO.Class
 import Control.Exception (IOException)
+import Data.ByteString (ByteString)
 import Data.Function (on)
-import Data.Maybe (isJust)
+import Data.Functor.Identity
+import Data.Map (Map)
 import Data.Text.Encoding.Error (UnicodeException)
+import Data.Traversable
 import MonadLib
 import System.Directory (canonicalizePath)
 
@@ -54,6 +65,7 @@ import Prelude.Compat
 data ImportSource
   = FromModule P.ModName
   | FromImport (Located P.Import)
+  | FromSigImport (Located P.ModName)
   | FromModuleInstance (Located P.ModName)
     deriving (Show, Generic, NFData)
 
@@ -64,6 +76,7 @@ instance PP ImportSource where
   ppPrec _ is = case is of
     FromModule n  -> text "module name" <+> pp n
     FromImport li -> text "import of module" <+> pp (P.iModule (P.thing li))
+    FromSigImport l -> text "import of interface" <+> pp (P.thing l)
     FromModuleInstance l ->
       text "instantiation of module" <+> pp (P.thing l)
 
@@ -73,6 +86,7 @@ importedModule is =
     FromModule n          -> n
     FromImport li         -> P.iModule (P.thing li)
     FromModuleInstance l  -> P.thing l
+    FromSigImport l       -> P.thing l
 
 
 data ModuleError
@@ -92,9 +106,11 @@ data ModuleError
     -- ^ Problems during the renaming phase
   | NoPatErrors ImportSource [NoPat.Error]
     -- ^ Problems during the NoPat phase
+  | ExpandPropGuardsError ImportSource ExpandPropGuards.Error
+    -- ^ Problems during the ExpandPropGuards phase
   | NoIncludeErrors ImportSource [NoInc.IncludeError]
     -- ^ Problems during the NoInclude phase
-  | TypeCheckingFailed ImportSource [(Range,T.Error)]
+  | TypeCheckingFailed ImportSource T.NameMap [(Range,T.Error)]
     -- ^ Problems during type checking
   | OtherFailure String
     -- ^ Problems after type checking, eg. specialization
@@ -102,11 +118,8 @@ data ModuleError
     -- ^ Module loaded by 'import' statement has the wrong module name
   | DuplicateModuleName P.ModName FilePath FilePath
     -- ^ Two modules loaded from different files have the same module name
-  | ImportedParamModule P.ModName
-    -- ^ Attempt to import a parametrized module that was not instantiated.
-  | FailedToParameterizeModDefs P.ModName [T.Name]
-    -- ^ Failed to add the module parameters to all definitions in a module.
-  | NotAParameterizedModule P.ModName
+  | FFILoadErrors P.ModName [FFILoadError]
+    -- ^ Errors loading foreign function implementations
 
   | ErrorInFile ModulePath ModuleError
     -- ^ This is just a tag on the error, indicating the file containing it.
@@ -125,16 +138,15 @@ instance NFData ModuleError where
     RecursiveModules mods                -> mods `deepseq` ()
     RenamerErrors src errs               -> src `deepseq` errs `deepseq` ()
     NoPatErrors src errs                 -> src `deepseq` errs `deepseq` ()
+    ExpandPropGuardsError src err        -> src `deepseq` err `deepseq` ()
     NoIncludeErrors src errs             -> src `deepseq` errs `deepseq` ()
-    TypeCheckingFailed src errs          -> src `deepseq` errs `deepseq` ()
+    TypeCheckingFailed nm src errs       -> nm `deepseq` src `deepseq` errs `deepseq` ()
     ModuleNameMismatch expected found    ->
       expected `deepseq` found `deepseq` ()
     DuplicateModuleName name path1 path2 ->
       name `deepseq` path1 `deepseq` path2 `deepseq` ()
     OtherFailure x                       -> x `deepseq` ()
-    ImportedParamModule x                -> x `deepseq` ()
-    FailedToParameterizeModDefs x xs     -> x `deepseq` xs `deepseq` ()
-    NotAParameterizedModule x            -> x `deepseq` ()
+    FFILoadErrors x errs                 -> x `deepseq` errs `deepseq` ()
     ErrorInFile x y                      -> x `deepseq` y `deepseq` ()
 
 instance PP ModuleError where
@@ -172,14 +184,16 @@ instance PP ModuleError where
 
     NoPatErrors _src errs -> vcat (map pp errs)
 
+    ExpandPropGuardsError _src err -> pp err
+
     NoIncludeErrors _src errs -> vcat (map NoInc.ppIncludeError errs)
 
-    TypeCheckingFailed _src errs -> vcat (map T.ppError errs)
+    TypeCheckingFailed _src nm errs -> vcat (map (T.ppNamedError nm) errs)
 
     ModuleNameMismatch expected found ->
       hang (text "[error]" <+> pp (P.srcRange found) <.> char ':')
          4 (vcat [ text "File name does not match module name:"
-                 , text "Saw:"      <+> pp (P.thing found)
+                 , text "  Actual:" <+> pp (P.thing found)
                  , text "Expected:" <+> pp expected
                  ])
 
@@ -190,16 +204,10 @@ instance PP ModuleError where
 
     OtherFailure x -> text x
 
-    ImportedParamModule p ->
-      text "[error] Import of a non-instantiated parameterized module:" <+> pp p
-
-    FailedToParameterizeModDefs x xs ->
-      hang (text "[error] Parameterized module" <+> pp x <+>
-            text "has polymorphic parameters:")
-        4 (hsep $ punctuate comma $ map pp xs)
-
-    NotAParameterizedModule x ->
-      text "[error] Module" <+> pp x <+> text "does not have parameters."
+    FFILoadErrors x errs ->
+      hang (text "[error] Failed to load foreign implementations for module"
+            <+> pp x <.> colon)
+         4 (vcat $ map pp errs)
 
     ErrorInFile _ x -> ppPrec prec x
 
@@ -232,15 +240,20 @@ noPatErrors errs = do
   src <- getImportSource
   ModuleT (raise (NoPatErrors src errs))
 
+expandPropGuardsError :: ExpandPropGuards.Error -> ModuleM a
+expandPropGuardsError err = do
+  src <- getImportSource
+  ModuleT (raise (ExpandPropGuardsError src err))
+
 noIncludeErrors :: [NoInc.IncludeError] -> ModuleM a
 noIncludeErrors errs = do
   src <- getImportSource
   ModuleT (raise (NoIncludeErrors src errs))
 
-typeCheckingFailed :: [(Range,T.Error)] -> ModuleM a
-typeCheckingFailed errs = do
+typeCheckingFailed :: T.NameMap -> [(Range,T.Error)] -> ModuleM a
+typeCheckingFailed nameMap errs = do
   src <- getImportSource
-  ModuleT (raise (TypeCheckingFailed src errs))
+  ModuleT (raise (TypeCheckingFailed src nameMap errs))
 
 moduleNameMismatch :: P.ModName -> Located P.ModName -> ModuleM a
 moduleNameMismatch expected found =
@@ -250,15 +263,8 @@ duplicateModuleName :: P.ModName -> FilePath -> FilePath -> ModuleM a
 duplicateModuleName name path1 path2 =
   ModuleT (raise (DuplicateModuleName name path1 path2))
 
-importParamModule :: P.ModName -> ModuleM a
-importParamModule x = ModuleT (raise (ImportedParamModule x))
-
-failedToParameterizeModDefs :: P.ModName -> [T.Name] -> ModuleM a
-failedToParameterizeModDefs x xs =
-  ModuleT (raise (FailedToParameterizeModDefs x xs))
-
-notAParameterizedModule :: P.ModName -> ModuleM a
-notAParameterizedModule x = ModuleT (raise (NotAParameterizedModule x))
+ffiLoadErrors :: P.ModName -> [FFILoadError] -> ModuleM a
+ffiLoadErrors x errs = ModuleT (raise (FFILoadErrors x errs))
 
 -- | Run the computation, and if it caused and error, tag the error
 -- with the given file.
@@ -271,22 +277,22 @@ errorInFile file (ModuleT m) = ModuleT (m `handle` h)
 -- Warnings --------------------------------------------------------------------
 
 data ModuleWarning
-  = TypeCheckWarnings [(Range,T.Warning)]
+  = TypeCheckWarnings T.NameMap [(Range,T.Warning)]
   | RenamerWarnings [RenamerWarning]
     deriving (Show, Generic, NFData)
 
 instance PP ModuleWarning where
   ppPrec _ w = case w of
-    TypeCheckWarnings ws -> vcat (map T.ppWarning ws)
+    TypeCheckWarnings nm ws -> vcat (map (T.ppNamedWarning nm) ws)
     RenamerWarnings ws   -> vcat (map pp ws)
 
 warn :: [ModuleWarning] -> ModuleM ()
 warn  = ModuleT . put
 
-typeCheckWarnings :: [(Range,T.Warning)] -> ModuleM ()
-typeCheckWarnings ws
+typeCheckWarnings :: T.NameMap -> [(Range,T.Warning)] -> ModuleM ()
+typeCheckWarnings nameMap ws
   | null ws   = return ()
-  | otherwise = warn [TypeCheckWarnings ws]
+  | otherwise = warn [TypeCheckWarnings nameMap ws]
 
 renamerWarnings :: [RenamerWarning] -> ModuleM ()
 renamerWarnings ws
@@ -296,16 +302,28 @@ renamerWarnings ws
 
 -- Module System Monad ---------------------------------------------------------
 
-data RO = RO { roLoading  :: [ImportSource]
-             , roEvalOpts :: EvalOpts
-             }
+data RO m =
+  RO { roLoading    :: [ImportSource]
+     , roEvalOpts   :: m EvalOpts
+     , roCallStacks :: Bool
+     , roFileReader :: FilePath -> m ByteString
+     , roTCSolver   :: SMT.Solver
+     }
 
-emptyRO :: EvalOpts -> RO
-emptyRO ev = RO { roLoading = [], roEvalOpts = ev }
+emptyRO :: ModuleInput m -> RO m
+emptyRO minp =
+  RO { roLoading = []
+     , roEvalOpts   = minpEvalOpts minp
+     , roCallStacks = minpCallStacks minp
+     , roFileReader = minpByteReader minp
+     , roTCSolver   = minpTCSolver minp
+     }
 
 newtype ModuleT m a = ModuleT
-  { unModuleT :: ReaderT RO (StateT ModuleEnv
-                    (ExceptionT ModuleError (WriterT [ModuleWarning] m))) a
+  { unModuleT :: ReaderT (RO m)
+                   (StateT ModuleEnv
+                     (ExceptionT ModuleError
+                       (WriterT [ModuleWarning] m))) a
   }
 
 instance Monad m => Functor (ModuleT m) where
@@ -321,10 +339,12 @@ instance Monad m => Applicative (ModuleT m) where
 
 instance Monad m => Monad (ModuleT m) where
   {-# INLINE return #-}
-  return x      = ModuleT (return x)
+  return        = pure
 
   {-# INLINE (>>=) #-}
   m >>= f       = ModuleT (unModuleT m >>= unModuleT . f)
+
+instance Fail.MonadFail m => Fail.MonadFail (ModuleT m) where
   {-# INLINE fail #-}
   fail          = ModuleT . raise . OtherFailure
 
@@ -342,31 +362,59 @@ instance Monad m => FreshM (ModuleT m) where
 instance MonadIO m => MonadIO (ModuleT m) where
   liftIO m = lift $ liftIO m
 
-runModuleT :: Monad m
-           => (EvalOpts,ModuleEnv)
-           -> ModuleT m a
-           -> m (Either ModuleError (a, ModuleEnv), [ModuleWarning])
-runModuleT (ev,env) m =
+
+data ModuleInput m =
+  ModuleInput
+  { minpCallStacks :: Bool
+  , minpEvalOpts   :: m EvalOpts
+  , minpByteReader :: FilePath -> m ByteString
+  , minpModuleEnv  :: ModuleEnv
+  , minpTCSolver   :: SMT.Solver
+  }
+
+runModuleT ::
+  Monad m =>
+  ModuleInput m ->
+  ModuleT m a ->
+  m (Either ModuleError (a, ModuleEnv), [ModuleWarning])
+runModuleT minp m =
     runWriterT
   $ runExceptionT
-  $ runStateT env
-  $ runReaderT (emptyRO ev)
+  $ runStateT (minpModuleEnv minp)
+  $ runReaderT (emptyRO minp)
   $ unModuleT m
 
 type ModuleM = ModuleT IO
 
-runModuleM :: (EvalOpts, ModuleEnv) -> ModuleM a
-           -> IO (Either ModuleError (a,ModuleEnv),[ModuleWarning])
+runModuleM ::
+  ModuleInput IO ->
+  ModuleM a ->
+  IO (Either ModuleError (a,ModuleEnv),[ModuleWarning])
 runModuleM = runModuleT
 
 
-
 io :: BaseM m IO => IO a -> ModuleT m a
 io m = ModuleT (inBase m)
 
+getByteReader :: Monad m => ModuleT m (FilePath -> m ByteString)
+getByteReader = ModuleT $ do
+  RO { roFileReader = readFileBytes } <- ask
+  return readFileBytes
+
+getCallStacks :: Monad m => ModuleT m Bool
+getCallStacks = ModuleT (roCallStacks <$> ask)
+
+readBytes :: Monad m => FilePath -> ModuleT m ByteString
+readBytes fn = do
+  fileReader <- getByteReader
+  ModuleT $ lift $ lift $ lift $ lift $ fileReader fn
+
 getModuleEnv :: Monad m => ModuleT m ModuleEnv
 getModuleEnv = ModuleT get
 
+getTCSolver :: Monad m => ModuleT m SMT.Solver
+getTCSolver = ModuleT (roTCSolver <$> ask)
+
 setModuleEnv :: Monad m => ModuleEnv -> ModuleT m ()
 setModuleEnv = ModuleT . set
 
@@ -375,13 +423,17 @@ modifyModuleEnv f = ModuleT $ do
   env <- get
   set $! f env
 
-getLoadedMaybe :: P.ModName -> ModuleM (Maybe LoadedModule)
+getLoadedMaybe :: P.ModName -> ModuleM (Maybe (LoadedModuleG T.TCTopEntity))
 getLoadedMaybe mn = ModuleT $
   do env <- get
-     return (lookupModule mn env)
+     return (lookupTCEntity mn env)
 
+-- | This checks if the given name is loaded---it might refer to either
+-- a module or a signature.
 isLoaded :: P.ModName -> ModuleM Bool
-isLoaded mn = isJust <$> getLoadedMaybe mn
+isLoaded mn =
+  do env <- ModuleT get
+     pure (MEnv.isLoaded mn (meLoadedModules env))
 
 loadingImport :: Located P.Import -> ModuleM a -> ModuleM a
 loadingImport  = loading . FromImport
@@ -400,12 +452,12 @@ interactive  = loadingModule interactiveName
 loading :: ImportSource -> ModuleM a -> ModuleM a
 loading src m = ModuleT $ do
   ro <- ask
-  let ro'  = ro { roLoading = src : roLoading ro }
+  let new = src : roLoading ro
 
   -- check for recursive modules
-  when (src `elem` roLoading ro) (raise (RecursiveModules (roLoading ro')))
+  when (src `elem` roLoading ro) (raise (RecursiveModules new))
 
-  local ro' (unModuleT m)
+  local ro { roLoading = new } (unModuleT m)
 
 -- | Get the currently focused import source.
 getImportSource :: ModuleM ImportSource
@@ -415,12 +467,16 @@ getImportSource  = ModuleT $ do
     is : _ -> return is
     _      -> return (FromModule noModuleName)
 
-getIface :: P.ModName -> ModuleM Iface
-getIface mn =
-  do env <- ModuleT get
-     case lookupModule mn env of
-       Just lm -> return (lmInterface lm)
-       Nothing -> panic "ModuleSystem" ["Interface not available", show (pp mn)]
+getIfaces :: ModuleM (Map P.ModName (Either T.ModParamNames Iface))
+getIfaces = toMap <$> ModuleT get
+  where
+  toMap env = cvt <$> getLoadedEntities (meLoadedModules env)
+
+  cvt ent =
+    case ent of
+      ALoadedInterface ifa -> Left (lmData ifa)
+      ALoadedFunctor mo -> Right (lmdInterface (lmData mo))
+      ALoadedModule mo -> Right (lmdInterface (lmData mo))
 
 getLoaded :: P.ModName -> ModuleM T.Module
 getLoaded mn = ModuleT $
@@ -429,6 +485,20 @@ getLoaded mn = ModuleT $
        Just lm -> return (lmModule lm)
        Nothing -> panic "ModuleSystem" ["Module not available", show (pp mn) ]
 
+getAllLoaded :: ModuleM (P.ModName -> Maybe (T.ModuleG (), IfaceG ()))
+getAllLoaded = ModuleT
+  do env <- get
+     pure \nm -> do lm <- lookupModule nm env
+                    pure ( (lmModule lm) { T.mName = () }
+                         , ifaceForgetName (lmInterface lm)
+                         )
+
+getAllLoadedSignatures :: ModuleM (P.ModName -> Maybe T.ModParamNames)
+getAllLoadedSignatures = ModuleT
+  do env <- get
+     pure \nm -> lmData <$> lookupSignature nm env
+
+
 getNameSeeds :: ModuleM T.NameSeeds
 getNameSeeds  = ModuleT (meNameSeeds `fmap` get)
 
@@ -453,33 +523,56 @@ setSupply supply = ModuleT $
   do env <- get
      set $! env { meSupply = supply }
 
-unloadModule :: (LoadedModule -> Bool) -> ModuleM ()
+unloadModule :: (forall a. LoadedModuleG a -> Bool) -> ModuleM ()
 unloadModule rm = ModuleT $ do
   env <- get
   set $! env { meLoadedModules = removeLoadedModule rm (meLoadedModules env) }
 
-loadedModule :: ModulePath -> Fingerprint -> T.Module -> ModuleM ()
-loadedModule path fp m = ModuleT $ do
+loadedModule ::
+  ModulePath ->
+  FileInfo ->
+  NamingEnv ->
+  Maybe ForeignSrc ->
+  T.TCTopEntity ->
+  ModuleM ()
+loadedModule path fi nameEnv fsrc m = ModuleT $ do
   env <- get
   ident <- case path of
              InFile p  -> unModuleT $ io (canonicalizePath p)
              InMem l _ -> pure l
 
-  set $! env { meLoadedModules = addLoadedModule path ident fp m (meLoadedModules env) }
+  let newLM =
+        case m of
+          T.TCTopModule mo -> addLoadedModule path ident fi nameEnv fsrc mo
+          T.TCTopSignature x s -> addLoadedSignature path ident fi nameEnv x s
 
-modifyEvalEnv :: (EvalEnv -> E.Eval EvalEnv) -> ModuleM ()
-modifyEvalEnv f = ModuleT $ do
+  set $! env { meLoadedModules = newLM (meLoadedModules env) }
+
+
+modifyEvalEnvM :: Traversable t =>
+  (EvalEnv -> E.Eval (t EvalEnv)) -> ModuleM (t ())
+modifyEvalEnvM f = ModuleT $ do
   env <- get
   let evalEnv = meEvalEnv env
-  evOpts <- unModuleT getEvalOpts
-  evalEnv' <- inBase $ E.runEval evOpts (f evalEnv)
-  set $! env { meEvalEnv = evalEnv' }
+  tenv <- inBase (E.runEval mempty (f evalEnv))
+  traverse (\evalEnv' -> set $! env { meEvalEnv = evalEnv' }) tenv
+
+modifyEvalEnv :: (EvalEnv -> E.Eval EvalEnv) -> ModuleM ()
+modifyEvalEnv = fmap runIdentity . modifyEvalEnvM . (fmap Identity .)
 
 getEvalEnv :: ModuleM EvalEnv
 getEvalEnv  = ModuleT (meEvalEnv `fmap` get)
 
+getEvalOptsAction :: ModuleM (IO EvalOpts)
+getEvalOptsAction = ModuleT (roEvalOpts `fmap` ask)
+
 getEvalOpts :: ModuleM EvalOpts
-getEvalOpts = ModuleT (roEvalOpts `fmap` ask)
+getEvalOpts =
+  do act <- getEvalOptsAction
+     liftIO act
+
+getNewtypes :: ModuleM (Map T.Name T.Newtype)
+getNewtypes = ModuleT (loadedNewtypes <$> get)
 
 getFocusedModule :: ModuleM (Maybe P.ModName)
 getFocusedModule  = ModuleT (meFocusedModule `fmap` get)
@@ -505,8 +598,7 @@ withPrependedSearchPath fps m = ModuleT $ do
   set $! env { meSearchPath = fps0 }
   return x
 
--- XXX improve error handling here
-getFocusedEnv :: ModuleM (IfaceParams,IfaceDecls,NamingEnv,NameDisp)
+getFocusedEnv :: ModuleM ModContext
 getFocusedEnv  = ModuleT (focusedEnv `fmap` get)
 
 getDynEnv :: ModuleM DynamicEnv
@@ -517,18 +609,7 @@ setDynEnv denv = ModuleT $ do
   me <- get
   set $! me { meDynEnv = denv }
 
-setSolver :: T.SolverConfig -> ModuleM ()
-setSolver cfg = ModuleT $ do
-  me <- get
-  set $! me { meSolverConfig = cfg }
-
-getSolverConfig :: ModuleM T.SolverConfig
-getSolverConfig  = ModuleT $ do
-  me <- get
-  return (meSolverConfig me)
-
 -- | Usefule for logging.  For example: @withLogger logPutStrLn "Hello"@
 withLogger :: (Logger -> a -> IO b) -> a -> ModuleM b
 withLogger f a = do l <- getEvalOpts
                     io (f (evalLogger l) a)
-
