@@ -9,18 +9,36 @@
 -- This module defines the scoping rules for value- and type-level
 -- names in Cryptol.
 
-module Cryptol.Parser.Names where
+{-# LANGUAGE Safe #-}
+
+module Cryptol.Parser.Names
+  ( tnamesNT
+  , tnamesT
+  , tnamesC
+
+  , namesD
+  , tnamesD
+  , namesB
+  , namesP
+  , namesNT
+
+  , boundNames
+  , boundNamesSet
+  ) where
 
 import Cryptol.Parser.AST
+import Cryptol.Utils.RecordMap
 
 import           Data.Set (Set)
 import qualified Data.Set as Set
 
-
 -- | The names defined by a newtype.
 tnamesNT :: Newtype name -> ([Located name], ())
 tnamesNT x = ([ nName x ], ())
 
+namesNT :: Newtype name -> ([Located name], ())
+namesNT x = ([ (nName x) { thing = nConName x } ], ())
+
 -- | The names defined and used by a group of mutually recursive declarations.
 namesDs :: Ord name => [Decl name] -> ([Located name], Set name)
 namesDs ds = (defs, boundLNames defs (Set.unions frees))
@@ -33,6 +51,8 @@ namesD :: Ord name => Decl name -> ([Located name], Set name)
 namesD decl =
   case decl of
     DBind b       -> namesB b
+    DRec bs       -> let (xs,ys) = unzip (map namesB bs)
+                     in (concat xs, Set.unions ys)  -- remove binders?
     DPatBind p e  -> (namesP p, namesE e)
     DSignature {} -> ([],Set.empty)
     DFixity{}     -> ([],Set.empty)
@@ -41,30 +61,17 @@ namesD decl =
     DProp {}      -> ([],Set.empty)
     DLocated d _  -> namesD d
 
--- | The names defined and used by a single declarations in such a way
--- that they cannot be duplicated in a file. For example, it is fine
--- to use @x@ on the RHS of two bindings, but not on the LHS of two
--- type signatures.
-allNamesD :: Ord name => Decl name -> [Located name]
-allNamesD decl =
-  case decl of
-    DBind b         -> fst (namesB b)
-    DPatBind p _    -> namesP p
-    DSignature ns _ -> ns
-    DFixity _ ns    -> ns
-    DPragma ns _    -> ns
-    DType ts        -> [tsName ts]
-    DProp ps        -> [psName ps]
-    DLocated d _    -> allNamesD d
-
 -- | The names defined and used by a single binding.
 namesB :: Ord name => Bind name -> ([Located name], Set name)
-namesB b = ([bName b], boundLNames (namesPs (bParams b)) (namesDef (thing (bDef b))))
+namesB b =
+  ([bName b], boundLNames (namesPs (bParams b)) (namesDef (thing (bDef b))))
 
 
 namesDef :: Ord name => BindDef name -> Set name
 namesDef DPrim     = Set.empty
+namesDef DForeign  = Set.empty
 namesDef (DExpr e) = namesE e
+namesDef (DPropGuards guards) = Set.unions (map (namesE . pgcExpr) guards)
 
 
 -- | The names used by an expression.
@@ -73,16 +80,17 @@ namesE expr =
   case expr of
     EVar x        -> Set.singleton x
     ELit _        -> Set.empty
-    ENeg e        -> namesE e
-    EComplement e -> namesE e
     EGenerate e   -> namesE e
     ETuple es     -> Set.unions (map namesE es)
-    ERecord fs    -> Set.unions (map (namesE . value) fs)
+    ERecord fs    -> Set.unions (map (namesE . snd) (recordElements fs))
     ESel e _      -> namesE e
     EUpd mb fs    -> let e = maybe Set.empty namesE mb
                      in Set.unions (e : map namesUF fs)
     EList es      -> Set.unions (map namesE es)
     EFromTo{}     -> Set.empty
+    EFromToBy{}   -> Set.empty
+    EFromToDownBy{} -> Set.empty
+    EFromToLessThan{} -> Set.empty
     EInfFrom e e' -> Set.union (namesE e) (maybe Set.empty namesE e')
     EComp e arms  -> let (dss,uss) = unzip (map namesArm arms)
                      in Set.union (boundLNames (concat dss) (namesE e))
@@ -94,12 +102,13 @@ namesE expr =
                      in Set.union (boundLNames bs (namesE e)) xs
     ETyped e _    -> namesE e
     ETypeVal _    -> Set.empty
-    EFun ps e     -> boundLNames (namesPs ps) (namesE e)
+    EFun _ ps e   -> boundLNames (namesPs ps) (namesE e)
     ELocated e _  -> namesE e
 
     ESplit e      -> namesE e
     EParens e     -> namesE e
     EInfix a o _ b-> Set.insert (thing o) (Set.union (namesE a) (namesE b))
+    EPrefix _ e   -> namesE e
 
 namesUF :: Ord name => UpdField name -> Set name
 namesUF (UpdField _ _ e) = namesE e
@@ -115,7 +124,7 @@ namesP pat =
     PVar x        -> [x]
     PWild         -> []
     PTuple ps     -> namesPs ps
-    PRecord fs    -> namesPs (map value fs)
+    PRecord fs    -> namesPs (map snd (recordElements fs))
     PList ps      -> namesPs ps
     PTyped p _    -> namesP p
     PSplit p1 p2  -> namesPs [p1,p2]
@@ -162,6 +171,7 @@ tnamesD decl =
     DFixity {}           -> ([], Set.empty)
     DPragma {}           -> ([], Set.empty)
     DBind b              -> ([], tnamesB b)
+    DRec bs              -> ([], Set.unions (map tnamesB bs))
     DPatBind _ e         -> ([], tnamesE e)
     DLocated d _         -> tnamesD d
     DType (TySyn n _ ps t)
@@ -181,7 +191,13 @@ tnamesB b = Set.unions [setS, setP, setE]
 
 tnamesDef :: Ord name => BindDef name -> Set name
 tnamesDef DPrim     = Set.empty
+tnamesDef DForeign  = Set.empty
 tnamesDef (DExpr e) = tnamesE e
+tnamesDef (DPropGuards guards) = Set.unions (map tnamesPropGuardCase guards)
+
+tnamesPropGuardCase :: Ord name => PropGuardCase name -> Set name
+tnamesPropGuardCase c =
+  Set.unions (tnamesE (pgcExpr c) : map (tnamesC . thing) (pgcProps c))
 
 -- | The type names used by an expression.
 tnamesE :: Ord name => Expr name -> Set name
@@ -189,11 +205,9 @@ tnamesE expr =
   case expr of
     EVar _          -> Set.empty
     ELit _          -> Set.empty
-    ENeg e          -> tnamesE e
-    EComplement e   -> tnamesE e
     EGenerate e     -> tnamesE e
     ETuple es       -> Set.unions (map tnamesE es)
-    ERecord fs      -> Set.unions (map (tnamesE . value) fs)
+    ERecord fs      -> Set.unions (map (tnamesE . snd) (recordElements fs))
     ESel e _        -> tnamesE e
     EUpd mb fs      -> let e = maybe Set.empty tnamesE mb
                        in Set.unions (e : map tnamesUF fs)
@@ -202,6 +216,10 @@ tnamesE expr =
                        `Set.union` maybe Set.empty tnamesT b
                        `Set.union` tnamesT c
                        `Set.union` maybe Set.empty tnamesT t
+    EFromToBy _ a b c t -> Set.unions [ tnamesT a, tnamesT b, tnamesT c, maybe Set.empty tnamesT t ]
+    EFromToDownBy _ a b c t -> Set.unions [ tnamesT a, tnamesT b, tnamesT c, maybe Set.empty tnamesT t ]
+    EFromToLessThan a b t -> tnamesT a `Set.union` tnamesT b
+                                       `Set.union` maybe Set.empty tnamesT t
     EInfFrom e e'   -> Set.union (tnamesE e) (maybe Set.empty tnamesE e')
     EComp e mss     -> Set.union (tnamesE e) (Set.unions (map tnamesM (concat mss)))
     EApp e1 e2      -> Set.union (tnamesE e1) (tnamesE e2)
@@ -211,12 +229,13 @@ tnamesE expr =
                        in Set.union (boundLNames bs (tnamesE e)) xs
     ETyped e t      -> Set.union (tnamesE e) (tnamesT t)
     ETypeVal t      -> tnamesT t
-    EFun ps e       -> Set.union (Set.unions (map tnamesP ps)) (tnamesE e)
+    EFun _ ps e     -> Set.union (Set.unions (map tnamesP ps)) (tnamesE e)
     ELocated e _    -> tnamesE e
 
     ESplit e        -> tnamesE e
     EParens e       -> tnamesE e
     EInfix a _ _ b  -> Set.union (tnamesE a) (tnamesE b)
+    EPrefix _ e     -> tnamesE e
 
 tnamesUF :: Ord name => UpdField name -> Set name
 tnamesUF (UpdField _ _ e) = tnamesE e
@@ -232,7 +251,7 @@ tnamesP pat =
     PVar _        -> Set.empty
     PWild         -> Set.empty
     PTuple ps     -> Set.unions (map tnamesP ps)
-    PRecord fs    -> Set.unions (map (tnamesP . value) fs)
+    PRecord fs    -> Set.unions (map (tnamesP . snd) (recordElements fs))
     PList ps      -> Set.unions (map tnamesP ps)
     PTyped p t    -> Set.union (tnamesP p) (tnamesT t)
     PSplit p1 p2  -> Set.union (tnamesP p1) (tnamesP p2)
@@ -266,9 +285,10 @@ tnamesT ty =
     TNum _        -> Set.empty
     TChar __      -> Set.empty
     TTuple ts     -> Set.unions (map tnamesT ts)
-    TRecord fs    -> Set.unions (map (tnamesT . value) fs)
+    TRecord fs    -> Set.unions (map (tnamesT . snd) (recordElements fs))
+    TTyApp fs     -> Set.unions (map (tnamesT . value) fs)
     TLocated t _  -> tnamesT t
     TUser x ts    -> Set.insert x (Set.unions (map tnamesT ts))
-    TParens t     -> tnamesT t
+    TParens t _   -> tnamesT t
     TInfix a x _ c-> Set.insert (thing x)
                                 (Set.union (tnamesT a) (tnamesT c))
