@@ -8,8 +8,8 @@
 --
 -- The purpose of this module is to convert all patterns to variable
 -- patterns.  It also eliminates pattern bindings by de-sugaring them
--- into `Bind`.  Furthermore, here we associate signatures and pragmas
--- with the names to which they belong.
+-- into `Bind`.  Furthermore, here we associate signatures, fixities,
+-- and pragmas with the names to which they belong.
 
 {-# LANGUAGE DeriveAnyClass #-}
 {-# LANGUAGE DeriveGeneric #-}
@@ -17,6 +17,7 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE BlockArguments #-}
 module Cryptol.Parser.NoPat (RemovePatterns(..),Error(..)) where
 
 import Cryptol.Parser.AST
@@ -24,17 +25,16 @@ import Cryptol.Parser.Position(Range(..),emptyRange,start,at)
 import Cryptol.Parser.Names (namesP)
 import Cryptol.Utils.PP
 import Cryptol.Utils.Panic(panic)
+import Cryptol.Utils.RecordMap
 
 import           MonadLib hiding (mapM)
 import           Data.Maybe(maybeToList)
 import qualified Data.Map as Map
+import           Data.Text (Text)
 
 import GHC.Generics (Generic)
 import Control.DeepSeq
 
-import Prelude ()
-import Prelude.Compat
-
 class RemovePatterns t where
   -- | Eliminate all patterns in a program.
   removePatterns :: t -> (t, [Error])
@@ -45,18 +45,23 @@ instance RemovePatterns (Program PName) where
 instance RemovePatterns (Expr PName) where
   removePatterns e = runNoPatM (noPatE e)
 
-instance RemovePatterns (Module PName) where
+instance RemovePatterns (ModuleG mname PName) where
   removePatterns m = runNoPatM (noPatModule m)
 
 instance RemovePatterns [Decl PName] where
   removePatterns ds = runNoPatM (noPatDs ds)
 
+instance RemovePatterns (NestedModule PName) where
+  removePatterns (NestedModule m) = (NestedModule m1,errs)
+    where (m1,errs) = removePatterns m
+
 simpleBind :: Located PName -> Expr PName -> Bind PName
 simpleBind x e = Bind { bName = x, bParams = []
                       , bDef = at e (Located emptyRange (DExpr e))
                       , bSignature = Nothing, bPragmas = []
                       , bMono = True, bInfix = False, bFixity = Nothing
                       , bDoc = Nothing
+                      , bExport = Public
                       }
 
 sel :: Pattern PName -> PName -> Selector -> Bind PName
@@ -101,12 +106,12 @@ noPat pat =
          return (pTy r x ty, zipWith getN as [0..] ++ concat dss)
 
     PRecord fs ->
-      do (as,dss) <- unzip `fmap` mapM (noPat . value) fs
+      do let (shape, els) = unzip (canonicalFields fs)
+         (as,dss) <- unzip `fmap` mapM (noPat . snd) els
          x <- newName
          r <- getRange
-         let shape    = map (thing . name) fs
-             ty       = TRecord (map (fmap (\_ -> TWild)) fs)
-             getN a n = sel a x (RecordSel n (Just shape))
+         let ty           = TRecord (fmap (\(rng,_) -> (rng,TWild)) fs)
+             getN a n     = sel a x (RecordSel n (Just shape))
          return (pTy r x ty, zipWith getN as shape ++ concat dss)
 
     PTyped p t ->
@@ -146,15 +151,16 @@ noPatE expr =
   case expr of
     EVar {}       -> return expr
     ELit {}       -> return expr
-    ENeg e        -> ENeg    <$> noPatE e
-    EComplement e -> EComplement <$> noPatE e
     EGenerate e   -> EGenerate <$> noPatE e
     ETuple es     -> ETuple  <$> mapM noPatE es
-    ERecord es    -> ERecord <$> mapM noPatF es
+    ERecord es    -> ERecord <$> traverse (traverse noPatE) es
     ESel e s      -> ESel    <$> noPatE e <*> return s
     EUpd mb fs    -> EUpd    <$> traverse noPatE mb <*> traverse noPatUF fs
     EList es      -> EList   <$> mapM noPatE es
     EFromTo {}    -> return expr
+    EFromToBy {}  -> return expr
+    EFromToDownBy {} -> return expr
+    EFromToLessThan{} -> return expr
     EInfFrom e e' -> EInfFrom <$> noPatE e <*> traverse noPatE e'
     EComp e mss   -> EComp  <$> noPatE e <*> mapM noPatArm mss
     EApp e1 e2    -> EApp   <$> noPatE e1 <*> noPatE e2
@@ -163,29 +169,38 @@ noPatE expr =
     EWhere e ds   -> EWhere <$> noPatE e <*> noPatDs ds
     ETyped e t    -> ETyped <$> noPatE e <*> return t
     ETypeVal {}   -> return expr
-    EFun ps e     -> do (ps1,e1) <- noPatFun ps e
-                        return (EFun ps1 e1)
+    EFun desc ps e -> noPatFun (funDescrName desc) (funDescrArgOffset desc) ps e
     ELocated e r1 -> ELocated <$> inRange r1 (noPatE e) <*> return r1
 
     ESplit e      -> ESplit  <$> noPatE e
     EParens e     -> EParens <$> noPatE e
     EInfix x y f z-> EInfix  <$> noPatE x <*> pure y <*> pure f <*> noPatE z
+    EPrefix op e  -> EPrefix op <$> noPatE e
 
-  where noPatF x = do e <- noPatE (value x)
-                      return x { value = e }
 
 noPatUF :: UpdField PName -> NoPatM (UpdField PName)
 noPatUF (UpdField h ls e) = UpdField h ls <$> noPatE e
 
-noPatFun :: [Pattern PName] -> Expr PName -> NoPatM ([Pattern PName], Expr PName)
-noPatFun ps e =
-  do (xs,bs) <- unzip <$> mapM noPat ps
-     e1 <- noPatE e
-     let body = case concat bs of
-                        [] -> e1
-                        ds -> EWhere e1 $ map DBind ds
-     return (xs, body)
-
+-- Desugar lambdas with multiple patterns into a sequence of
+-- lambdas with a single, simple pattern each.  Bindings required
+-- to simplify patterns are placed inside "where" blocks that are
+-- interspersed into the lambdas to ensure that the lexical
+-- structure is reliable, with names on the right shadowing names
+-- on the left.
+noPatFun :: Maybe PName -> Int -> [Pattern PName] -> Expr PName -> NoPatM (Expr PName)
+noPatFun _   _      []     e = noPatE e
+noPatFun mnm offset (p:ps) e =
+  do (p',ds) <- noPat p
+     e' <- noPatFun mnm (offset+1) ps e
+     let body = case ds of
+                  [] -> e'
+                  _  -> EWhere e' $ map DBind (reverse ds)
+                           --                  ^
+                           -- This reverse isn't strictly necessary, but yields more sensible
+                           -- variable ordering results from type inference.  I'm not entirely
+                           -- sure why.
+     let desc = FunDesc mnm offset
+     return (EFun desc [p'] body)
 
 noPatArm :: [Match PName] -> NoPatM [Match PName]
 noPatArm ms = concat <$> mapM noPatM ms
@@ -205,9 +220,28 @@ noMatchB b =
           | otherwise        -> panic "NoPat" [ "noMatchB: primitive with params"
                                               , show b ]
 
+    DForeign
+      | null (bParams b) -> return b
+      | otherwise        -> panic "NoPat" [ "noMatchB: foreign with params"
+                                          , show b ]
+
     DExpr e ->
-      do (ps,e') <- noPatFun (bParams b) e
-         return b { bParams = ps, bDef = DExpr e' <$ bDef b }
+      do e' <- noPatFun (Just (thing (bName b))) 0 (bParams b) e
+         return b { bParams = [], bDef = DExpr e' <$ bDef b }
+
+    DPropGuards guards ->
+      do let nm = thing (bName b)
+             ps = bParams b
+         gs <- mapM (noPatPropGuardCase nm ps) guards
+         pure b { bParams = [], bDef = DPropGuards gs <$ bDef b }
+
+noPatPropGuardCase ::
+  PName ->
+  [Pattern PName] ->
+  PropGuardCase PName -> NoPatM (PropGuardCase PName)
+noPatPropGuardCase f xs pc =
+  do e <- noPatFun (Just f) 0 xs (pgcExpr pc)
+     pure pc { pgcExpr = e }
 
 noMatchD :: Decl PName -> NoPatM [Decl PName]
 noMatchD decl =
@@ -218,6 +252,7 @@ noMatchD decl =
 
     DBind b         -> do b1 <- noMatchB b
                           return [DBind b1]
+    DRec {}         -> panic "noMatchD" [ "DRec" ]
 
     DPatBind p e    -> do (p',bs) <- noPat p
                           let (x,ts) = splitSimpleP p'
@@ -232,6 +267,7 @@ noMatchD decl =
                                               , bInfix = False
                                               , bFixity = Nothing
                                               , bDoc = Nothing
+                                              , bExport = Public
                                               } : map DBind bs
     DType {}        -> return [decl]
     DProp {}        -> return [decl]
@@ -315,10 +351,14 @@ noPatTopDs tds =
 noPatProg :: Program PName -> NoPatM (Program PName)
 noPatProg (Program topDs) = Program <$> noPatTopDs topDs
 
-noPatModule :: Module PName -> NoPatM (Module PName)
+noPatModule :: ModuleG mname PName -> NoPatM (ModuleG mname PName)
 noPatModule m =
-  do ds1 <- noPatTopDs (mDecls m)
-     return m { mDecls = ds1 }
+  do def <-
+       case mDef m of
+         NormalModule ds -> NormalModule <$> noPatTopDs ds
+         FunctorInstance f as i -> pure (FunctorInstance f as i)
+         InterfaceModule s -> pure (InterfaceModule s)
+     pure m { mDef = def }
 
 --------------------------------------------------------------------------------
 
@@ -327,7 +367,7 @@ data AnnotMap = AnnotMap
   , annSigs     :: Map.Map PName [Located (Schema PName)]
   , annValueFs  :: Map.Map PName [Located  Fixity       ]
   , annTypeFs   :: Map.Map PName [Located  Fixity       ]
-  , annDocs     :: Map.Map PName [Located  String       ]
+  , annDocs     :: Map.Map PName [Located  Text         ]
   }
 
 type Annotates a = a -> StateT AnnotMap NoPatM a
@@ -355,28 +395,22 @@ annotTopDs tds =
              let d1 = DPrimType tl { tlValue = pt }
              (d1 :) <$> annotTopDs ds
 
-        DParameterType p ->
-          do p1 <- annotParameterType p
-             (DParameterType p1 :) <$> annotTopDs ds
-
-        DParameterConstraint {} -> (d :) <$> annotTopDs ds
-
-        DParameterFun p ->
-          do AnnotMap { .. } <- get
-             let rm _ _ = Nothing
-                 name = thing (pfName p)
-             case Map.updateLookupWithKey rm name annValueFs of
-               (Nothing,_)  -> (d :) <$> annotTopDs ds
-               (Just f,fs1) ->
-                 do mbF <- lift (checkFixs name f)
-                    set AnnotMap { annValueFs = fs1, .. }
-                    let p1 = p { pfFixity = mbF }
-                    (DParameterFun p1 :) <$> annotTopDs ds
+        DParamDecl {} -> (d :) <$> annotTopDs ds
+        DInterfaceConstraint {} -> (d :) <$> annotTopDs ds
 
         -- XXX: we may want to add pragmas to newtypes?
         TDNewtype {} -> (d :) <$> annotTopDs ds
         Include {}   -> (d :) <$> annotTopDs ds
 
+        DModule m ->
+          case removePatterns (tlValue m) of
+            (m1,errs) -> do lift (mapM_ recordError errs)
+                            (DModule m { tlValue = m1 } :) <$> annotTopDs ds
+
+        DImport {} -> (d :) <$> annotTopDs ds
+
+        DModParam {} -> (d :) <$> annotTopDs ds
+
     [] -> return []
 
 
@@ -395,6 +429,7 @@ annotD :: Decl PName -> ExceptionT () (StateT AnnotMap NoPatM) (Decl PName)
 annotD decl =
   case decl of
     DBind b       -> DBind <$> lift (annotB b)
+    DRec {}       -> panic "annotD" [ "DRec" ]
     DSignature {} -> raise ()
     DFixity{}     -> raise ()
     DPragma {}    -> raise ()
@@ -458,15 +493,6 @@ annotPrimType pt =
   do f <- annotTyThing (thing (primTName pt))
      pure pt { primTFixity = f }
 
--- | Annotate a module's type parameter.
-annotParameterType :: Annotates (ParameterType PName)
-annotParameterType pt =
-  do f <- annotTyThing (thing (ptName pt))
-     pure pt { ptFixity = f }
-
-
-
-
 -- | Check for multiple signatures.
 checkSigs :: PName -> [Located (Schema PName)] -> NoPatM (Maybe (Schema PName))
 checkSigs _ []             = return Nothing
@@ -481,7 +507,7 @@ checkFixs f fs@(x:_) = do recordError $ MultipleFixities f $ map srcRange fs
                           return (Just (thing x))
 
 
-checkDocs :: PName -> [Located String] -> NoPatM (Maybe String)
+checkDocs :: PName -> [Located Text] -> NoPatM (Maybe Text)
 checkDocs _ []       = return Nothing
 checkDocs _ [d]      = return (Just (thing d))
 checkDocs f ds@(d:_) = do recordError $ MultipleDocs f (map srcRange ds)
@@ -506,7 +532,7 @@ toFixity (DFixity f ns) = [ (thing n, [Located (srcRange n) f]) | n <- ns ]
 toFixity _              = []
 
 -- | Does this top-level declaration provide a documentation string?
-toDocs :: TopLevel (Decl PName) -> [(PName, [Located String])]
+toDocs :: TopLevel (Decl PName) -> [(PName, [Located Text])]
 toDocs TopLevel { .. }
   | Just txt <- tlDoc = go txt tlValue
   | otherwise = []
@@ -516,6 +542,7 @@ toDocs TopLevel { .. }
       DSignature ns _ -> [ (thing n, [txt]) | n <- ns ]
       DFixity _ ns    -> [ (thing n, [txt]) | n <- ns ]
       DBind b         -> [ (thing (bName b), [txt]) ]
+      DRec {}         -> panic "toDocs" [ "DRec" ]
       DLocated d _    -> go txt d
       DPatBind p _    -> [ (thing n, [txt]) | n <- namesP p ]
 
@@ -539,13 +566,13 @@ data Error  = MultipleSignatures PName [Located (Schema PName)]
               deriving (Show,Generic, NFData)
 
 instance Functor NoPatM where fmap = liftM
-instance Applicative NoPatM where pure = return; (<*>) = ap
+instance Applicative NoPatM where
+  pure x = M (pure x)
+  (<*>) = ap
 instance Monad NoPatM where
-  return x  = M (return x)
-  fail x    = M (fail x)
+  return    = pure
   M x >>= k = M (x >>= unM . k)
 
-
 -- | Pick a new name, to be used when desugaring patterns.
 newName :: NoPatM PName
 newName = M $ sets $ \s -> let x = names s
