@@ -20,7 +20,7 @@ import           System.FilePath(takeExtension)
 
 import           Cryptol.Utils.Panic
 
-data PreProc = None | Markdown | LaTeX
+data PreProc = None | Markdown | LaTeX | RST
 
 knownExts :: [String]
 knownExts  =
@@ -28,6 +28,7 @@ knownExts  =
   , "tex"
   , "markdown"
   , "md"
+  , "rst"
   ]
 
 guessPreProc :: FilePath -> PreProc
@@ -35,6 +36,7 @@ guessPreProc file = case takeExtension file of
                       ".tex"      -> LaTeX
                       ".markdown" -> Markdown
                       ".md"       -> Markdown
+                      ".rst"      -> RST
                       _           -> None
 
 unLit :: PreProc -> Text -> Text
@@ -47,6 +49,7 @@ preProc p =
     None     -> return . Code
     Markdown -> markdown
     LaTeX    -> latex
+    RST      -> rst
 
 
 data Block = Code [Text] | Comment [Text]
@@ -78,13 +81,14 @@ markdown = blanks []
   where
   comment current []    = mk Comment current
   comment current (l : ls)
+    | Just op <- isOpenFence l = mk Comment (l : current) ++ fenced op [] ls
     | isBlank l         = blanks (l : current) ls
     | otherwise         = comment (l : current) ls
 
   blanks current []     = mk Comment current
   blanks current (l : ls)
-    | isCodeLine l             = mk Comment current ++ code [l] ls
     | Just op <- isOpenFence l = mk Comment (l : current) ++ fenced op [] ls
+    | isCodeLine l             = mk Comment current ++ code [l] ls
     | isBlank l                = blanks  (l : current) ls
     | otherwise                = comment (l : current) ls
 
@@ -101,7 +105,7 @@ markdown = blanks []
 
   isOpenFence l
     | "```" `Text.isPrefixOf` l' =
-      Just $ case Text.drop 3 l' of
+      Just $ case Text.dropWhile isSpace (Text.drop 3 l') of
                l'' | "cryptol" `Text.isPrefixOf` l'' -> Code
                    | isBlank l''                     -> Code
                    | otherwise                       -> Comment
@@ -110,7 +114,7 @@ markdown = blanks []
     where
     l' = Text.dropWhile isSpace l
 
-  isCloseFence l = "```" `Text.isPrefixOf` l
+  isCloseFence l = "```" `Text.isPrefixOf` Text.dropWhile isSpace l
   isBlank l      = Text.all isSpace l
   isCodeLine l   = "\t" `Text.isPrefixOf` l || "    " `Text.isPrefixOf` l
 
@@ -133,6 +137,35 @@ latex = comment []
   isBeginCode l = "\\begin{code}" `Text.isPrefixOf` l
   isEndCode l   = "\\end{code}"   `Text.isPrefixOf` l
 
+rst :: [Text] -> [Block]
+rst = comment []
+  where
+  isBeginCode l = case filter (not . Text.null) (Text.split isSpace l) of
+                    ["..", dir, "cryptol"] -> dir == "code-block::" ||
+                                              dir == "sourcecode::"
+                    _ -> False
 
+  isEmpty       = Text.all isSpace
+  isCode l      = case Text.uncons l of
+                    Just (c, _) -> isSpace c
+                    Nothing     -> True
+
+  comment acc ls =
+    case ls of
+      [] -> mk Comment acc
+      l : ls1 | isBeginCode l -> codeOptions (l : acc) ls1
+              | otherwise     -> comment (l : acc) ls1
+
+  codeOptions acc ls =
+    case ls of
+      [] -> mk Comment acc
+      l : ls1 | isEmpty l -> mk Comment (l : acc) ++ code [] ls1
+              | otherwise -> codeOptions (l : acc) ls1
+
+  code acc ls =
+    case ls of
+      [] -> mk Code acc
+      l : ls1 | isCode l   -> code (l : acc) ls1
+              | otherwise  -> mk Code acc ++ comment [] ls
 
 
