@@ -6,6 +6,7 @@
 -- Stability   :  provisional
 -- Portability :  portable
 
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
@@ -14,6 +15,7 @@
 {-# LANGUAGE TypeFamilies #-}
 {-# LANGUAGE ViewPatterns #-}
 {-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE OverloadedStrings #-}
 
 module Cryptol.REPL.Monad (
     -- * REPL Monad
@@ -35,9 +37,13 @@ module Cryptol.REPL.Monad (
   , getFocusedEnv
   , getModuleEnv, setModuleEnv
   , getDynEnv, setDynEnv
+  , getCallStacks
+  , getTCSolver
+  , resetTCSolver
   , uniqify, freshName
-  , getTSyns, getNewtypes, getVars
   , whenDebug
+  , getEvalOptsAction
+  , getPPValOpts
   , getExprNames
   , getTypeNames
   , getPropertyNames
@@ -49,21 +55,26 @@ module Cryptol.REPL.Monad (
   , shouldContinue
   , unlessBatch
   , asBatch
-  , disableLet
-  , enableLet
-  , getLetEnabled
   , validEvalContext
   , updateREPLTitle
   , setUpdateREPLTitle
+  , withRandomGen
+  , setRandomGen
+  , getRandomGen
 
     -- ** Config Options
   , EnvVal(..)
   , OptionDescr(..)
   , setUser, getUser, getKnownUser, tryGetUser
   , userOptions
+  , userOptionsWithAliases
   , getUserSatNum
   , getUserShowProverStats
   , getUserProverValidate
+  , parsePPFloatFormat
+  , parseFieldOrder
+  , getProverConfig
+  , parseSearchPath
 
     -- ** Configurable Output
   , getPutStr
@@ -78,7 +89,7 @@ module Cryptol.REPL.Monad (
 
 import Cryptol.REPL.Trie
 
-import Cryptol.Eval (EvalError)
+import Cryptol.Eval (EvalErrorEx, Unsupported, WordTooWide,EvalOpts(..))
 import qualified Cryptol.ModuleSystem as M
 import qualified Cryptol.ModuleSystem.Env as M
 import qualified Cryptol.ModuleSystem.Name as M
@@ -89,32 +100,42 @@ import Cryptol.Parser.NoPat (Error)
 import Cryptol.Parser.Position (emptyRange, Range(from))
 import qualified Cryptol.TypeCheck.AST as T
 import qualified Cryptol.TypeCheck as T
+import qualified Cryptol.TypeCheck.Solver.SMT as SMT
 import qualified Cryptol.IR.FreeVars as T
 import qualified Cryptol.Utils.Ident as I
 import Cryptol.Utils.PP
 import Cryptol.Utils.Panic (panic)
 import Cryptol.Utils.Logger(Logger, logPutStr, funLogger)
 import qualified Cryptol.Parser.AST as P
-import Cryptol.Symbolic (proverNames, lookupProver, SatNum(..))
+import Cryptol.Symbolic (SatNum(..))
+import Cryptol.Symbolic.SBV (SBVPortfolioException)
+import Cryptol.Symbolic.What4 (W4Exception)
+import qualified Cryptol.Symbolic.SBV as SBV (proverNames, setupProver, defaultProver, SBVProverConfig)
+import qualified Cryptol.Symbolic.What4 as W4 (proverNames, setupProver, W4ProverConfig)
+
+
 
 import Control.Monad (ap,unless,when)
 import Control.Monad.Base
+import qualified Control.Monad.Catch as Ex
 import Control.Monad.IO.Class
 import Control.Monad.Trans.Control
-import Data.Char (isSpace)
+import Data.Char (isSpace, toLower)
 import Data.IORef
-    (IORef,newIORef,readIORef,modifyIORef,atomicModifyIORef)
+    (IORef,newIORef,readIORef,atomicModifyIORef)
 import Data.List (intercalate, isPrefixOf, unfoldr, sortBy)
 import Data.Maybe (catMaybes)
 import Data.Ord (comparing)
 import Data.Typeable (Typeable)
 import System.Directory (findExecutable)
+import System.FilePath (splitSearchPath, searchPathSeparator)
 import qualified Control.Exception as X
 import qualified Data.Map as Map
 import qualified Data.Set as Set
 import Text.Read (readMaybe)
 
-import Data.SBV.Dynamic (sbvCheckSolverInstallation)
+import Data.SBV (SBVException)
+import qualified System.Random.TF as TF
 
 import Prelude ()
 import Prelude.Compat
@@ -151,18 +172,38 @@ data RW = RW
   , eLogger      :: Logger
     -- ^ Use this to send messages to the user
 
-  , eLetEnabled  :: Bool
-    -- ^ Should we allow `let` on the command line
+  , eCallStacks :: Bool
 
   , eUpdateTitle :: REPL ()
     -- ^ Execute this every time we load a module.
     -- This is used to change the title of terminal when loading a module.
+
+  , eProverConfig :: Either SBV.SBVProverConfig W4.W4ProverConfig
+
+  , eTCConfig :: T.SolverConfig
+    -- ^ Solver configuration to be used for typechecking
+
+  , eTCSolver :: Maybe SMT.Solver
+    -- ^ Solver instance to be used for typechecking
+
+  , eTCSolverRestarts :: !Int
+    -- ^ Counts how many times we've restarted the solver.
+    -- Used as a kind of id for the current solver, which helps avoid
+    -- a race condition where the callback of a dead solver runs after
+    -- a new one has been started.
+
+  , eRandomGen :: TF.TFGen
+    -- ^ Random number generator for things like QC and dumpTests
   }
 
+
 -- | Initial, empty environment.
-defaultRW :: Bool -> Logger -> IO RW
-defaultRW isBatch l = do
+defaultRW :: Bool -> Bool -> Logger -> IO RW
+defaultRW isBatch callStacks l = do
   env <- M.initialModuleEnv
+  rng <- TF.newTFGen
+  let searchPath = M.meSearchPath env
+  let solverConfig = T.defaultSolverConfig searchPath
   return RW
     { eLoadedMod   = Nothing
     , eEditFile    = Nothing
@@ -171,8 +212,13 @@ defaultRW isBatch l = do
     , eModuleEnv   = env
     , eUserEnv     = mkUserEnv userOptions
     , eLogger      = l
-    , eLetEnabled  = True
+    , eCallStacks  = callStacks
     , eUpdateTitle = return ()
+    , eProverConfig = Left SBV.defaultProver
+    , eTCConfig    = solverConfig
+    , eTCSolver    = Nothing
+    , eTCSolverRestarts = 0
+    , eRandomGen = rng
     }
 
 -- | Build up the prompt for the REPL.
@@ -182,16 +228,17 @@ mkPrompt rw
   | detailedPrompt = withEdit ++ "> "
   | otherwise      = modLn ++ "> "
   where
-  detailedPrompt = False
+  detailedPrompt = id False
 
   modLn   =
     case lName =<< eLoadedMod rw of
       Nothing -> show (pp I.preludeName)
       Just m
-        | M.isLoadedParamMod m (M.meLoadedModules (eModuleEnv rw)) ->
-                 modName ++ "(parameterized)"
+        | M.isLoadedParamMod m loaded -> modName ++ "(parameterized)"
+        | M.isLoadedInterface m loaded -> modName ++ "(interface)"
         | otherwise -> modName
         where modName = pretty m
+              loaded = M.meLoadedModules (eModuleEnv rw)
 
   withFocus =
     case eLoadedMod rw of
@@ -217,10 +264,12 @@ mkPrompt rw
 newtype REPL a = REPL { unREPL :: IORef RW -> IO a }
 
 -- | Run a REPL action with a fresh environment.
-runREPL :: Bool -> Logger -> REPL a -> IO a
-runREPL isBatch l m = do
-  ref <- newIORef =<< defaultRW isBatch l
-  unREPL m ref
+runREPL :: Bool -> Bool -> Logger -> REPL a -> IO a
+runREPL isBatch callStacks l m = do
+  Ex.bracket
+    (newIORef =<< defaultRW isBatch callStacks l)
+    (unREPL resetTCSolver)
+    (unREPL m)
 
 instance Functor REPL where
   {-# INLINE fmap #-}
@@ -228,13 +277,13 @@ instance Functor REPL where
 
 instance Applicative REPL where
   {-# INLINE pure #-}
-  pure = return
+  pure x = REPL (\_ -> pure x)
   {-# INLINE (<*>) #-}
   (<*>) = ap
 
 instance Monad REPL where
   {-# INLINE return #-}
-  return x = REPL (\_ -> return x)
+  return = pure
 
   {-# INLINE (>>=) #-}
   m >>= f = REPL $ \ref -> do
@@ -258,6 +307,25 @@ instance M.FreshM REPL where
     let (a,s') = f (M.meSupply eModuleEnv)
      in (RW { eModuleEnv = eModuleEnv { M.meSupply = s' }, .. },a)
 
+instance Ex.MonadThrow REPL where
+  throwM e = liftIO $ X.throwIO e
+
+instance Ex.MonadCatch REPL where
+  catch op handler = control $ \runInBase -> Ex.catch (runInBase op) (runInBase . handler)
+
+instance Ex.MonadMask REPL where
+  mask op = REPL $ \ref -> Ex.mask $ \u -> unREPL (op (q u)) ref
+    where q u (REPL b) = REPL (\ref -> u (b ref))
+
+  uninterruptibleMask op = REPL $ \ref ->
+    Ex.uninterruptibleMask $ \u -> unREPL (op (q u)) ref
+    where q u (REPL b) = REPL (\ref -> u (b ref))
+
+  generalBracket acq rls op = control $ \runInBase ->
+    Ex.generalBracket (runInBase acq)
+    (\saved -> \e -> runInBase (restoreM saved >>= \a -> rls a e))
+    (\saved -> runInBase (restoreM saved >>= op))
+
 -- Exceptions ------------------------------------------------------------------
 
 -- | REPL exceptions.
@@ -267,12 +335,18 @@ data REPLException
   | DirectoryNotFound FilePath
   | NoPatError [Error]
   | NoIncludeError [IncludeError]
-  | EvalError EvalError
+  | EvalError EvalErrorEx
+  | TooWide WordTooWide
+  | Unsupported Unsupported
   | ModuleSystemError NameDisp M.ModuleError
   | EvalPolyError T.Schema
+  | InstantiationsNotFound T.Schema
   | TypeNotTestable T.Type
-  | EvalInParamModule [M.Name]
+  | EvalInParamModule [T.TParam] [M.Name]
   | SBVError String
+  | SBVException SBVException
+  | SBVPortfolioException SBVPortfolioException
+  | W4Exception W4Exception
     deriving (Show,Typeable)
 
 instance X.Exception REPLException
@@ -292,14 +366,22 @@ instance PP REPLException where
     NoIncludeError es    -> vcat (map ppIncludeError es)
     ModuleSystemError ns me -> fixNameDisp ns (pp me)
     EvalError e          -> pp e
+    Unsupported e        -> pp e
+    TooWide e            -> pp e
     EvalPolyError s      -> text "Cannot evaluate polymorphic value."
                          $$ text "Type:" <+> pp s
+    InstantiationsNotFound s -> text "Cannot find instantiations for some type variables."
+                             $$ text "Type:" <+> pp s
     TypeNotTestable t    -> text "The expression is not of a testable type."
                          $$ text "Type:" <+> pp t
-    EvalInParamModule xs ->
-      text "Expression depends on definitions from a parameterized module:"
-        $$ nest 2 (vcat (map pp xs))
+    EvalInParamModule as xs -> nest 2 $ vsep $
+      [ text "Expression depends on definitions from a parameterized module:" ]
+      ++ map pp as
+      ++ map pp xs
     SBVError s           -> text "SBV error:" $$ text s
+    SBVException e       -> text "SBV exception:" $$ text (show e)
+    SBVPortfolioException e -> text "SBV exception:" $$ text (show e)
+    W4Exception e        -> text "What4 exception:" $$ text (show e)
 
 -- | Raise an exception.
 raise :: REPLException -> REPL a
@@ -314,16 +396,23 @@ finally m1 m2 = REPL (\ref -> unREPL m1 ref `X.finally` unREPL m2 ref)
 
 
 rethrowEvalError :: IO a -> IO a
-rethrowEvalError m = run `X.catch` rethrow
+rethrowEvalError m =
+    run `X.catch` rethrow
+        `X.catch` rethrowTooWide
+        `X.catch` rethrowUnsupported
   where
   run = do
     a <- m
     return $! a
 
-  rethrow :: EvalError -> IO a
+  rethrow :: EvalErrorEx -> IO a
   rethrow exn = X.throwIO (EvalError exn)
 
+  rethrowTooWide :: WordTooWide -> IO a
+  rethrowTooWide exn = X.throwIO (TooWide exn)
 
+  rethrowUnsupported :: Unsupported -> IO a
+  rethrowUnsupported exn = X.throwIO (Unsupported exn)
 
 
 -- Primitives ------------------------------------------------------------------
@@ -339,12 +428,73 @@ modifyRW :: (RW -> (RW,a)) -> REPL a
 modifyRW f = REPL (\ ref -> atomicModifyIORef ref f)
 
 modifyRW_ :: (RW -> RW) -> REPL ()
-modifyRW_ f = REPL (\ ref -> modifyIORef ref f)
+modifyRW_ f = REPL (\ ref -> atomicModifyIORef ref (\x -> (f x, ())))
 
 -- | Construct the prompt for the current environment.
 getPrompt :: REPL String
 getPrompt  = mkPrompt `fmap` getRW
 
+getCallStacks :: REPL Bool
+getCallStacks = eCallStacks <$> getRW
+
+-- This assumes that we are not starting solvers in parallel, otherwise
+-- there are a bunch of race conditions here.
+getTCSolver :: REPL SMT.Solver
+getTCSolver =
+  do rw <- getRW
+     case eTCSolver rw of
+       Just s -> return s
+       Nothing ->
+         do ref <- REPL (\ref -> pure ref)
+            let now = eTCSolverRestarts rw + 1
+                upd new = if eTCSolverRestarts new == now
+                             then new { eTCSolver = Nothing }
+                             else new
+                onExit = atomicModifyIORef ref (\s -> (upd s, ()))
+            s <- io (SMT.startSolver onExit (eTCConfig rw))
+            modifyRW_ (\rw' -> rw'{ eTCSolver = Just s
+                                  , eTCSolverRestarts = now })
+            return s
+
+
+resetTCSolver :: REPL ()
+resetTCSolver =
+  do mtc <- eTCSolver <$> getRW
+     case mtc of
+       Nothing -> return ()
+       Just s  ->
+         do io (SMT.stopSolver s)
+            modifyRW_ (\rw -> rw{ eTCSolver = Nothing })
+
+-- Get the setting we should use for displaying values.
+getPPValOpts :: REPL PPOpts
+getPPValOpts =
+  do base       <- getKnownUser "base"
+     ascii      <- getKnownUser "ascii"
+     infLength  <- getKnownUser "infLength"
+
+     fpBase     <- getKnownUser "fpBase"
+     fpFmtTxt   <- getKnownUser "fpFormat"
+     fieldOrder <- getKnownUser "fieldOrder"
+     let fpFmt = case parsePPFloatFormat fpFmtTxt of
+                   Just f  -> f
+                   Nothing -> panic "getPPOpts"
+                                      [ "Failed to parse fp-format" ]
+
+     return PPOpts { useBase       = base
+                   , useAscii      = ascii
+                   , useInfLength  = infLength
+                   , useFPBase     = fpBase
+                   , useFPFormat   = fpFmt
+                   , useFieldOrder = fieldOrder
+                   }
+
+getEvalOptsAction :: REPL (IO EvalOpts)
+getEvalOptsAction = REPL $ \rwRef -> pure $
+  do ppOpts <- unREPL getPPValOpts rwRef
+     l      <- unREPL getLogger rwRef
+     return EvalOpts { evalPPOpts = ppOpts, evalLogger = l }
+
 clearLoadedMod :: REPL ()
 clearLoadedMod = do modifyRW_ (\rw -> rw { eLoadedMod = upd <$> eLoadedMod rw })
                     updateREPLTitle
@@ -376,11 +526,17 @@ setSearchPath :: [FilePath] -> REPL ()
 setSearchPath path = do
   me <- getModuleEnv
   setModuleEnv $ me { M.meSearchPath = path }
+  setUserDirect "path" (EnvString (renderSearchPath path))
 
 prependSearchPath :: [FilePath] -> REPL ()
 prependSearchPath path = do
   me <- getModuleEnv
-  setModuleEnv $ me { M.meSearchPath = path ++ M.meSearchPath me }
+  let path' = path ++ M.meSearchPath me
+  setModuleEnv $ me { M.meSearchPath = path' }
+  setUserDirect "path" (EnvString (renderSearchPath path'))
+
+getProverConfig :: REPL (Either SBV.SBVProverConfig W4.W4ProverConfig)
+getProverConfig = eProverConfig <$> getRW
 
 shouldContinue :: REPL Bool
 shouldContinue  = eContinue `fmap` getRW
@@ -403,16 +559,6 @@ asBatch body = do
   modifyRW_ $ (\ rw -> rw { eIsBatch = wasBatch })
   return a
 
-disableLet :: REPL ()
-disableLet  = modifyRW_ (\ rw -> rw { eLetEnabled = False })
-
-enableLet :: REPL ()
-enableLet  = modifyRW_ (\ rw -> rw { eLetEnabled = True })
-
--- | Are let-bindings enabled in this REPL?
-getLetEnabled :: REPL Bool
-getLetEnabled = fmap eLetEnabled getRW
-
 -- | Is evaluation enabled.  If the currently focused module is
 -- parameterized, then we cannot evalute.
 validEvalContext :: T.FreeVars a => a -> REPL ()
@@ -421,17 +567,20 @@ validEvalContext a =
      let ds      = T.freeVars a
          badVals = foldr badName Set.empty (T.valDeps ds)
          bad     = foldr badName badVals (T.tyDeps ds)
+         badTs   = T.tyParams ds
 
          badName nm bs =
            case M.nameInfo nm of
-             M.Declared m _
+
+             -- XXX: Changes if focusing on nested modules
+             M.GlobalName _ I.OrigName { ogModule = I.TopModule m }
                | M.isLoadedParamMod m (M.meLoadedModules me) -> Set.insert nm bs
+               | M.isLoadedInterface m (M.meLoadedModules me) -> Set.insert nm bs
+
              _ -> bs
 
-     unless (Set.null bad) $
-       raise (EvalInParamModule (Set.toList bad))
-
-
+     unless (Set.null bad && Set.null badTs) $
+       raise (EvalInParamModule (Set.toList badTs) (Set.toList bad))
 
 -- | Update the title
 updateREPLTitle :: REPL ()
@@ -471,75 +620,33 @@ rPutStrLn str = rPutStr $ str ++ "\n"
 rPrint :: Show a => a -> REPL ()
 rPrint x = rPutStrLn (show x)
 
-getFocusedEnv :: REPL (M.IfaceParams,M.IfaceDecls,M.NamingEnv,NameDisp)
-getFocusedEnv  = do
-  me <- getModuleEnv
-  -- dyNames is a NameEnv that removes the #Uniq prefix from interactively-bound
-  -- variables.
-  let (dyDecls,dyNames,dyDisp) = M.dynamicEnv me
-  let (fParams,fDecls,fNames,fDisp) = M.focusedEnv me
-  return ( fParams
-         , dyDecls `mappend` fDecls
-         , dyNames `M.shadowing` fNames
-         , dyDisp `mappend` fDisp)
-
-  -- -- the subtle part here is removing the #Uniq prefix from
-  -- -- interactively-bound variables, and also excluding any that are
-  -- -- shadowed and thus can no longer be referenced
-  -- let (fDecls,fNames,fDisp) = M.focusedEnv me
-  --     edecls = M.ifDecls dyDecls
-  --     -- is this QName something the user might actually type?
-  --     isShadowed (qn@(P.QName (Just (P.unModName -> ['#':_])) name), _) =
-  --         case Map.lookup localName neExprs of
-  --           Nothing -> False
-  --           Just uniqueNames -> isNamed uniqueNames
-  --       where localName = P.QName Nothing name
-  --             isNamed us = any (== qn) (map M.qname us)
-  --             neExprs = M.neExprs (M.deNames (M.meDynEnv me))
-  --     isShadowed _ = False
-  --     unqual ((P.QName _ name), ifds) = (P.QName Nothing name, ifds)
-  --     edecls' = Map.fromList
-  --             . map unqual
-  --             . filter isShadowed
-  --             $ Map.toList edecls
-  -- return (decls `mappend` mempty { M.ifDecls = edecls' }, names `mappend` dyNames)
-
-getVars :: REPL (Map.Map M.Name M.IfaceDecl)
-getVars  = do
-  (_,decls,_,_) <- getFocusedEnv
-  return (M.ifDecls decls)
-
-getTSyns :: REPL (Map.Map M.Name T.TySyn)
-getTSyns  = do
-  (_,decls,_,_) <- getFocusedEnv
-  return (M.ifTySyns decls)
-
-getNewtypes :: REPL (Map.Map M.Name T.Newtype)
-getNewtypes = do
-  (_,decls,_,_) <- getFocusedEnv
-  return (M.ifNewtypes decls)
+getFocusedEnv :: REPL M.ModContext
+getFocusedEnv  = M.focusedEnv <$> getModuleEnv
 
 -- | Get visible variable names.
+-- This is used for command line completition.
 getExprNames :: REPL [String]
 getExprNames =
-  do (_,_, fNames, _) <- getFocusedEnv
-     return (map (show . pp) (Map.keys (M.neExprs fNames)))
+  do fNames <- M.mctxNames <$> getFocusedEnv
+     return (map (show . pp) (Map.keys (M.namespaceMap M.NSValue fNames)))
 
 -- | Get visible type signature names.
+-- This is used for command line completition.
 getTypeNames :: REPL [String]
 getTypeNames  =
-  do (_,_, fNames, _) <- getFocusedEnv
-     return (map (show . pp) (Map.keys (M.neTypes fNames)))
+  do fNames <- M.mctxNames <$> getFocusedEnv
+     return (map (show . pp) (Map.keys (M.namespaceMap M.NSType fNames)))
 
 -- | Return a list of property names, sorted by position in the file.
-getPropertyNames :: REPL ([M.Name],NameDisp)
+getPropertyNames :: REPL ([(M.Name,M.IfaceDecl)],NameDisp)
 getPropertyNames =
-  do (_,decls,_,names) <- getFocusedEnv
-     let xs = M.ifDecls decls
-         ps = sortBy (comparing (from . M.nameLoc))
-            $ [ x | (x,d) <- Map.toList xs, T.PragmaProperty `elem` M.ifDeclPragmas d ]
+  do fe <- getFocusedEnv
+     let xs = M.ifDecls (M.mctxDecls fe)
+         ps = sortBy (comparing (from . M.nameLoc . fst))
+              [ (x,d) | (x,d) <- Map.toList xs,
+                    T.PragmaProperty `elem` M.ifDeclPragmas d ]
 
-     return (ps, names)
+     return (ps, M.mctxNameDisp fe)
 
 getModNames :: REPL [I.ModName]
 getModNames =
@@ -560,6 +667,18 @@ setDynEnv denv = do
   me <- getModuleEnv
   setModuleEnv (me { M.meDynEnv = denv })
 
+getRandomGen :: REPL TF.TFGen
+getRandomGen = eRandomGen <$> getRW
+
+setRandomGen :: TF.TFGen -> REPL ()
+setRandomGen rng = modifyRW_ (\s -> s { eRandomGen = rng })
+
+withRandomGen :: (TF.TFGen -> REPL (a, TF.TFGen)) -> REPL a
+withRandomGen repl =
+  do  g <- getRandomGen
+      (result, g') <- repl g
+      setRandomGen g'
+      pure result
 
 -- | Given an existing qualified name, prefix it with a
 -- relatively-unique string. We make it unique by prefixing with a
@@ -568,10 +687,12 @@ uniqify :: M.Name -> REPL M.Name
 
 uniqify name =
   case M.nameInfo name of
-    M.Declared ns s ->
-      M.liftSupply (M.mkDeclared ns s (M.nameIdent name) (M.nameFixity name) (M.nameLoc name))
+    M.GlobalName s og ->
+      M.liftSupply (M.mkDeclared (M.nameNamespace name)
+                  (I.ogModule og) s
+                  (M.nameIdent name) (M.nameFixity name) (M.nameLoc name))
 
-    M.Parameter ->
+    M.LocalName {} ->
       panic "[REPL] uniqify" ["tried to uniqify a parameter: " ++ pretty name]
 
 
@@ -589,9 +710,28 @@ uniqify name =
 -- the "<interactive>" namespace.
 freshName :: I.Ident -> M.NameSource -> REPL M.Name
 freshName i sys =
-  M.liftSupply (M.mkDeclared I.interactiveName sys i Nothing emptyRange)
+  M.liftSupply (M.mkDeclared I.NSValue mpath sys i Nothing emptyRange)
+  where mpath = M.TopModule I.interactiveName
 
 
+parseSearchPath :: String -> [String]
+parseSearchPath path = path'
+#if defined(mingw32_HOST_OS) || defined(__MINGW32__)
+      -- Windows paths search from end to beginning
+      where path' = reverse (splitSearchPath path)
+#else
+      where path' = splitSearchPath path
+#endif
+
+renderSearchPath :: [String] -> String
+renderSearchPath pathSegs = path
+#if defined(mingw32_HOST_OS) || defined(__MINGW32__)
+      -- Windows paths search from end to beginning
+      where path = intercalate [searchPathSeparator] (reverse pathSegs)
+#else
+      where path = intercalate [searchPathSeparator] pathSegs
+#endif
+
 -- User Environment Interaction ------------------------------------------------
 
 -- | User modifiable environment, for things like numeric base.
@@ -612,7 +752,7 @@ mkUserEnv opts = Map.fromList $ do
 
 -- | Set a user option.
 setUser :: String -> String -> REPL ()
-setUser name val = case lookupTrieExact name userOptions of
+setUser name val = case lookupTrieExact name userOptionsWithAliases of
 
   [opt] -> setUserOpt opt
   []    -> rPutStrLn ("Unknown env value `" ++ name ++ "`")
@@ -639,7 +779,7 @@ setUser name val = case lookupTrieExact name userOptions of
       | otherwise ->
         rPutStrLn ("Failed to parse boolean for field, `" ++ name ++ "`")
     where
-    doCheck v = do (r,ws) <- io (optCheck opt v)
+    doCheck v = do (r,ws) <- optCheck opt v
                    case r of
                      Just err -> rPutStrLn err
                      Nothing  -> do mapM_ rPutStrLn ws
@@ -684,6 +824,10 @@ getUser name = do
     Just ev -> return ev
     Nothing -> panic "[REPL] getUser" ["option `" ++ name ++ "` does not exist"]
 
+setUserDirect :: String -> EnvVal -> REPL ()
+setUserDirect optName val = do
+  modifyRW_ (\rw -> rw { eUserEnv = Map.insert optName val (eUserEnv rw) })
+
 getKnownUser :: IsEnvVal a => String -> REPL a
 getKnownUser x = fromEnvVal <$> getUser x
 
@@ -710,15 +854,21 @@ instance IsEnvVal String where
                    EnvString b -> b
                    _           -> badIsEnv "String"
 
+instance IsEnvVal FieldOrder where
+  fromEnvVal x = case x of
+                   EnvString s | Just o <- parseFieldOrder s
+                     -> o
+                   _ -> badIsEnv "display` or `canonical"
+
 badIsEnv :: String -> a
 badIsEnv x = panic "fromEnvVal" [ "[REPL] Expected a `" ++ x ++ "` value." ]
 
 
 getUserShowProverStats :: REPL Bool
-getUserShowProverStats = getKnownUser "prover-stats"
+getUserShowProverStats = getKnownUser "proverStats"
 
 getUserProverValidate :: REPL Bool
-getUserProverValidate = getKnownUser "prover-validate"
+getUserProverValidate = getKnownUser "proverValidate"
 
 -- Environment Options ---------------------------------------------------------
 
@@ -730,72 +880,97 @@ mkOptionMap  = foldl insert emptyTrie
   insert m d = insertTrie (optName d) d m
 
 -- | Returns maybe an error, and some warnings
-type Checker = EnvVal -> IO (Maybe String, [String])
+type Checker = EnvVal -> REPL (Maybe String, [String])
 
 noCheck :: Checker
 noCheck _ = return (Nothing, [])
 
-noWarns :: Maybe String -> IO (Maybe String, [String])
+noWarns :: Maybe String -> REPL (Maybe String, [String])
 noWarns mb = return (mb, [])
 
 data OptionDescr = OptionDescr
   { optName    :: String
+  , optAliases :: [String]
   , optDefault :: EnvVal
   , optCheck   :: Checker
   , optHelp    :: String
   , optEff     :: EnvVal -> REPL ()
   }
 
-simpleOpt :: String -> EnvVal -> Checker -> String -> OptionDescr
-simpleOpt optName optDefault optCheck optHelp =
+simpleOpt :: String -> [String] -> EnvVal -> Checker -> String -> OptionDescr
+simpleOpt optName optAliases optDefault optCheck optHelp =
   OptionDescr { optEff = \ _ -> return (), .. }
 
+userOptionsWithAliases :: OptionMap
+userOptionsWithAliases = foldl insert userOptions (leaves userOptions)
+  where
+  insert m d = foldl (\m' n -> insertTrie n d m') m (optAliases d)
+
 userOptions :: OptionMap
 userOptions  = mkOptionMap
-  [ simpleOpt "base" (EnvNum 16) checkBase
+  [ simpleOpt "base" [] (EnvNum 16) checkBase
     "The base to display words at (2, 8, 10, or 16)."
-  , simpleOpt "debug" (EnvBool False) noCheck
+  , simpleOpt "debug" [] (EnvBool False) noCheck
     "Enable debugging output."
-  , simpleOpt "ascii" (EnvBool False) noCheck
+  , simpleOpt "ascii" [] (EnvBool False) noCheck
     "Whether to display 7- or 8-bit words using ASCII notation."
-  , simpleOpt "infLength" (EnvNum 5) checkInfLength
+  , simpleOpt "infLength" ["inf-length"] (EnvNum 5) checkInfLength
     "The number of elements to display for infinite sequences."
-  , simpleOpt "tests" (EnvNum 100) noCheck
+  , simpleOpt "tests" [] (EnvNum 100) noCheck
     "The number of random tests to try with ':check'."
-  , simpleOpt "satNum" (EnvString "1") checkSatNum
+  , simpleOpt "satNum" ["sat-num"] (EnvString "1") checkSatNum
     "The maximum number of :sat solutions to display (\"all\" for no limit)."
-  , simpleOpt "prover" (EnvString "z3") checkProver $
+  , simpleOpt "prover" [] (EnvString "z3") checkProver $
     "The external SMT solver for ':prove' and ':sat'\n(" ++ proverListString ++ ")."
-  , simpleOpt "warnDefaulting" (EnvBool True) noCheck
+  , simpleOpt "warnDefaulting" ["warn-defaulting"] (EnvBool False) noCheck
     "Choose whether to display warnings when defaulting."
-  , simpleOpt "warnShadowing" (EnvBool True) noCheck
+  , simpleOpt "warnShadowing" ["warn-shadowing"] (EnvBool True) noCheck
     "Choose whether to display warnings when shadowing symbols."
-  , simpleOpt "smtfile" (EnvString "-") noCheck
+  , simpleOpt "warnPrefixAssoc" ["warn-prefix-assoc"] (EnvBool True) noCheck
+    "Choose whether to display warnings when expression association has changed due to new prefix operator fixities."
+  , simpleOpt "warnUninterp" ["warn-uninterp"] (EnvBool True) noCheck
+    "Choose whether to issue a warning when uninterpreted functions are used to implement primitives in the symbolic simulator."
+  , simpleOpt "warnNonExhaustiveConstraintGuards" ["warn-nonexhaustive-constraintguards"] (EnvBool True) noCheck
+    "Choose whether to issue a warning when a use of constraint guards is not determined to be exhaustive."
+  , simpleOpt "smtFile" ["smt-file"] (EnvString "-") noCheck
     "The file to use for SMT-Lib scripts (for debugging or offline proving).\nUse \"-\" for stdout."
-  , OptionDescr "mono-binds" (EnvBool True) noCheck
+  , OptionDescr "path" [] (EnvString "") noCheck
+    "The search path for finding named Cryptol modules." $
+    \case EnvString path ->
+            do let segs = parseSearchPath path
+               me <- getModuleEnv
+               setModuleEnv me { M.meSearchPath = segs }
+          _ -> return ()
+
+  , OptionDescr "monoBinds" ["mono-binds"] (EnvBool True) noCheck
     "Whether or not to generalize bindings in a 'where' clause." $
     \case EnvBool b -> do me <- getModuleEnv
                           setModuleEnv me { M.meMonoBinds = b }
           _         -> return ()
 
-  , OptionDescr "tc-solver" (EnvProg "z3" [ "-smt2", "-in" ])
+  , OptionDescr "tcSolver" ["tc-solver"] (EnvProg "z3" [ "-smt2", "-in" ])
     noCheck  -- TODO: check for the program in the path
     "The solver that will be used by the type checker." $
-    \case EnvProg prog args -> do me <- getModuleEnv
-                                  let cfg = M.meSolverConfig me
-                                  setModuleEnv me { M.meSolverConfig =
-                                                      cfg { T.solverPath = prog
-                                                          , T.solverArgs = args } }
+    \case EnvProg prog args -> do modifyRW_ (\rw -> rw { eTCConfig = (eTCConfig rw)
+                                                                      { T.solverPath = prog
+                                                                      , T.solverArgs = args
+                                                                      }})
+                                  resetTCSolver
           _                 -> return ()
 
-  , OptionDescr "tc-debug" (EnvNum 0)
+  , OptionDescr "tcDebug" ["tc-debug"] (EnvNum 0)
     noCheck
-    "Enable type-checker debugging output." $
-    \case EnvNum n -> do me <- getModuleEnv
-                         let cfg = M.meSolverConfig me
-                         setModuleEnv me { M.meSolverConfig = cfg{ T.solverVerbose = n } }
+    (unlines
+      [ "Enable type-checker debugging output:"
+      , "  *  0  no debug output"
+      , "  *  1  show type-checker debug info"
+      , "  * >1  show type-checker debug info and interactions with SMT solver"]) $
+    \case EnvNum n -> do changed <- modifyRW (\rw -> ( rw{ eTCConfig = (eTCConfig rw){ T.solverVerbose = n } }
+                                                     , n /= T.solverVerbose (eTCConfig rw)
+                                                     ))
+                         when changed resetTCSolver
           _        -> return ()
-  , OptionDescr "core-lint" (EnvBool False)
+  , OptionDescr "coreLint" ["core-lint"] (EnvBool False)
     noCheck
     "Enable sanity checking of type-checker." $
       let setIt x = do me <- getModuleEnv
@@ -804,17 +979,87 @@ userOptions  = mkOptionMap
                EnvBool False -> setIt M.NoCoreLint
                _             -> return ()
 
-  , simpleOpt "prover-stats" (EnvBool True) noCheck
+  , simpleOpt "hashConsing" ["hash-consing"] (EnvBool True) noCheck
+    "Enable hash-consing in the What4 symbolic backends."
+
+  , simpleOpt "proverStats" ["prover-stats"] (EnvBool True) noCheck
     "Enable prover timing statistics."
 
-  , simpleOpt "prover-validate" (EnvBool False) noCheck
+  , simpleOpt "proverValidate" ["prover-validate"] (EnvBool False) noCheck
     "Validate :sat examples and :prove counter-examples for correctness."
 
-  , simpleOpt "show-examples" (EnvBool True) noCheck
+  , simpleOpt "showExamples" ["show-examples"] (EnvBool True) noCheck
     "Print the (counter) example after :sat or :prove"
+
+  , simpleOpt "fpBase" ["fp-base"] (EnvNum 16) checkBase
+    "The base to display floating point numbers at (2, 8, 10, or 16)."
+
+  , simpleOpt "fpFormat" ["fp-format"] (EnvString "free") checkPPFloatFormat
+    $ unlines
+    [ "Specifies the format to use when showing floating point numbers:"
+    , "  * free      show using as many digits as needed"
+    , "  * free+exp  like `free` but always show exponent"
+    , "  * .NUM      show NUM (>=1) digits after floating point"
+    , "  * NUM       show using NUM (>=1) significant digits"
+    , "  * NUM+exp   like NUM but always show exponent"
+    ]
+
+  , simpleOpt "ignoreSafety" ["ignore-safety"] (EnvBool False) noCheck
+    "Ignore safety predicates when performing :sat or :prove checks"
+
+  , simpleOpt "fieldOrder" ["field-order"] (EnvString "display") checkFieldOrder
+    $ unlines
+    [ "The order that record fields are displayed in."
+    , "  * display      try to match the order they were written in the source code"
+    , "  * canonical    use a predictable, canonical order"
+    ]
+
+  , simpleOpt "timeMeasurementPeriod" ["time-measurement-period"] (EnvNum 10)
+    checkTimeMeasurementPeriod
+    $ unlines
+    [ "The period of time in seconds to spend collecting measurements when"
+    , "  running :time."
+    , "This is a lower bound and the actual time taken might be higher if the"
+    , "  evaluation takes a long time."
+    ]
+
+  , simpleOpt "timeQuiet" ["time-quiet"] (EnvBool False) noCheck
+    "Suppress output of :time command and only bind result to `it`."
   ]
 
 
+parsePPFloatFormat :: String -> Maybe PPFloatFormat
+parsePPFloatFormat s =
+  case s of
+    "free"     -> Just $ FloatFree AutoExponent
+    "free+exp" -> Just $ FloatFree ForceExponent
+    '.' : xs   -> FloatFrac <$> readMaybe xs
+    _ | (as,res) <- break (== '+') s
+      , Just n   <- readMaybe as
+      , Just e   <- case res of
+                      "+exp" -> Just ForceExponent
+                      ""     -> Just AutoExponent
+                      _      -> Nothing
+        -> Just (FloatFixed n e)
+    _  -> Nothing
+
+checkPPFloatFormat :: Checker
+checkPPFloatFormat val =
+  case val of
+    EnvString s | Just _ <- parsePPFloatFormat s -> noWarns Nothing
+    _ -> noWarns $ Just "Failed to parse `fp-format`"
+
+parseFieldOrder :: String -> Maybe FieldOrder
+parseFieldOrder "canonical" = Just CanonicalOrder
+parseFieldOrder "display" = Just DisplayOrder
+parseFieldOrder _ = Nothing
+
+checkFieldOrder :: Checker
+checkFieldOrder val =
+  case val of
+    EnvString s | Just _ <- parseFieldOrder s -> noWarns Nothing
+    _ -> noWarns $ Just "Failed to parse field-order (expected one of \"canonical\" or \"display\")"
+
 -- | Check the value to the `base` option.
 checkBase :: Checker
 checkBase val = case val of
@@ -832,22 +1077,30 @@ checkInfLength val = case val of
 
 checkProver :: Checker
 checkProver val = case val of
-  EnvString s
-    | s `notElem` proverNames ->
-      noWarns $ Just $ "Prover must be " ++ proverListString
-    | s `elem` ["offline", "any"] -> noWarns Nothing
+  EnvString (map toLower -> s)
+    | s `elem` W4.proverNames ->
+      io (W4.setupProver s) >>= \case
+        Left msg -> noWarns (Just msg)
+        Right (ws, cfg) ->
+          do modifyRW_ (\rw -> rw{ eProverConfig = Right cfg })
+             return (Nothing, ws)
+    | s `elem` SBV.proverNames ->
+      io (SBV.setupProver s) >>= \case
+        Left msg -> noWarns (Just msg)
+        Right (ws, cfg) ->
+          do modifyRW_ (\rw -> rw{ eProverConfig = Left cfg })
+             return (Nothing, ws)
+
     | otherwise ->
-      do let prover = lookupProver s
-         available <- sbvCheckSolverInstallation prover
-         let ws = if available
-                     then []
-                     else ["Warning: " ++ s ++ " installation not found"]
-         return (Nothing, ws)
+      noWarns $ Just $ "Prover must be " ++ proverListString
 
   _ -> noWarns $ Just "unable to parse a value for prover"
 
+allProvers :: [String]
+allProvers = SBV.proverNames ++ W4.proverNames
+
 proverListString :: String
-proverListString = concatMap (++ ", ") (init proverNames) ++ "or " ++ last proverNames
+proverListString = concatMap (++ ", ") (init allProvers) ++ "or " ++ last allProvers
 
 checkSatNum :: Checker
 checkSatNum val = case val of
@@ -867,6 +1120,14 @@ getUserSatNum = do
     _                         -> panic "REPL.Monad.getUserSatNum"
                                    [ "invalid satNum option" ]
 
+checkTimeMeasurementPeriod :: Checker
+checkTimeMeasurementPeriod (EnvNum n)
+  | n >= 1 = noWarns Nothing
+  | otherwise = noWarns $
+    Just "timeMeasurementPeriod must be a positive integer"
+checkTimeMeasurementPeriod _ = noWarns $
+  Just "unable to parse value for timeMeasurementPeriod"
+
 -- Environment Utilities -------------------------------------------------------
 
 whenDebug :: REPL () -> REPL ()
@@ -901,4 +1162,3 @@ z3exists = do
   case mPath of
     Nothing -> return (Just Z3NotFound)
     Just _  -> return Nothing
-
