@@ -10,62 +10,62 @@
 -- slow down in some cases.  What's the problem?  Consider the following (common)
 -- patterns:
 --
---     fibs = [0,1] # [ x + y | x <- fibs, y <- drop`{1} fibs ]
+-- >    fibs = [0,1] # [ x + y | x <- fibs, y <- drop`{1} fibs ]
 --
--- The type of `fibs` is:
+-- The type of @fibs@ is:
 --
---     {a} (a >= 1, fin a) => [inf][a]
+-- >    {a} (a >= 1, fin a) => [inf][a]
 --
--- Here `a` is the number of bits to be used in the values computed by `fibs`.
--- When we evaluate `fibs`, `a` becomes a parameter to `fibs`, which works
--- except that now `fibs` is a function, and we don't get any of the memoization
+-- Here @a@ is the number of bits to be used in the values computed by @fibs@.
+-- When we evaluate @fibs@, @a@ becomes a parameter to @fibs@, which works
+-- except that now @fibs@ is a function, and we don't get any of the memoization
 -- we might expect!  What looked like an efficient implementation has all
 -- of a sudden become exponential!
 --
--- Note that this is only a problem for polymorphic values: if `fibs` was
+-- Note that this is only a problem for polymorphic values: if @fibs@ was
 -- already a function, it would not be that surprising that it does not
 -- get cached.
 --
 -- So, to avoid this, we try to spot recursive polymorphic values,
 -- where the recursive occurrences have the exact same type parameters
--- as the definition.  For example, this is the case in `fibs`: each
--- recursive call to `fibs` is instantiated with exactly the same
--- type parameter (i.e., `a`).  The rewrite we do is as follows:
+-- as the definition.  For example, this is the case in @fibs@: each
+-- recursive call to @fibs@ is instantiated with exactly the same
+-- type parameter (i.e., @a@).  The rewrite we do is as follows:
 --
---     fibs : {a} (a >= 1, fin a) => [inf][a]
---     fibs = \{a} (a >= 1, fin a) -> fibs'
---       where fibs' : [inf][a]
---             fibs' = [0,1] # [ x + y | x <- fibs', y <- drop`{1} fibs' ]
+-- >    fibs : {a} (a >= 1, fin a) => [inf][a]
+-- >    fibs = \{a} (a >= 1, fin a) -> fibs'
+-- >      where fibs' : [inf][a]
+-- >            fibs' = [0,1] # [ x + y | x <- fibs', y <- drop`{1} fibs' ]
 --
 -- After the rewrite, the recursion is monomorphic (i.e., we are always using
--- the same type).  As a result, `fibs'` is an ordinary recursive value,
+-- the same type).  As a result, @fibs'@ is an ordinary recursive value,
 -- where we get the benefit of caching.
 --
 -- The rewrite is a bit more complex, when there are multiple mutually
 -- recursive functions.  Here is an example:
 --
---     zig : {a} (a >= 2, fin a) => [inf][a]
---     zig = [1] # zag
---
---     zag : {a} (a >= 2, fin a) => [inf][a]
---     zag = [2] # zig
+-- >    zig : {a} (a >= 2, fin a) => [inf][a]
+-- >    zig = [1] # zag
+-- >
+-- >    zag : {a} (a >= 2, fin a) => [inf][a]
+-- >    zag = [2] # zig
 --
 -- This gets rewritten to:
 --
---     newName : {a} (a >= 2, fin a) => ([inf][a], [inf][a])
---     newName = \{a} (a >= 2, fin a) -> (zig', zag')
---       where
---       zig' : [inf][a]
---       zig' = [1] # zag'
---
---       zag' : [inf][a]
---       zag' = [2] # zig'
---
---     zig : {a} (a >= 2, fin a) => [inf][a]
---     zig = \{a} (a >= 2, fin a) -> (newName a <> <> ).1
---
---     zag : {a} (a >= 2, fin a) => [inf][a]
---     zag = \{a} (a >= 2, fin a) -> (newName a <> <> ).2
+-- >    newName : {a} (a >= 2, fin a) => ([inf][a], [inf][a])
+-- >    newName = \{a} (a >= 2, fin a) -> (zig', zag')
+-- >      where
+-- >      zig' : [inf][a]
+-- >      zig' = [1] # zag'
+-- >
+-- >      zag' : [inf][a]
+-- >      zag' = [2] # zig'
+-- >
+-- >    zig : {a} (a >= 2, fin a) => [inf][a]
+-- >    zig = \{a} (a >= 2, fin a) -> (newName a <> <> ).1
+-- >
+-- >    zag : {a} (a >= 2, fin a) => [inf][a]
+-- >    zag = \{a} (a >= 2, fin a) -> (newName a <> <> ).2
 --
 -- NOTE:  We are assuming that no capture would occur with binders.
 -- For values, this is because we replaces things with freshly chosen variables.
@@ -79,21 +79,22 @@
 module Cryptol.Transform.MonoValues (rewModule) where
 
 import Cryptol.ModuleSystem.Name
-        (SupplyT,liftSupply,Supply,mkDeclared,NameSource(..))
+        (SupplyT,liftSupply,Supply,mkDeclared,NameSource(..),ModPath(..))
 import Cryptol.Parser.Position (emptyRange)
 import Cryptol.TypeCheck.AST hiding (splitTApp) -- XXX: just use this one
 import Cryptol.TypeCheck.TypeMap
-import Cryptol.Utils.Ident (ModName)
-import Data.List(sortBy,groupBy)
+import Cryptol.Utils.Ident(Namespace(..))
+import Data.List(sortBy)
 import Data.Either(partitionEithers)
 import Data.Map (Map)
+import qualified Data.List.NonEmpty as NE
 import MonadLib hiding (mapM)
 
 import Prelude ()
 import Prelude.Compat
 
-{- (f,t,n) |--> x  means that when we spot instantiations of `f` with `ts` and
-`n` proof argument, we should replace them with `Var x` -}
+{- (f,t,n) |--> x  means that when we spot instantiations of @f@ with @ts@ and
+@n@ proof argument, we should replace them with @Var x@ -}
 newtype RewMap' a = RM (Map Name (TypesMap (Map Int a)))
 type RewMap = RewMap' Name
 
@@ -132,7 +133,7 @@ instance TrieMap RewMap' (Name,[Type],Int) where
 -- | Note that this assumes that this pass will be run only once for each
 -- module, otherwise we will get name collisions.
 rewModule :: Supply -> Module -> (Module,Supply)
-rewModule s m = runM body (mName m) s
+rewModule s m = runM body (TopModule (mName m)) s
   where
   body = do ds <- mapM (rewDeclGroup emptyTM) (mDecls m)
             return m { mDecls = ds }
@@ -140,13 +141,13 @@ rewModule s m = runM body (mName m) s
 --------------------------------------------------------------------------------
 
 type M  = ReaderT RO (SupplyT Id)
-type RO = ModName
+type RO = ModPath
 
 -- | Produce a fresh top-level name.
 newName :: M Name
 newName  =
   do ns <- ask
-     liftSupply (mkDeclared ns SystemName "$mono" Nothing emptyRange)
+     liftSupply (mkDeclared NSValue ns SystemName "$mono" Nothing emptyRange)
 
 newTopOrLocalName :: M Name
 newTopOrLocalName  = newName
@@ -179,12 +180,12 @@ rewE rews = go
                           Nothing  -> EProofApp <$> go e
                           Just yes -> return yes
 
+      ELocated r t    -> ELocated r <$> go t
       EList es t      -> EList   <$> mapM go es <*> return t
       ETuple es       -> ETuple  <$> mapM go es
-      ERec fs         -> ERec    <$> (forM fs $ \(f,e) -> do e1 <- go e
-                                                             return (f,e1))
+      ERec fs         -> ERec    <$> traverse go fs
       ESel e s        -> ESel    <$> go e  <*> return s
-      ESet e s v      -> ESet    <$> go e  <*> return s <*> go v
+      ESet ty e s v   -> ESet ty <$> go e  <*> return s <*> go v
       EIf e1 e2 e3    -> EIf     <$> go e1 <*> go e2 <*> go e3
 
       EComp len t e mss -> EComp len t <$> go e  <*> mapM (mapM (rewM rews)) mss
@@ -200,6 +201,8 @@ rewE rews = go
       EWhere e dgs    -> EWhere      <$> go e <*> inLocal
                                                   (mapM (rewDeclGroup rews) dgs)
 
+      EPropGuards guards ty -> EPropGuards <$> (\(props, e) -> (,) <$> pure props <*> go e) `traverse` guards <*> pure ty
+
 
 rewM :: RewMap -> Match -> M Match
 rewM rews ma =
@@ -215,8 +218,9 @@ rewD rews d = do e <- rewDef rews (dDefinition d)
                  return d { dDefinition = e }
 
 rewDef :: RewMap -> DeclDef -> M DeclDef
-rewDef rews (DExpr e) = DExpr <$> rewE rews e
-rewDef _    DPrim     = return DPrim
+rewDef rews (DExpr e)    = DExpr <$> rewE rews e
+rewDef _    DPrim        = return DPrim
+rewDef _    (DForeign t) = return $ DForeign t
 
 rewDeclGroup :: RewMap -> DeclGroup -> M DeclGroup
 rewDeclGroup rews dg =
@@ -224,7 +228,7 @@ rewDeclGroup rews dg =
     NonRecursive d -> NonRecursive <$> rewD rews d
     Recursive ds ->
       do let (leave,rew) = partitionEithers (map consider ds)
-             rewGroups   = groupBy sameTParams
+             rewGroups   = NE.groupBy sameTParams
                          $ sortBy compareTParams rew
          ds1 <- mapM (rewD rews) leave
          ds2 <- mapM rewSame rewGroups
@@ -236,17 +240,18 @@ rewDeclGroup rews dg =
 
   consider d   =
     case dDefinition d of
-      DPrim   -> Left d
-      DExpr e -> let (tps,props,e') = splitTParams e
-                 in if not (null tps) && notFun e'
-                     then Right (d, tps, props, e')
-                     else Left d
+      DPrim      -> Left d
+      DForeign _ -> Left d
+      DExpr e    -> let (tps,props,e') = splitTParams e
+                    in if not (null tps) && notFun e'
+                        then Right (d, tps, props, e')
+                        else Left d
 
   rewSame ds =
-     do new <- forM ds $ \(d,_,_,e) ->
+     do new <- forM (NE.toList ds) $ \(d,_,_,e) ->
                  do x <- newName
                     return (d, x, e)
-        let (_,tps,props,_) : _ = ds
+        let (_,tps,props,_) NE.:| _ = ds
             tys            = map (TVar . tpVar) tps
             proofNum       = length props
             addRew (d,x,_) = insertTM (dName d,tys,proofNum) x
