@@ -1,3 +1,6 @@
+-- See Note [-Wincomplete-uni-patterns and irrefutable patterns] in Cryptol.TypeCheck.TypePat
+{-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
+
 -- |
 -- Module      :  Cryptol.Transform.Specialize
 -- Copyright   :  (c) 2013-2016 Galois, Inc.
@@ -16,20 +19,22 @@ import qualified Cryptol.ModuleSystem as M
 import qualified Cryptol.ModuleSystem.Env as M
 import qualified Cryptol.ModuleSystem.Monad as M
 import Cryptol.ModuleSystem.Name
+import Cryptol.Utils.Ident(OrigName(..))
+import Cryptol.Eval (checkProp)
 
 import           Data.Map (Map)
 import qualified Data.Map as Map
 import Data.Maybe (catMaybes)
+import qualified Data.List as List
 
 import MonadLib hiding (mapM)
+import Cryptol.ModuleSystem.Base (getPrimMap)
 
-import Prelude ()
-import Prelude.Compat
 
 -- Specializer Monad -----------------------------------------------------------
 
--- | A Name should have an entry in the SpecCache iff it is
--- specializable. Each Name starts out with an empty TypesMap.
+-- | A 'Name' should have an entry in the 'SpecCache' iff it is
+-- specializable. Each 'Name' starts out with an empty 'TypesMap'.
 type SpecCache = Map Name (Decl, TypesMap (Name, Maybe Decl))
 
 -- | The specializer monad.
@@ -58,26 +63,27 @@ modify f = get >>= (set . f)
 
 -- Specializer -----------------------------------------------------------------
 
--- | Add a `where` clause to the given expression containing
+-- | Add a @where@ clause to the given expression containing
 -- type-specialized versions of all functions called (transitively) by
 -- the body of the expression.
 specialize :: Expr -> M.ModuleCmd Expr
-specialize expr (ev,modEnv) = run $ do
-  let extDgs = allDeclGroups modEnv
+specialize expr minp = run $ do
+  let extDgs = allDeclGroups (M.minpModuleEnv minp)
   let (tparams, expr') = destETAbs expr
   spec' <- specializeEWhere expr' extDgs
   return (foldr ETAbs spec' tparams)
   where
-  run = M.runModuleT (ev,modEnv) . fmap fst . runSpecT Map.empty
+  run = M.runModuleT minp . fmap fst . runSpecT Map.empty
 
 specializeExpr :: Expr -> SpecM Expr
 specializeExpr expr =
   case expr of
+    ELocated r e  -> ELocated r <$> specializeExpr e
     EList es t    -> EList <$> traverse specializeExpr es <*> pure t
     ETuple es     -> ETuple <$> traverse specializeExpr es
-    ERec fs       -> ERec <$> traverse (traverseSnd specializeExpr) fs
+    ERec fs       -> ERec <$> traverse specializeExpr fs
     ESel e s      -> ESel <$> specializeExpr e <*> pure s
-    ESet e s v    -> ESet <$> specializeExpr e <*> pure s <*> specializeExpr v
+    ESet ty e s v -> ESet ty <$> specializeExpr e <*> pure s <*> specializeExpr v
     EIf e1 e2 e3  -> EIf <$> specializeExpr e1 <*> specializeExpr e2 <*> specializeExpr e3
     EComp len t e mss -> EComp len t <$> specializeExpr e <*> traverse (traverse specializeMatch) mss
     -- Bindings within list comprehensions always have monomorphic types.
@@ -88,19 +94,29 @@ specializeExpr expr =
       e' <- specializeExpr e
       setSpecCache cache
       return (ETAbs t e')
-    -- We need to make sure that after processing `e`, no specialized
-    -- decls mentioning type variable `t` escape outside the
-    -- `ETAbs`. To avoid this, we reset to an empty SpecCache while we
-    -- run `specializeExpr e`, and restore it afterward: this
+    -- We need to make sure that after processing @e@, no specialized
+    -- decls mentioning type variable @t@ escape outside the
+    -- 'ETAbs'. To avoid this, we reset to an empty 'SpecCache' while we
+    -- run @'specializeExpr' e@, and restore it afterward: this
     -- effectively prevents the specializer from registering any type
-    -- instantiations involving `t` for any decls bound outside the
-    -- scope of `t`.
+    -- instantiations involving @t@ for any decls bound outside the
+    -- scope of @t@.
     ETApp {}      -> specializeConst expr
     EApp e1 e2    -> EApp <$> specializeExpr e1 <*> specializeExpr e2
     EAbs qn t e   -> EAbs qn t <$> specializeExpr e
     EProofAbs p e -> EProofAbs p <$> specializeExpr e
     EProofApp {}  -> specializeConst expr
     EWhere e dgs  -> specializeEWhere e dgs
+    -- The type should be monomorphic, and the guarded expressions should
+    -- already be normalized, so we just need to choose the first expression
+    -- that's true.
+    EPropGuards guards ty -> 
+      case List.find (all checkProp . fst) guards of
+        Just (_, e) -> specializeExpr e
+        Nothing -> do
+          pm <- liftSpecT getPrimMap
+          pure $ eError pm ty "no constraint guard was satisfied"
+
 
 specializeMatch :: Match -> SpecM Match
 specializeMatch (From qn l t e) = From qn l t <$> specializeExpr e
@@ -144,9 +160,9 @@ withDeclGroups dgs action = do
   modifySpecCache (Map.union savedCache . flip Map.difference newCache)
   return (result, dgs', nameTable)
 
--- | Compute the specialization of `EWhere e dgs`. A decl within `dgs`
+-- | Compute the specialization of @'EWhere' e dgs@. A decl within @dgs@
 -- is replicated once for each monomorphic type instance at which it
--- is used; decls not mentioned in `e` (even monomorphic ones) are
+-- is used; decls not mentioned in @e@ (even monomorphic ones) are
 -- simply dropped.
 specializeEWhere :: Expr -> [DeclGroup] -> SpecM Expr
 specializeEWhere e dgs = do
@@ -186,9 +202,10 @@ specializeConst e0 = do
                  sig' <- instantiateSchema ts n (dSignature decl)
                  modifySpecCache (Map.adjust (fmap (insertTM ts (qname', Nothing))) qname)
                  rhs' <- case dDefinition decl of
-                           DExpr e -> do e' <- specializeExpr =<< instantiateExpr ts n e
-                                         return (DExpr e')
-                           DPrim   -> return DPrim
+                           DExpr e    -> do e' <- specializeExpr =<< instantiateExpr ts n e
+                                            return (DExpr e')
+                           DPrim      -> return DPrim
+                           DForeign t -> return $ DForeign t
                  let decl' = decl { dName = qname', dSignature = sig', dDefinition = rhs' }
                  modifySpecCache (Map.adjust (fmap (insertTM ts (qname', Just decl'))) qname)
                  return (EVar qname')
@@ -224,7 +241,7 @@ destETAbs = go []
 -- Any top-level declarations in the current module can be found in the
 -- ModuleEnv's LoadedModules, and so we can count of freshName to avoid
 -- collisions with them.  Any generated name for a
--- specialized function will be qualified with the current @ModName@, so genned
+-- specialized function will be qualified with the current 'ModName', so genned
 -- names will not collide with local decls either.
 -- freshName :: Name -> [Type] -> SpecM Name
 -- freshName n [] = return n
@@ -244,9 +261,10 @@ destETAbs = go []
 freshName :: Name -> [Type] -> SpecM Name
 freshName n _ =
   case nameInfo n of
-    Declared ns s -> liftSupply (mkDeclared ns s ident fx loc)
-    Parameter     -> liftSupply (mkParameter ident loc)
+    GlobalName s og -> liftSupply (mkDeclared ns (ogModule og) s ident fx loc)
+    LocalName {}    -> liftSupply (mkLocal ns ident loc)
   where
+  ns    = nameNamespace n
   fx    = nameFixity n
   ident = nameIdent n
   loc   = nameLoc n
@@ -322,12 +340,12 @@ instantiateSchema ts n (Forall params props ty)
   | otherwise                  = return $ Forall [] [] (apSubst sub ty)
   where sub = listParamSubst (zip params ts)
 
--- | Reduce `length ts` outermost type abstractions and `n` proof abstractions.
+-- | Reduce @length ts@ outermost type abstractions and @n@ proof abstractions.
 instantiateExpr :: [Type] -> Int -> Expr -> SpecM Expr
 instantiateExpr [] 0 e = return e
 instantiateExpr [] n (EProofAbs _ e) = instantiateExpr [] (n - 1) e
 instantiateExpr (t : ts) n (ETAbs param e) =
-  instantiateExpr ts n (apSubst (singleSubst (tpVar param) t) e)
+  instantiateExpr ts n (apSubst (singleTParamSubst param t) e)
 instantiateExpr _ _ _ = fail "instantiateExpr: wrong number of type/proof arguments"
 
 
