@@ -9,26 +9,31 @@
 
 module Cryptol.TypeCheck
   ( tcModule
-  , tcModuleInst
   , tcExpr
   , tcDecls
   , InferInput(..)
   , InferOutput(..)
   , SolverConfig(..)
+  , defaultSolverConfig
   , NameSeeds
   , nameSeeds
   , Error(..)
   , Warning(..)
   , ppWarning
   , ppError
+  , WithNames(..)
+  , NameMap
+  , ppNamedWarning
+  , ppNamedError
   ) where
 
+import Data.Map(Map)
+
 import           Cryptol.ModuleSystem.Name
-                    (liftSupply,mkDeclared,NameSource(..))
+                    (liftSupply,mkDeclared,NameSource(..),ModPath(..))
 import qualified Cryptol.Parser.AST as P
 import           Cryptol.Parser.Position(Range,emptyRange)
 import           Cryptol.TypeCheck.AST
-import           Cryptol.TypeCheck.Depends (FromDecl)
 import           Cryptol.TypeCheck.Error
 import           Cryptol.TypeCheck.Monad
                    ( runInferM
@@ -37,33 +42,21 @@ import           Cryptol.TypeCheck.Monad
                    , NameSeeds
                    , nameSeeds
                    , lookupVar
+                   , newLocalScope, endLocalScope
                    )
-import           Cryptol.TypeCheck.Infer (inferModule, inferBinds, inferDs)
-import           Cryptol.TypeCheck.InferTypes(VarType(..), SolverConfig(..))
-import           Cryptol.TypeCheck.Solve(proveModuleTopLevel)
-import           Cryptol.TypeCheck.CheckModuleInstance(checkModuleInstance)
-import           Cryptol.TypeCheck.Monad(withParamType,withParameterConstraints)
-import           Cryptol.Utils.Ident (exprModName,packIdent)
-import           Cryptol.Utils.PP
-import           Cryptol.Utils.Panic(panic)
+import Cryptol.TypeCheck.Infer (inferTopModule, inferBinds, checkTopDecls)
+import Cryptol.TypeCheck.InferTypes(VarType(..), SolverConfig(..), defaultSolverConfig)
+import Cryptol.TypeCheck.Solve(proveModuleTopLevel)
+-- import Cryptol.TypeCheck.Monad(withParamType,withParameterConstraints)
+import Cryptol.TypeCheck.PP(WithNames(..),NameMap)
+import Cryptol.Utils.Ident (exprModName,packIdent,Namespace(..))
+import Cryptol.Utils.PP
+import Cryptol.Utils.Panic(panic)
 
-tcModule :: P.Module Name -> InferInput -> IO (InferOutput Module)
-tcModule m inp = runInferM inp (inferModule m)
 
--- | Check a module instantiation, assuming that the functor has already
--- been checked.
-tcModuleInst :: Module                  {- ^ functor -} ->
-                P.Module Name           {- params -} ->
-                InferInput              {- ^ TC settings -} ->
-                IO (InferOutput Module) {- ^ new version of instance -}
-tcModuleInst func m inp = runInferM inp
-                        $ do x <- inferModule m
-                             y <- checkModuleInstance func x
-                             flip (foldr withParamType) (mParamTypes x) $
-                               withParameterConstraints (mParamConstraints x) $
-                               proveModuleTopLevel
 
-                             return y
+tcModule :: P.Module Name -> InferInput -> IO (InferOutput TCTopEntity)
+tcModule m inp = runInferM inp (inferTopModule m)
 
 tcExpr :: P.Expr Name -> InferInput -> IO (InferOutput (Expr,Schema))
 tcExpr e0 inp = runInferM inp
@@ -74,7 +67,9 @@ tcExpr e0 inp = runInferM inp
   where
   go loc expr =
     case expr of
-      P.ELocated e loc' -> go loc' e
+      P.ELocated e loc' ->
+        do (te, sch) <- go loc' e
+           pure $! if inpCallStacks inp then (ELocated loc' te, sch) else (te,sch)
       P.EVar x  ->
         do res <- lookupVar x
            case res of
@@ -84,8 +79,9 @@ tcExpr e0 inp = runInferM inp
                              , show e'
                              , show t
                              ]
-      _ -> do fresh <- liftSupply (mkDeclared exprModName SystemName
-                                      (packIdent "(expression)") Nothing loc)
+      _ -> do fresh <- liftSupply $
+                        mkDeclared NSValue (TopModule exprModName) SystemName
+                                      (packIdent "(expression)") Nothing loc
               res   <- inferBinds True False
                 [ P.Bind
                     { P.bName      = P.Located { P.srcRange = loc, P.thing = fresh }
@@ -97,6 +93,7 @@ tcExpr e0 inp = runInferM inp
                     , P.bInfix     = False
                     , P.bFixity    = Nothing
                     , P.bDoc       = Nothing
+                    , P.bExport    = Public
                     } ]
 
               case res of
@@ -111,15 +108,24 @@ tcExpr e0 inp = runInferM inp
                           : map show res
                           )
 
-tcDecls :: FromDecl d => [d] -> InferInput -> IO (InferOutput [DeclGroup])
-tcDecls ds inp = runInferM inp $ inferDs ds $ \dgs -> do
-                   proveModuleTopLevel
-                   return dgs
+tcDecls :: [P.TopDecl Name] -> InferInput -> IO (InferOutput ([DeclGroup],Map Name TySyn))
+tcDecls ds inp = runInferM inp $
+  do newLocalScope
+     checkTopDecls ds
+     proveModuleTopLevel
+     endLocalScope
 
 ppWarning :: (Range,Warning) -> Doc
-ppWarning (r,w) = text "[warning] at" <+> pp r <.> colon $$ nest 2 (pp w)
+ppWarning (r,w) = nest 2 (text "[warning] at" <+> pp r <.> colon $$ pp w)
 
 ppError :: (Range,Error) -> Doc
-ppError (r,w) = text "[error] at" <+> pp r <.> colon $$ nest 2 (pp w)
+ppError (r,w) = nest 2 (text "[error] at" <+> pp r <.> colon $$ pp w)
 
+ppNamedWarning :: NameMap -> (Range,Warning) -> Doc
+ppNamedWarning nm (r,w) =
+  nest 2 (text "[warning] at" <+> pp r <.> colon $$ pp (WithNames w nm))
+
+ppNamedError :: NameMap -> (Range,Error) -> Doc
+ppNamedError nm (r,e) =
+  nest 2 (text "[error] at" <+> pp r <.> colon $$ pp (WithNames e nm))
 
