@@ -18,6 +18,8 @@
 {-# LANGUAGE ViewPatterns #-}
 module Cryptol.TypeCheck.InferTypes where
 
+import           Control.Monad(guard)
+
 import           Cryptol.Parser.Position
 import           Cryptol.ModuleSystem.Name (asPrim,nameLoc)
 import           Cryptol.TypeCheck.AST
@@ -25,10 +27,9 @@ import           Cryptol.TypeCheck.PP
 import           Cryptol.TypeCheck.Subst
 import           Cryptol.TypeCheck.TypePat
 import           Cryptol.TypeCheck.SimpType(tMax)
-import           Cryptol.Utils.Ident (ModName, identText)
+import           Cryptol.Utils.Ident (PrimIdent(..), preludeName)
 import           Cryptol.Utils.Panic(panic)
 import           Cryptol.Utils.Misc(anyJust)
-import           Cryptol.Utils.Patterns(matchMaybe)
 
 import           Data.Set ( Set )
 import qualified Data.Set as Set
@@ -46,6 +47,19 @@ data SolverConfig = SolverConfig
     -- ^ Look for the solver prelude in these locations.
   } deriving (Show, Generic, NFData)
 
+
+-- | A default configuration for using Z3, where
+--   the solver prelude is expected to be found
+--   in the given search path.
+defaultSolverConfig :: [FilePath] -> SolverConfig
+defaultSolverConfig searchPath =
+  SolverConfig
+  { solverPath = "z3"
+  , solverArgs = [ "-smt2", "-in" ]
+  , solverVerbose = 0
+  , solverPreludePath = searchPath
+  }
+
 -- | The types of variables in the environment.
 data VarType = ExtVar Schema
                -- ^ Known type
@@ -60,8 +74,15 @@ data Goals = Goals
   { goalSet :: Set Goal
     -- ^ A bunch of goals, not including the ones in 'literalGoals'.
 
+  , saturatedPropSet :: Set Prop
+    -- ^ The set of nonliteral goals, saturated by all superclass implications
+
   , literalGoals :: Map TVar LitGoal
     -- ^ An entry @(a,t)@ corresponds to @Literal t a@.
+
+  , literalLessThanGoals :: Map TVar LitGoal
+    -- ^ An entry @(a,t)@ corresponds to @LiteralLessThan t a@.
+
   } deriving (Show)
 
 -- | This abuses the type 'Goal' a bit. The 'goal' field contains
@@ -80,30 +101,72 @@ goalToLitGoal g =
      return (a, g { goal = tn })
 
 
+litLessThanGoalToGoal :: (TVar,LitGoal) -> Goal
+litLessThanGoalToGoal (a,g) = g { goal = pLiteralLessThan (goal g) (TVar a) }
+
+goalToLitLessThanGoal :: Goal -> Maybe (TVar,LitGoal)
+goalToLitLessThanGoal g =
+  do (tn,a) <- matchMaybe $ do (tn,b) <- aLiteralLessThan (goal g)
+                               a      <- aTVar b
+                               return (tn,a)
+     return (a, g { goal = tn })
+
 
 emptyGoals :: Goals
-emptyGoals  = Goals { goalSet = Set.empty, literalGoals = Map.empty }
+emptyGoals  =
+  Goals
+  { goalSet = Set.empty
+  , saturatedPropSet = Set.empty
+  , literalGoals = Map.empty
+  , literalLessThanGoals = Map.empty
+  }
 
 nullGoals :: Goals -> Bool
-nullGoals gs = Set.null (goalSet gs) && Map.null (literalGoals gs)
+nullGoals gs =
+  Set.null (goalSet gs) &&
+  Map.null (literalGoals gs) &&
+  Map.null (literalLessThanGoals gs)
 
 fromGoals :: Goals -> [Goal]
 fromGoals gs = map litGoalToGoal (Map.toList (literalGoals gs)) ++
+               map litLessThanGoalToGoal (Map.toList (literalLessThanGoals gs)) ++
                Set.toList (goalSet gs)
 
 goalsFromList :: [Goal] -> Goals
 goalsFromList = foldr insertGoal emptyGoals
 
 insertGoal :: Goal -> Goals -> Goals
-insertGoal g gs
+insertGoal g gls
   | Just (a,newG) <- goalToLitGoal g =
-                gs { literalGoals = Map.insertWith jn a newG (literalGoals gs) }
-  | otherwise = gs { goalSet = Set.insert g (goalSet gs) }
+       -- XXX: here we are arbitrarily using the info of the first goal,
+       -- which could lead to a confusing location for a constraint.
+       let jn g1 g2 = g1 { goal = tMax (goal g1) (goal g2) } in
+       gls { literalGoals = Map.insertWith jn a newG (literalGoals gls)
+           , saturatedPropSet = Set.insert (pFin (TVar a)) (saturatedPropSet gls)
+           }
+
+  | Just (a,newG) <- goalToLitLessThanGoal g =
+       let jn g1 g2 = g1 { goal = tMax (goal g1) (goal g2) } in
+       gls { literalLessThanGoals = Map.insertWith jn a newG (literalLessThanGoals gls)
+           }
+
+  -- If the goal is already implied by some other goal, skip it
+  | Set.member (goal g) (saturatedPropSet gls) = gls
+
+  -- Otherwise, it is not already implied, add it and saturate
+  | otherwise =
+       gls { goalSet = gs', saturatedPropSet = sps'  }
+
+       where
+       ips  = superclassSet (goal g)
+       igs  = Set.map (\p -> g{ goal = p}) ips
+
+       -- remove all the goals that are implied by ips
+       gs'  = Set.insert g (Set.difference (goalSet gls) igs)
+
+       -- add the goal and all its implied toals to the saturated set
+       sps' = Set.insert (goal g) (Set.union (saturatedPropSet gls) ips)
 
-  where
-  jn g1 g2 = g1 { goal = tMax (goal g1) (goal g2) }
-  -- XXX: here we are arbitrarily using the info of the first goal,
-  -- which could lead to a confusing location for a constraint.
 
 -- | Something that we need to find evidence for.
 data Goal = Goal
@@ -119,7 +182,8 @@ instance Ord Goal where
   compare x y = compare (goal x) (goal y)
 
 data HasGoal = HasGoal
-  { hasName :: !Int
+  { hasName :: !Int -- ^ This is the "name" of the constraint,
+                    -- used to find the solution for ellaboration.
   , hasGoal :: Goal
   } deriving Show
 
@@ -155,10 +219,21 @@ data ConstraintSource
   | CtDefaulting          -- ^ Just defaulting on the command line
   | CtPartialTypeFun Name -- ^ Use of a partial type function.
   | CtImprovement
-  | CtPattern Doc         -- ^ Constraints arising from type-checking patterns
-  | CtModuleInstance ModName -- ^ Instantiating a parametrized module
+  | CtPattern TypeSource  -- ^ Constraints arising from type-checking patterns
+  | CtModuleInstance Range -- ^ Instantiating a parametrized module
+  | CtPropGuardsExhaustive Name -- ^ Checking that a use of prop guards is exhastive
+  | CtFFI Name            -- ^ Constraints on a foreign declaration required
+                          --   by the FFI (e.g. sequences must be finite)
     deriving (Show, Generic, NFData)
 
+selSrc :: Selector -> TypeSource
+selSrc l = case l of
+             RecordSel la _ -> TypeOfRecordField la
+             TupleSel n _   -> TypeOfTupleField n
+             ListSel _ _    -> TypeOfSeqElement
+
+
+
 
 
 instance TVars ConstraintSource where
@@ -176,6 +251,8 @@ instance TVars ConstraintSource where
       CtImprovement    -> src
       CtPattern _      -> src
       CtModuleInstance _ -> src
+      CtPropGuardsExhaustive _ -> src
+      CtFFI _          -> src
 
 
 instance FVS Goal where
@@ -244,24 +321,25 @@ cppKind ki =
 addTVarsDescsAfter :: FVS t => NameMap -> t -> Doc -> Doc
 addTVarsDescsAfter nm t d
   | Set.null vs = d
+-- TODO? use `hang` here instead to indent things after "where"
   | otherwise   = d $$ text "where" $$ vcat (map desc (Set.toList vs))
   where
   vs     = fvs t
   desc v = ppWithNames nm v <+> text "is" <+> pp (tvInfo v)
 
 addTVarsDescsBefore :: FVS t => NameMap -> t -> Doc -> Doc
-addTVarsDescsBefore nm t d = frontMsg $$ d $$ backMsg
+addTVarsDescsBefore nm t d = vcat (frontMsg ++ [d] ++ backMsg)
   where
   (vs1,vs2)  = Set.partition isFreeTV (fvs t)
 
-  frontMsg | null vs1  = empty
-           | otherwise = "Failed to infer the following types:"
-                         $$ nest 2 (vcat (map desc1 (Set.toList vs1)))
+  frontMsg | null vs1  = []
+           | otherwise = [hang "Failed to infer the following types:"
+                             2 (vcat (map desc1 (Set.toList vs1)))]
   desc1 v    = "•" <+> ppWithNames nm v <.> comma <+> pp (tvInfo v)
 
-  backMsg  | null vs2  = empty
-           | otherwise = "where"
-                         $$ nest 2 (vcat (map desc2 (Set.toList vs2)))
+  backMsg  | null vs2  = []
+           | otherwise = [hang "where"
+                             2 (vcat (map desc2 (Set.toList vs2)))]
   desc2 v    = ppWithNames nm v <+> text "is" <+> pp (tvInfo v)
 
 
@@ -279,35 +357,44 @@ instance PP ConstraintSource where
       CtDefaulting    -> "defaulting"
       CtPartialTypeFun f -> "use of partial type function" <+> pp f
       CtImprovement   -> "examination of collected goals"
-      CtPattern desc  -> "checking a pattern:" <+> desc
-      CtModuleInstance n -> "module instantiation" <+> pp n
+      CtPattern ad    -> "checking a pattern:" <+> pp ad
+      CtModuleInstance r -> "module instantiation at" <+> pp r
+      CtPropGuardsExhaustive n -> "exhaustion check for prop guards used in defining" <+> pp n
+      CtFFI f         -> "declaration of foreign function" <+> pp f
 
 ppUse :: Expr -> Doc
 ppUse expr =
   case expr of
-    EVar (asPrim -> Just prim)
-      | identText prim == "number"       -> "literal or demoted expression"
-      | identText prim == "infFrom"      -> "infinite enumeration"
-      | identText prim == "infFromThen"  -> "infinite enumeration (with step)"
-      | identText prim == "fromTo"       -> "finite enumeration"
-      | identText prim == "fromThenTo"   -> "finite enumeration"
-    _                                    -> "expression" <+> pp expr
+    EVar (isPrelPrim -> Just prim)
+      | prim == "number"       -> "literal or demoted expression"
+      | prim == "fraction"     -> "fractional literal"
+      | prim == "infFrom"      -> "infinite enumeration"
+      | prim == "infFromThen"  -> "infinite enumeration (with step)"
+      | prim == "fromTo"       -> "finite enumeration"
+      | prim == "fromThenTo"   -> "finite enumeration"
+    _                          -> "expression" <+> pp expr
+  where
+  isPrelPrim x = do PrimIdent p i <- asPrim x
+                    guard (p == preludeName)
+                    pure i
 
 instance PP (WithNames Goal) where
   ppPrec _ (WithNames g names) =
-      (ppWithNames names (goal g)) $$
-               nest 2 (text "arising from" $$
-                       pp (goalSource g)   $$
-                       text "at" <+> pp (goalRange g))
+      hang (ppWithNames names (goal g))
+         2 (text "arising from" $$
+            pp (goalSource g)   $$
+            text "at" <+> pp (goalRange g))
 
 instance PP (WithNames DelayedCt) where
   ppPrec _ (WithNames d names) =
-    sig $$ "we need to show that" $$
-    nest 2 (vcat [ vars, asmps, "the following constraints hold:"
-                 , nest 2 $ vcat
-                          $ bullets
-                          $ map (ppWithNames ns1)
-                          $ dctGoals d ])
+    sig $$
+    hang "we need to show that"
+       2 (vcat ( vars ++ asmps ++ 
+               [ hang "the following constraints hold:"
+                    2 (vcat
+                       $ bullets
+                       $ map (ppWithNames ns1)
+                       $ dctGoals d )]))
     where
     bullets xs = [ "•" <+> x | x <- xs ]
 
@@ -318,15 +405,11 @@ instance PP (WithNames DelayedCt) where
 
     name  = dctSource d
     vars = case dctForall d of
-             [] -> empty
-             xs -> "for any type" <+>
-                      fsep (punctuate comma (map (ppWithNames ns1 ) xs))
+             [] -> []
+             xs -> ["for any type" <+> commaSep (map (ppWithNames ns1) xs)]
     asmps = case dctAsmps d of
-              [] -> empty
-              xs -> "assuming" $$
-                    nest 2 (vcat (bullets (map (ppWithNames ns1) xs)))
+              [] -> []
+              xs -> [hang "assuming"
+                       2 (vcat (bullets (map (ppWithNames ns1) xs)))]
 
     ns1 = addTNames (dctForall d) names
-
-
-
