@@ -7,7 +7,9 @@
 -- Portability :  portable
 
 {-# LANGUAGE RecursiveDo #-}
-{-# LANGUAGE Safe #-}
+{-# LANGUAGE BlockArguments #-}
+-- See Note [-Wincomplete-uni-patterns and irrefutable patterns] in Cryptol.TypeCheck.TypePat
+{-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
 module Cryptol.TypeCheck.Kind
   ( checkType
   , checkSchema
@@ -17,10 +19,10 @@ module Cryptol.TypeCheck.Kind
   , checkPropSyn
   , checkParameterType
   , checkParameterConstraints
+  , checkPropGuards
   ) where
 
 import qualified Cryptol.Parser.AST as P
-import           Cryptol.Parser.AST (Named(..))
 import           Cryptol.Parser.Position
 import           Cryptol.TypeCheck.AST
 import           Cryptol.TypeCheck.Error
@@ -30,14 +32,14 @@ import           Cryptol.TypeCheck.SimpleSolver(simplify)
 import           Cryptol.TypeCheck.Solve (simplifyAllConstraints)
 import           Cryptol.TypeCheck.Subst(listSubst,apSubst)
 import           Cryptol.Utils.Panic (panic)
+import           Cryptol.Utils.RecordMap
 
 import qualified Data.Map as Map
 import           Data.List(sortBy,groupBy)
 import           Data.Maybe(fromMaybe)
 import           Data.Function(on)
-import           Control.Monad(unless,forM,when)
-
-
+import           Data.Text (Text)
+import           Control.Monad(unless,when,mplus)
 
 -- | Check a type signature.  Returns validated schema, and any implicit
 -- constraints that we inferred.
@@ -52,7 +54,7 @@ checkSchema withWild (P.Forall xs ps t mb) =
      -- XXX: We probably shouldn't do this, as we are changing what the
      -- user is doing.  We do it so that things are in a propal normal form,
      -- but we should probably figure out another time to do this.
-     let newPs = concatMap pSplitAnd $ map (simplify Map.empty)
+     let newPs = concatMap pSplitAnd $ map (simplify mempty)
                                      $ map tRebuild ps1
      return ( Forall xs1 newPs (tRebuild t1)
             , [ g { goal = tRebuild (goal g) } | g <- gs ]
@@ -63,18 +65,42 @@ checkSchema withWild (P.Forall xs ps t mb) =
           Nothing -> id
           Just r  -> inRange r
 
+{- | Validate parsed propositions that appear in the guard of a PropGuard.
+
+  * Note that we don't validate the well-formedness constraints here---instead,
+    they'd be validated when the signature for the auto generated
+    function corresponding guard is checked.
+
+  * We also check that there are no wild-cards in the constraints.
+-}
+checkPropGuards :: [Located (P.Prop Name)] -> InferM [Prop]
+checkPropGuards props =
+  do (newPs,_gs) <- collectGoals (mapM check props)
+     pure newPs
+  where
+  check lp =
+    inRange (srcRange lp)
+    do let p = thing lp
+       (_,ps) <- withTParams NoWildCards schemaParam [] (checkProp p)
+       case tNoUser ps of
+         TCon (PC x) _ | x `elem` [PEqual,PNeq,PGeq,PFin,PTrue] -> pure ()
+         _ -> recordError (InvalidConstraintGuard ps)
+       pure ps
+
+
+
 -- | Check a module parameter declarations.  Nothing much to check,
 -- we just translate from one syntax to another.
-checkParameterType :: P.ParameterType Name -> Maybe String -> InferM ModTParam
-checkParameterType a mbDoc =
-  do let k = cvtK (P.ptKind a)
+checkParameterType :: P.ParameterType Name -> InferM ModTParam
+checkParameterType a =
+  do let mbDoc = P.ptDoc a
+         k = cvtK (P.ptKind a)
          n = thing (P.ptName a)
-     return ModTParam { mtpKind = k, mtpName = n, mtpDoc = mbDoc
-                      , mtpNumber = P.ptNumber a }
+     return ModTParam { mtpKind = k, mtpName = n, mtpDoc = mbDoc }
 
 
 -- | Check a type-synonym declaration.
-checkTySyn :: P.TySyn Name -> Maybe String -> InferM TySyn
+checkTySyn :: P.TySyn Name -> Maybe Text -> InferM TySyn
 checkTySyn (P.TySyn x _ as t) mbD =
   do ((as1,t1),gs) <- collectGoals
                     $ inRange (srcRange x)
@@ -90,7 +116,7 @@ checkTySyn (P.TySyn x _ as t) mbD =
                   }
 
 -- | Check a constraint-synonym declaration.
-checkPropSyn :: P.PropSyn Name -> Maybe String -> InferM TySyn
+checkPropSyn :: P.PropSyn Name -> Maybe Text -> InferM TySyn
 checkPropSyn (P.PropSyn x _ as ps) mbD =
   do ((as1,t1),gs) <- collectGoals
                     $ inRange (srcRange x)
@@ -107,30 +133,28 @@ checkPropSyn (P.PropSyn x _ as ps) mbD =
 
 -- | Check a newtype declaration.
 -- XXX: Do something with constraints.
-checkNewtype :: P.Newtype Name -> Maybe String -> InferM Newtype
-checkNewtype (P.Newtype x as fs) mbD =
+checkNewtype :: P.Newtype Name -> Maybe Text -> InferM Newtype
+checkNewtype (P.Newtype x as con fs) mbD =
   do ((as1,fs1),gs) <- collectGoals $
        inRange (srcRange x) $
        do r <- withTParams NoWildCards newtypeParam as $
-               forM fs $ \field ->
-                 let n = name field
-                 in kInRange (srcRange n) $
-                    do t1 <- doCheckType (value field) (Just KType)
-                       return (thing n, t1)
+               flip traverseRecordMap fs $ \_n (rng,f) ->
+                 kInRange rng $ doCheckType f (Just KType)
           simplifyAllConstraints
           return r
 
      return Newtype { ntName   = thing x
                     , ntParams = as1
                     , ntConstraints = map goal gs
+                    , ntConName = con
                     , ntFields = fs1
                     , ntDoc = mbD
                     }
 
-checkPrimType :: P.PrimType Name -> Maybe String -> InferM AbstractType
+checkPrimType :: P.PrimType Name -> Maybe Text -> InferM AbstractType
 checkPrimType p mbD =
   do let (as,cs) = P.primTCts p
-     (as',cs') <- withTParams NoWildCards (TPOther . Just) as $
+     (as',cs') <- withTParams NoWildCards TPPrimParam as $
                     mapM checkProp cs
      pure AbstractType { atName = thing (P.primTName p)
                        , atKind = cvtK (thing (P.primTKind p))
@@ -263,10 +287,13 @@ checkTUser x ts k =
        checkKind (TUser x ts1 t1) k k1
 
   checkNewTypeUse nt =
-    do let tc = newtypeTyCon nt
-       (ts1,_) <- appTy ts (kindOf tc)
-       ts2 <- checkParams (ntParams nt) ts1
-       return (TCon tc ts2)
+    do (ts1,k1) <- appTy ts (kindOf nt)
+       let as = ntParams nt
+       ts2 <- checkParams as ts1
+       let su = zip as ts2
+       ps1 <- mapM (`kInstantiateT` su) (ntConstraints nt)
+       kNewGoals (CtPartialTypeFun (ntName nt)) ps1
+       checkKind (TNewtype nt ts2) k k1
 
   checkAbstractTypeUse absT =
     do let tc   = abstractTypeTC absT
@@ -301,7 +328,8 @@ checkTUser x ts k =
     do let ty = tpVar (mtpParam a)
        (ts1,k1) <- appTy ts (kindOf ty)
        case k of
-         Just ks | ks /= k1 -> kRecordError $ KindMismatch ks k1
+         Just ks
+           | ks /= k1 -> kRecordError (KindMismatch Nothing ks k1)
          _ -> return ()
 
        unless (null ts1) $
@@ -376,26 +404,33 @@ doCheckType ty k =
 
     P.TTuple ts     -> tcon (TC (TCTuple (length ts))) ts k
 
-    P.TRecord fs    -> do t1 <- TRec `fmap` mapM checkF fs
+    P.TRecord fs    -> do t1 <- TRec <$> traverseRecordMap checkF fs
                           checkKind t1 k KType
     P.TLocated t r1 -> kInRange r1 $ doCheckType t k
 
     P.TUser x ts    -> checkTUser x ts k
 
-    P.TParens t     -> doCheckType t k
+    P.TParens t mb  ->
+      do newK <- case (k, cvtK <$> mb) of
+                   (Just a, Just b) ->
+                      do unless (a == b)
+                           (kRecordError (KindMismatch Nothing a b))
+                         pure (Just b)
+                   (a,b) -> pure (mplus a b)
+
+         doCheckType t newK
 
     P.TInfix t x _ u-> doCheckType (P.TUser (thing x) [t, u]) k
 
+    P.TTyApp _fs    -> do kRecordError BareTypeApp
+                          kNewType TypeWildCard KNum
+
   where
-  checkF f = do t <- kInRange (srcRange (name f))
-                   $ doCheckType (value f) (Just KType)
-                return (thing (name f), t)
-
-
+  checkF _nm (rng,v) = kInRange rng $ doCheckType v (Just KType)
 
 -- | Validate a parsed proposition.
 checkProp :: P.Prop Name      -- ^ Proposition that need to be checked
-          -> KindM Type       -- ^ Checked representation
+          -> KindM Prop -- ^ Checked representation
 checkProp (P.CType t) = doCheckType t (Just KProp)
 
 
@@ -405,8 +440,6 @@ checkKind :: Type         -- ^ Kind-checked type
           -> Kind         -- ^ Inferred kind
           -> KindM Type   -- ^ A type consistent with expectations.
 checkKind _ (Just k1) k2
-  | k1 /= k2    = do kRecordError (KindMismatch k1 k2)
+  | k1 /= k2    = do kRecordError (KindMismatch Nothing k1 k2)
                      kNewType TypeErrorPlaceHolder k1
 checkKind t _ _ = return t
-
-
