@@ -17,24 +17,38 @@ module Cryptol.TypeCheck.Parseable
   , ShowParseable(..)
   ) where
 
+import Data.Void
+import Prettyprinter
+
 import Cryptol.TypeCheck.AST
 import Cryptol.Utils.Ident (Ident,unpackIdent)
+import Cryptol.Utils.RecordMap (canonicalFields)
 import Cryptol.Parser.AST ( Located(..))
 import Cryptol.ModuleSystem.Name
-import Text.PrettyPrint hiding ((<>))
-import qualified Text.PrettyPrint as PP ((<>))
+
+
+infixl 5 $$
+($$) :: Doc a -> Doc a -> Doc a
+($$) x y = sep [x, y]
+
+text :: String -> Doc a
+text = pretty
+
+int :: Int -> Doc a
+int = pretty
 
 -- ShowParseable prints out a cryptol program in a way that it's parseable by Coq (and likely other things)
 -- Used mainly for reasoning about the semantics of cryptol programs in Coq (https://github.com/GaloisInc/cryptol-semantics)
 class ShowParseable t where
-  showParseable :: t -> Doc
+  showParseable :: t -> Doc Void
 
 instance ShowParseable Expr where
+  showParseable (ELocated _ e) = showParseable e -- TODO? emit range information
   showParseable (EList es _) = parens (text "EList" <+> showParseable es)
   showParseable (ETuple es) = parens (text "ETuple" <+> showParseable es)
-  showParseable (ERec ides) = parens (text "ERec" <+> showParseable ides)
+  showParseable (ERec ides) = parens (text "ERec" <+> showParseable (canonicalFields ides))
   showParseable (ESel e s) = parens (text "ESel" <+> showParseable e <+> showParseable s)
-  showParseable (ESet e s v) = parens (text "ESet" <+>
+  showParseable (ESet _ty e s v) = parens (text "ESet" <+>
                                 showParseable e <+> showParseable s
                                                 <+> showParseable v)
   showParseable (EIf c t f) = parens (text "EIf" <+> showParseable c $$ showParseable t $$ showParseable f)
@@ -49,9 +63,11 @@ instance ShowParseable Expr where
   --NOTE: erase all "proofs" for now (change the following two lines to change that)
   showParseable (EProofAbs {-p-}_ e) = showParseable e --"(EProofAbs " ++ show p ++ showParseable e ++ ")"
   showParseable (EProofApp e) = showParseable e --"(EProofApp " ++ showParseable e ++ ")"
+  
+  showParseable (EPropGuards guards _) = parens (text "EPropGuards" $$ showParseable guards)
 
 instance (ShowParseable a, ShowParseable b) => ShowParseable (a,b) where
-  showParseable (x,y) = parens (showParseable x PP.<> comma PP.<> showParseable y)
+  showParseable (x,y) = parens (showParseable x <> comma <> showParseable y)
 
 instance ShowParseable Int where
   showParseable i = int i
@@ -61,7 +77,7 @@ instance ShowParseable Ident where
 
 instance ShowParseable Type where
   showParseable (TUser n lt t) = parens (text "TUser" <+> showParseable n <+> showParseable lt <+> showParseable t)
-  showParseable (TRec lidt) = parens (text "TRec" <+> showParseable lidt)
+  showParseable (TRec lidt) = parens (text "TRec" <+> showParseable (canonicalFields lidt))
   showParseable t = parens $ text $ show t
 
 instance ShowParseable Selector where
@@ -79,6 +95,7 @@ instance ShowParseable Decl where
 
 instance ShowParseable DeclDef where
   showParseable DPrim = text (show DPrim)
+  showParseable (DForeign t) = text (show $ DForeign t)
   showParseable (DExpr e) = parens (text "DExpr" $$ showParseable e)
 
 instance ShowParseable DeclGroup where
@@ -103,11 +120,11 @@ instance (ShowParseable a) => ShowParseable (Located a) where
   showParseable l = showParseable (thing l)
 
 instance ShowParseable TParam where
-  showParseable tp = parens (text (show (tpUnique tp)) PP.<> comma PP.<> maybeNameDoc (tpName tp))
+  showParseable tp = parens (text (show (tpUnique tp)) <> comma <> maybeNameDoc (tpName tp))
 
-maybeNameDoc :: Maybe Name -> Doc
-maybeNameDoc Nothing = doubleQuotes empty
+maybeNameDoc :: Maybe Name -> Doc Void
+maybeNameDoc Nothing = dquotes mempty
 maybeNameDoc (Just n) = showParseable (nameIdent n)
 
 instance ShowParseable Name where
-  showParseable n = parens (text (show (nameUnique n)) PP.<> comma PP.<> showParseable (nameIdent n))
+  showParseable n = parens (text (show (nameUnique n)) <> comma <> showParseable (nameIdent n))
