@@ -5,25 +5,30 @@
 -- Maintainer  :  cryptol@galois.com
 -- Stability   :  provisional
 -- Portability :  portable
+{-# Language OverloadedStrings #-}
 module Cryptol.TypeCheck.Sanity
   ( tcExpr
   , tcDecls
   , tcModule
   , ProofObligation
+  , onlyNonTrivial
   , Error(..)
+  , AreSame(..)
   , same
   ) where
 
 
-import Cryptol.Parser.Position(thing)
+import Cryptol.Parser.Position(thing,Range,emptyRange)
 import Cryptol.TypeCheck.AST
-import Cryptol.TypeCheck.Subst (apSubst, singleSubst)
+import Cryptol.TypeCheck.Subst (apSubst, singleTParamSubst)
 import Cryptol.TypeCheck.Monad(InferInput(..))
+import Cryptol.ModuleSystem.Name(nameLoc)
 import Cryptol.Utils.Ident
+import Cryptol.Utils.RecordMap
+import Cryptol.Utils.PP
 
+import Data.List (sort)
 import qualified Data.Set as Set
-import Data.List (sort, sortBy)
-import Data.Function (on)
 import MonadLib
 import qualified Control.Applicative as A
 
@@ -31,15 +36,15 @@ import           Data.Map ( Map )
 import qualified Data.Map as Map
 
 
-tcExpr :: InferInput -> Expr -> Either Error (Schema, [ ProofObligation ])
+tcExpr :: InferInput -> Expr -> Either (Range, Error) (Schema, [ ProofObligation ])
 tcExpr env e = runTcM env (exprSchema e)
 
-tcDecls :: InferInput -> [DeclGroup] -> Either Error [ ProofObligation ]
+tcDecls :: InferInput -> [DeclGroup] -> Either (Range, Error) [ ProofObligation ]
 tcDecls env ds0 = case runTcM env (checkDecls ds0) of
                     Left err     -> Left err
                     Right (_,ps) -> Right ps
 
-tcModule :: InferInput -> Module -> Either Error [ ProofObligation ]
+tcModule :: InferInput -> Module -> Either (Range, Error) [ ProofObligation ]
 tcModule env m = case runTcM env check of
                    Left err -> Left err
                    Right (_,ps) -> Right ps
@@ -48,6 +53,20 @@ tcModule env m = case runTcM env check of
         k2    = withVars (Map.toList (fmap mvpType (mParamFuns m)))
               $ checkDecls (mDecls m)
 
+onlyNonTrivial :: [ProofObligation] -> [ProofObligation]
+onlyNonTrivial = filter (not . trivialProofObligation)
+
+-- | Identify proof obligations that are obviously true.
+-- We can filter these to avoid clutter
+trivialProofObligation :: ProofObligation -> Bool
+trivialProofObligation oblig = pIsTrue goal || simpleEq || goal `elem` asmps
+  where
+  goal  = sType oblig
+  asmps = sProps oblig
+  simpleEq = case pIsEqual goal of
+               Just (t1,t2) -> t1 == t2
+               Nothing      -> False
+
 
 --------------------------------------------------------------------------------
 
@@ -68,10 +87,14 @@ checkType ty =
       do ks <- mapM checkType ts
          checkKind (kindOf tc) ks
 
+    TNewtype nt ts ->
+      do ks <- mapM checkType ts
+         checkKind (kindOf nt) ks
+
     TVar tv -> lookupTVar tv
 
     TRec fs ->
-      do forM_ fs $ \(_,t) ->
+      do forM_ fs $ \t ->
            do k <- checkType t
               unless (k == KType) $ reportError $ KindMismatch KType k
          return KType
@@ -103,23 +126,68 @@ checkSchema (Forall as ps t) = foldr withTVar check as
   where check = do mapM_ (checkTypeIs KProp) ps
                    checkTypeIs KType t
 
+data AreSame = SameIf [Prop]
+             | NotSame
+
+areSame :: AreSame
+areSame = SameIf []
+
+sameAnd :: AreSame -> AreSame -> AreSame
+sameAnd x y =
+  case (x,y) of
+    (SameIf xs, SameIf ys) -> SameIf (xs ++ ys)
+    _                      -> NotSame
+
+sameBool :: Bool -> AreSame
+sameBool b = if b then areSame else NotSame
+
+sameTypes :: String -> Type -> Type -> TcM ()
+sameTypes msg x y = sameSchemas msg (tMono x) (tMono y)
+
+sameSchemas :: String -> Schema -> Schema -> TcM ()
+sameSchemas msg x y =
+  case same x y of
+    NotSame   -> reportError (TypeMismatch msg x y)
+    SameIf ps -> mapM_ proofObligation ps
+
+
+
 
 class Same a where
-  same :: a -> a -> Bool
+  same :: a -> a -> AreSame
 
 instance Same a => Same [a] where
-  same [] [] = True
-  same (x : xs) (y : ys) = same x y && same xs ys
-  same _ _ = False
+  same [] [] = areSame
+  same (x : xs) (y : ys) = same x y `sameAnd` same xs ys
+  same _ _ = NotSame
+
+data Field a b = Field a b
+
+instance (Eq a, Same b) => Same (Field a b) where
+  same (Field x a) (Field y b) = sameBool (x == y) `sameAnd` same a b
 
 instance Same Type where
-  same t1 t2 = tNoUser t1 == tNoUser t2
+  same t1 t2
+    | k1 /= k2    = NotSame
+    | k1 == KNum  = if t1 == t2 then SameIf [] else SameIf [ t1 =#= t2 ]
+    | otherwise   =
+      case (tNoUser t1, tNoUser t2) of
+        (TVar x, TVar y)               -> sameBool (x == y)
+        (TRec x, TRec y)               -> same (mkRec x) (mkRec y)
+        (TNewtype x xs, TNewtype y ys) -> same (Field x xs) (Field y ys)
+        (TCon x xs, TCon y ys)         -> same (Field x xs) (Field y ys)
+        _                              -> NotSame
+      where
+      k1 = kindOf t1
+      k2 = kindOf t2
+      mkRec r = [ Field x y | (x,y) <- canonicalFields r ]
 
 instance Same Schema where
-  same (Forall xs ps s) (Forall ys qs t) = same xs ys && same ps qs && same s t
+  same (Forall xs ps s) (Forall ys qs t) =
+    same xs ys `sameAnd` same ps qs `sameAnd` same s t
 
 instance Same TParam where
-  same x y = tpName x == tpName y && tpKind x == tpKind y
+  same x y = sameBool (tpName x == tpName y && tpKind x == tpKind y)
 
 
 
@@ -143,12 +211,13 @@ exprSchema :: Expr -> TcM Schema
 exprSchema expr =
   case expr of
 
+    ELocated rng t -> withRange rng (exprSchema t)
+
     EList es t ->
       do checkTypeIs KType t
          forM_ es $ \e ->
            do t1 <- exprType e
-              unless (same t1 t) $
-                reportError $ TypeMismatch "EList" (tMono t) (tMono t1)
+              sameTypes "EList" t1 t
 
          return $ tMono $ tSeq (tNum (length es)) t
 
@@ -156,17 +225,15 @@ exprSchema expr =
       fmap (tMono . tTuple) (mapM exprType es)
 
     ERec fs ->
-      do fs1 <- forM fs $ \(f,e) -> do t <- exprType e
-                                       return (f,t)
+      do fs1 <- traverse exprType fs
          return $ tMono $ TRec fs1
 
-    ESet e x v -> do ty  <- exprType e
-                     expe <- checkHas ty x
-                     has <- exprType v
-                     unless (same expe has) $
-                        reportError $
-                          TypeMismatch "ESet" (tMono expe) (tMono has)
-                     return (tMono ty)
+    ESet _ e x v ->
+       do ty  <- exprType e
+          expe <- checkHas ty x
+          has <- exprType v
+          sameTypes "ESet" expe has
+          return (tMono ty)
 
     ESel e sel -> do ty <- exprType e
                      ty1 <- checkHas ty sel
@@ -174,13 +241,11 @@ exprSchema expr =
 
     EIf e1 e2 e3 ->
       do ty <- exprType e1
-         unless (same tBit ty) $
-           reportError $ TypeMismatch "EIf_condition" (tMono tBit) (tMono ty)
+         sameTypes "EIf_condition" tBit ty
 
          t1 <- exprType e2
          t2 <- exprType e3
-         unless (same t1 t2) $
-           reportError $ TypeMismatch "EIf_arms" (tMono t1) (tMono t2)
+         sameTypes "EIf_arms" t1 t2
 
          return $ tMono t1
 
@@ -221,7 +286,7 @@ exprSchema expr =
                 let k' = kindOf a
                 unless (k == k') $ reportError $ KindMismatch k' k
 
-                let su = singleSubst (tpVar a) t
+                let su = singleTParamSubst a t
                 return $ Forall as (apSubst su ps) (apSubst su t1)
 
            Forall [] _ _ -> reportError BadInstantiation
@@ -232,7 +297,9 @@ exprSchema expr =
 
          case tNoUser t1 of
            TCon (TC TCFun) [ a, b ]
-              | same a t2 -> return (tMono b)
+              | SameIf ps <- same a t2 ->
+                do mapM_ proofObligation ps
+                   return (tMono b)
            tf -> reportError (BadApplication tf t1)
 
 
@@ -264,6 +331,9 @@ exprSchema expr =
       in go dgs
 
 
+    EPropGuards _guards typ -> 
+      pure Forall {sVars = [], sProps = [], sType = typ}
+
 checkHas :: Type -> Selector -> TcM Type
 checkHas t sel =
   case sel of
@@ -297,13 +367,13 @@ checkHas t sel =
           do case mb of
                Nothing -> return ()
                Just fs1 ->
-                 do let ns  = sort (map fst fs)
+                 do let ns  = Set.toList (fieldSet fs)
                         ns1 = sort fs1
                     unless (ns == ns1) $
                       reportError $ UnexpectedRecordShape ns1 ns
 
-             case lookup f fs of
-               Nothing -> reportError $ MissingField f $ map fst fs
+             case lookupField f fs of
+               Nothing -> reportError $ MissingField f $ displayOrder fs
                Just ft -> return ft
 
         TCon (TC TCSeq) [s,elT] -> do res <- checkHas elT sel
@@ -367,14 +437,17 @@ convertible t1 t2 = go t1 t2
                                | tc1 == tc2 -> goMany ts1 ts2
                             _ -> err
 
+         TNewtype nt1 ts1 ->
+            case other of
+              TNewtype nt2 ts2
+                | nt1 == nt2 -> goMany ts1 ts2
+              _ -> err
+
          TRec fs ->
            case other of
              TRec gs ->
-               do let order = sortBy (compare `on` fst)
-                      fs1   = order fs
-                      gs1   = order gs
-                  unless (map fst fs1 == map fst gs1) err
-                  goMany (map snd fs1) (map snd gs1)
+               do unless (fieldSet fs == fieldSet gs) err
+                  goMany (recordElements fs) (recordElements gs)
              _ -> err
 
 
@@ -389,12 +462,19 @@ checkDecl checkSig d =
       do when checkSig $ checkSchema $ dSignature d
          return (dName d, dSignature d)
 
+    DForeign _ ->
+      do when checkSig $ checkSchema $ dSignature d
+         return (dName d, dSignature d)
+
     DExpr e ->
       do let s = dSignature d
          when checkSig $ checkSchema s
          s1 <- exprSchema e
-         unless (same s s1) $
-           reportError $ TypeMismatch "DExpr" s s1
+         let nm = dName d
+             loc = "definition of " ++ show (pp nm) ++
+                            ", at " ++ show (pp (nameLoc nm))
+         sameSchemas loc s s1
+
          return (dName d, s)
 
 checkDeclGroup :: DeclGroup -> TcM [(Name, Schema)]
@@ -418,7 +498,9 @@ checkMatch ma =
          t1 <- exprType e
          case tNoUser t1 of
            TCon (TC TCSeq) [ l, el ]
-             | same elt el -> return ((x, tMono elt), l)
+             | SameIf ps <- same elt el ->
+               do mapM_ proofObligation ps
+                  return ((x, tMono elt), l)
              | otherwise -> reportError $ TypeMismatch "From" (tMono elt) (tMono el)
 
 
@@ -447,6 +529,7 @@ checkArm (m : ms) =
 data RO = RO
   { roTVars   :: Map Int TParam
   , roAsmps   :: [Prop]
+  , roRange   :: Range
   , roVars    :: Map Name Schema
   }
 
@@ -456,34 +539,36 @@ data RW = RW
   { woProofObligations :: [ProofObligation]
   }
 
-newtype TcM a = TcM (ReaderT RO (ExceptionT Error (StateT RW Id)) a)
+newtype TcM a = TcM (ReaderT RO (ExceptionT (Range, Error) (StateT RW Id)) a)
 
 instance Functor TcM where
   fmap = liftM
 
 instance A.Applicative TcM where
-  pure  = return
+  pure a = TcM (pure a)
   (<*>) = ap
 
 instance Monad TcM where
-  return a    = TcM (return a)
-  fail x      = TcM (fail x)
+  return      = pure
   TcM m >>= f = TcM (do a <- m
                         let TcM m1 = f a
                         m1)
 
-runTcM :: InferInput -> TcM a -> Either Error (a, [ProofObligation])
+runTcM :: InferInput -> TcM a -> Either (Range, Error) (a, [ProofObligation])
 runTcM env (TcM m) =
   case runM m ro rw of
     (Left err, _) -> Left err
     (Right a, s)  -> Right (a, woProofObligations s)
   where
+  allPs = inpParams env
+
   ro = RO { roTVars = Map.fromList [ (tpUnique x, x)
-                                      | tp <- Map.elems (inpParamTypes env)
+                                      | tp <- Map.elems (mpnTypes allPs)
                                       , let x = mtpParam tp ]
-          , roAsmps = map thing (inpParamConstraints env)
+          , roAsmps = map thing (mpnConstraints allPs)
+          , roRange = emptyRange
           , roVars  = Map.union
-                        (fmap mvpType (inpParamFuns env))
+                        (fmap mvpType (mpnFuns allPs))
                         (inpVars env)
           }
   rw = RW { woProofObligations = [] }
@@ -515,13 +600,20 @@ data Error =
     deriving Show
 
 reportError :: Error -> TcM a
-reportError e = TcM (raise e)
+reportError e = TcM $
+  do ro <- ask
+     raise (roRange ro, e)
 
 withTVar :: TParam -> TcM a -> TcM a
 withTVar a (TcM m) = TcM $
   do ro <- ask
      local ro { roTVars = Map.insert (tpUnique a) a (roTVars ro) } m
 
+withRange :: Range -> TcM a -> TcM a
+withRange rng (TcM m) = TcM $
+  do ro <- ask
+     local ro { roRange = rng } m
+
 withAsmp :: Prop -> TcM a -> TcM a
 withAsmp p (TcM m) = TcM $
   do ro <- ask
@@ -562,3 +654,114 @@ lookupVar x =
      case Map.lookup x (roVars ro) of
        Just s -> return s
        Nothing -> reportError $ UndefinedVariable x
+
+
+instance PP Error where
+  ppPrec _ err =
+    case err of
+
+      TypeMismatch what expected actual ->
+        ppErr ("Type mismatch in" <+> text what)
+          [ "Expected:" <+> pp expected
+          , "Actual  :" <+> pp actual
+          ]
+
+      ExpectedMono s ->
+        ppErr "Not a monomorphic type"
+          [ pp s ]
+
+      TupleSelectorOutOfRange sel sz ->
+        ppErr "Tuple selector out of range"
+          [ "Selector:" <+> int sel
+          , "Size    :" <+> int sz
+          ]
+
+      MissingField f fs ->
+        ppErr "Invalid record selector"
+          [ "Field: " <+> pp f
+          , "Fields:" <+> commaSep (map pp fs)
+          ]
+
+      UnexpectedTupleShape expected actual ->
+        ppErr "Unexpected tuple shape"
+          [ "Expected:" <+> int expected
+          , "Actual  :" <+> int actual
+          ]
+
+      UnexpectedRecordShape expected actual ->
+        ppErr "Unexpected record shape"
+          [ "Expected:" <+> commaSep (map pp expected)
+          , "Actual  :" <+> commaSep (map pp actual)
+          ]
+
+      UnexpectedSequenceShape n t ->
+        ppErr "Unexpected sequence shape"
+          [ "Expected:" <+> int n
+          , "Actual  :" <+> pp t
+          ]
+
+      BadSelector sel t ->
+        ppErr "Bad selector"
+          [ "Selector:" <+> pp sel
+          , "Type    :" <+> pp t
+          ]
+
+      BadInstantiation ->
+        ppErr "Bad instantiation" []
+
+      Captured x ->
+        ppErr "Captured type variable"
+          [ "Variable:" <+> pp x ]
+
+      BadProofNoAbs ->
+        ppErr "Proof application without a proof abstraction" []
+
+      BadProofTyVars xs ->
+        ppErr "Proof application with type abstraction"
+          [ "Type parameter:" <+> pp x | x <- xs ]
+
+      KindMismatch expected actual ->
+        ppErr "Kind mismatch"
+          [ "Expected:" <+> pp expected
+          , "Actual  :" <+> pp actual
+          ]
+
+      NotEnoughArgumentsInKind k ->
+        ppErr "Not enough arguments in kind" [ pp k ]
+
+      BadApplication t1 t2 ->
+        ppErr "Bad application"
+          [ "Function:" <+> pp t1
+          , "Argument:" <+> pp t2
+          ]
+
+      FreeTypeVariable x ->
+        ppErr "Free type variable"
+          [ "Variable:" <+> pp x ]
+
+      BadTypeApplication kf ka ->
+        ppErr "Bad type application"
+          [ "Function :" <+> pp kf
+          , "Arguments:" <+> commaSep (map pp ka)
+          ]
+
+      RepeatedVariableInForall x ->
+        ppErr "Repeated variable in forall"
+          [ "Variable:" <+> pp x ]
+
+      BadMatch t ->
+        ppErr "Bad match"
+          [ "Type:" <+> pp t ]
+
+      EmptyArm -> ppErr "Empty comprehension arm" []
+
+      UndefinedTypeVaraible x ->
+        ppErr "Undefined type variable"
+          [ "Variable:" <+> pp x ]
+
+      UndefinedVariable x ->
+        ppErr "Undefined variable"
+          [ "Variable:" <+> pp x ]
+
+    where
+    ppErr x ys = hang x 2 (vcat [ "•" <+> y | y <- ys ])
