@@ -7,10 +7,10 @@
 -- Portability :  portable
 
 {-# LANGUAGE PatternGuards, BangPatterns, RecordWildCards #-}
-{-# LANGUAGE Safe #-}
 module Cryptol.TypeCheck.Solve
   ( simplifyAllConstraints
   , proveImplication
+  , tryProveImplication
   , proveModuleTopLevel
   , defaultAndSimplify
   , defaultReplExpr
@@ -30,32 +30,30 @@ import qualified Cryptol.TypeCheck.SimpleSolver as Simplify
 import           Cryptol.TypeCheck.Solver.Types
 import           Cryptol.TypeCheck.Solver.Selector(tryHasGoal)
 
-
 import           Cryptol.TypeCheck.Solver.SMT(Solver,proveImp,isNumeric)
 import           Cryptol.TypeCheck.Solver.Improve(improveProp,improveProps)
 import           Cryptol.TypeCheck.Solver.Numeric.Interval
-import           Cryptol.Utils.PP (text,vcat,(<+>))
 import           Cryptol.Utils.Patterns(matchMaybe)
 
 import           Control.Applicative ((<|>))
 import           Control.Monad(mzero)
+import           Data.Containers.ListUtils (nubOrd)
+import           Data.Map (Map)
 import qualified Data.Map as Map
 import           Data.Set ( Set )
 import qualified Data.Set as Set
 import           Data.List(partition)
-import           Data.Maybe(listToMaybe)
+import           Data.Maybe(listToMaybe,fromMaybe)
 
 
 
 
 
-quickSolverIO :: Ctxt -> [Goal] -> IO (Either Goal (Subst,[Goal]))
+quickSolverIO :: Ctxt -> [Goal] -> IO (Either Error (Subst,[Goal]))
 quickSolverIO _ [] = return (Right (emptySubst, []))
 quickSolverIO ctxt gs =
   case quickSolver ctxt gs of
-    Left err ->
-      do msg (text "Contradiction:" <+> pp (goal err))
-         return (Left err)
+    Left err -> return (Left err)
     Right (su,gs') ->
       do msg (vcat (map (pp . goal) gs' ++ [pp su]))
          return (Right (su,gs'))
@@ -76,7 +74,7 @@ quickSolverIO ctxt gs =
 
 quickSolver :: Ctxt   -- ^ Facts we can know
             -> [Goal] -- ^ Need to solve these
-            -> Either Goal (Subst,[Goal])
+            -> Either Error (Subst,[Goal])
             -- ^ Left: contradicting goals,
             --   Right: inferred types, unsolved goals.
 quickSolver ctxt gs0 = go emptySubst [] gs0
@@ -84,24 +82,65 @@ quickSolver ctxt gs0 = go emptySubst [] gs0
   go su [] [] = Right (su,[])
 
   go su unsolved [] =
-    case matchMaybe (findImprovement unsolved) of
-      Nothing            -> Right (su,unsolved)
-      Just (newSu, subs) -> go (newSu @@ su) [] (subs ++ apSubst newSu unsolved)
+    case matchMaybe (findImprovement noIncompatible unsolved) of
+      Nothing -> Right (su,unsolved)
+      Just imp ->
+        case imp of
+          Right (newSu, subs) ->
+            go (newSu @@ su) [] (subs ++ apSubst newSu unsolved)
+          Left err -> Left err
+
+  go su unsolved (g : gs)
+    | Set.member (goal g) (saturatedAsmps ctxt) = go su unsolved gs
 
   go su unsolved (g : gs) =
     case Simplify.simplifyStep ctxt (goal g) of
-      Unsolvable _        -> Left g
+      Unsolvable          -> Left (UnsolvableGoals [g])
       Unsolved            -> go su (g : unsolved) gs
       SolvedIf subs       ->
         let cvt x = g { goal = x }
         in go su unsolved (map cvt subs ++ gs)
 
   -- Probably better to find more than one.
-  findImprovement []       = mzero
-  findImprovement (g : gs) =
+  findImprovement inc [] =
+    do let bad = Map.intersectionWith (,) (integralTVars inc) (fracTVars inc)
+       case Map.minView bad of
+         Just ((g1,g2),_) -> pure $ Left $ UnsolvableGoals [g1,g2]
+         Nothing -> mzero
+
+  findImprovement inc (g : gs) =
     do (su,ps) <- improveProp False ctxt (goal g)
-       return (su, [ g { goal = p } | p <- ps ])
-    <|> findImprovement gs
+       return (Right (su, [ g { goal = p } | p <- ps ]))
+    <|>
+    findImprovement (addIncompatible g inc) gs
+
+
+--------------------------------------------------------------------------------
+-- Look for type variable with incompatible constraints
+
+data Incompatible = Incompatible
+  { integralTVars :: Map TVar Goal    -- ^ Integral a
+  , fracTVars     :: Map TVar Goal    -- ^ Field a or FLiteral 
+  }
+
+noIncompatible :: Incompatible
+noIncompatible = Incompatible
+  { integralTVars = Map.empty
+  , fracTVars     = Map.empty
+  }
+
+addIncompatible :: Goal -> Incompatible -> Incompatible
+addIncompatible g i =
+  fromMaybe i $
+  do tv <- tIsVar =<< pIsIntegral (goal g)
+     pure i { integralTVars = Map.insert tv g (integralTVars i) }
+  <|>
+  do tv <- tIsVar =<< pIsField (goal g)
+     pure i { fracTVars = Map.insert tv g (fracTVars i) }
+  <|>
+  do (_,_,_,t) <- pIsFLiteral (goal g)
+     tv        <- tIsVar t
+     pure i { fracTVars = Map.insert tv g (fracTVars i) }
 
 
 
@@ -118,7 +157,7 @@ defaultReplExpr sol expr sch =
      case mb of
        Nothing -> return Nothing
        Just numBinds -> return $
-         do optss <- mapM tryDefVar otherVs
+         do let optss = map tryDefVar otherVs
             su    <- listToMaybe
                        [ binds | nonSu <- sequence optss
                                , let binds = nonSu ++ numBinds
@@ -142,28 +181,40 @@ defaultReplExpr sol expr sch =
                   , goalRange = emptyRange
                   , goalSource = CtDefaulting } | p <- otherPs ]
 
-  tryDefVar a =
-    do let a' = TVBound a
-       gt <- Map.lookup a' (literalGoals gSet)
-       let ok p = not (Set.member a' (fvs p))
-       return [ (a,t) | t <- [ tInteger, tBit, tWord (tWidth (goal gt)) ]
-                      , ok t ]
+  fLitGoals = flitDefaultCandidates gSet
 
+  tryDefVar :: TParam -> [(TParam, Type)]
+  tryDefVar a
+    -- REPL defaulting for floating-point literals
+    | Just m <- Map.lookup (TVBound a) fLitGoals
+    = case m of
+        Just ((_,t),_) -> [(a,t)]
+        Nothing        -> []
+
+    -- REPL defaulting for integer literals
+    | Just gt <- Map.lookup (TVBound a) (literalGoals gSet)
+    = let ok p = not (Set.member (TVBound a) (fvs p)) in
+      [ (a,t) | t <- [ tInteger, tWord (tWidth (goal gt)) ]
+              , ok t ]
+
+    -- REPL defaulting for variables unconstrained by a literal constraint
+    | otherwise = [ (a,t) | t <- [tInteger, tRational, tBit] ]
 
   appExpr tys = foldl (\e1 _ -> EProofApp e1)
                       (foldl ETApp expr tys)
                       (sProps sch)
 
 
-defaultAndSimplify :: [TVar] -> [Goal] -> ([TVar],[Goal],Subst,[Warning])
+defaultAndSimplify :: [TVar] -> [Goal] -> ([TVar],[Goal],Subst,[Warning],[Error])
 defaultAndSimplify as gs =
-  let (as1, gs1, su1, ws1) = defLit
-      (as2, gs2, su2, ws2) = improveByDefaultingWithPure as1 gs1
-  in (as2,gs2,su2 @@ su1, ws1 ++ ws2)
+  let (as1, gs1, su1, ws) = defLit
+      (as2, gs2, su2, errs) = improveByDefaultingWithPure as1 gs1
+  in (as2,gs2,su2 @@ su1, ws, errs)
+
   where
   defLit
     | isEmptySubst su = nope
-    | otherwise       = case quickSolver Map.empty (apSubst su gs) of
+    | otherwise       = case quickSolver mempty (apSubst su gs) of
                           Left _ -> nope -- hm?
                           Right (su1,gs1) -> (as1,gs1,su1@@su,ws)
     where (as1,su,ws) = defaultLiterals as gs
@@ -178,8 +229,8 @@ simplifyAllConstraints =
      case gs of
        [] -> return ()
        _ ->
-        case quickSolver Map.empty gs of
-          Left badG      -> recordError (UnsolvedGoals True [badG])
+        case quickSolver mempty gs of
+          Left err -> recordError err
           Right (su,gs1) ->
             do extendSubst su
                addGoals gs1
@@ -207,48 +258,70 @@ proveModuleTopLevel =
   do simplifyAllConstraints
      gs <- getGoals
      let vs = Set.toList (Set.filter isFreeTV (fvs gs))
-         (_,gs1,su1,ws) = defaultAndSimplify vs gs
+         (_,gs1,su1,ws,errs) = defaultAndSimplify vs gs
      extendSubst su1
      mapM_ recordWarning ws
+     mapM_ recordError errs
 
      cs <- getParamConstraints
      case cs of
        [] -> addGoals gs1
-       _  -> do su2 <- proveImplication Nothing [] [] gs1
+       _  -> do su2 <- proveImplication False Nothing [] [] gs1
                 extendSubst su2
 
 -- | Prove an implication, and return any improvements that we computed.
 -- Records errors, if any of the goals couldn't be solved.
-proveImplication :: Maybe Name -> [TParam] -> [Prop] -> [Goal] -> InferM Subst
-proveImplication lnam as ps gs =
+proveImplication :: Bool -> Maybe Name ->
+  [TParam] -> [Prop] -> [Goal] -> InferM Subst
+proveImplication dedupErrs lnam as ps gs =
   do evars <- varsWithAsmps
      solver <- getSolver
 
      extraAs <- (map mtpParam . Map.elems) <$> getParamTypes
      extra   <- map thing <$> getParamConstraints
 
-     (mbErr,su) <- io (proveImplicationIO solver lnam evars
+     (mbErr,su) <- io (proveImplicationIO solver dedupErrs lnam evars
                             (extraAs ++ as) (extra ++ ps) gs)
      case mbErr of
-       Right ws -> mapM_ recordWarning ws
-       Left err -> recordError err
+       Right ws  -> mapM_ recordWarning ws
+       Left errs -> mapM_ recordError errs
      return su
 
+-- | Tries to prove an implication. If proved, then returns `Right (m_su ::
+-- InferM Subst)` where `m_su` is an `InferM` computation that results in the
+-- solution substitution, and records any warning invoked during proving. If not
+-- proved, then returns `Left (m_err :: InferM ())`, which records all errors
+-- invoked during proving.
+tryProveImplication :: 
+  Maybe Name -> [TParam] -> [Prop] -> [Goal] -> InferM Bool
+tryProveImplication lnam as ps gs =
+  do evars <- varsWithAsmps
+     solver <- getSolver
+
+     extraAs <- (map mtpParam . Map.elems) <$> getParamTypes
+     extra   <- map thing <$> getParamConstraints
+
+     (mbErr,_su) <- io (proveImplicationIO solver False lnam evars
+                            (extraAs ++ as) (extra ++ ps) gs)
+     case mbErr of
+       Left {}  -> pure False
+       Right {} -> pure True
 
 proveImplicationIO :: Solver
+                   -> Bool     -- ^ Whether to remove duplicate goals in errors
                    -> Maybe Name     -- ^ Checking this function
                    -> Set TVar -- ^ These appear in the env., and we should
                                -- not try to default them
                    -> [TParam] -- ^ Type parameters
                    -> [Prop]   -- ^ Assumed constraint
                    -> [Goal]   -- ^ Collected constraints
-                   -> IO (Either Error [Warning], Subst)
-proveImplicationIO _   _     _         _  [] [] = return (Right [], emptySubst)
-proveImplicationIO s f varsInEnv ps asmps0 gs0 =
-  do let ctxt = assumptionIntervals Map.empty asmps
+                   -> IO (Either [Error] [Warning], Subst)
+proveImplicationIO _ _         _ _         _  []     []  = return (Right [], emptySubst)
+proveImplicationIO s dedupErrs f varsInEnv ps asmps0 gs0 =
+  do let ctxt = buildSolverCtxt asmps
      res <- quickSolverIO ctxt gs
      case res of
-       Left bad -> return (Left (UnsolvedGoals True [bad]), emptySubst)
+       Left erro -> return (Left [erro], emptySubst)
        Right (su,[]) -> return (Right [], su)
        Right (su,gs1) ->
          do gs2 <- proveImp s asmps gs1
@@ -259,32 +332,34 @@ proveImplicationIO s f varsInEnv ps asmps0 gs0 =
                             $ Set.toList
                             $ Set.difference (fvs (map goal gs3)) varsInEnv
                    case defaultAndSimplify free gs3 of
-                     (_,_,newSu,_)
+                     (_,_,newSu,_,errs)
                         | isEmptySubst newSu ->
-                                 return (err gs3, su) -- XXX: Old?
-                     (_,newGs,newSu,ws) ->
+                                 return (Left (err gs3:errs), su) -- XXX: Old?
+                     (_,newGs,newSu,ws,errs) ->
                        do let su1 = newSu @@ su
-                          (res1,su2) <- proveImplicationIO s f varsInEnv ps
+                          (res1,su2) <- proveImplicationIO s dedupErrs f varsInEnv ps
                                                  (apSubst su1 asmps0) newGs
                           let su3 = su2 @@ su1
                           case res1 of
-                            Left bad -> return (Left bad, su3)
-                            Right ws1 -> return (Right (ws++ws1),su3)
+                            Left bad -> return (Left (bad ++ errs), su3)
+                            Right ws1
+                              | null errs -> return (Right (ws++ws1),su3)
+                              | otherwise -> return (Left errs, su3)
   where
-  err us =  Left $ cleanupError
-                 $ UnsolvedDelayedCt
-                 $ DelayedCt { dctSource = f
-                             , dctForall = ps
-                             , dctAsmps  = asmps0
-                             , dctGoals  = us
-                             }
+  err us = cleanupError
+           $ UnsolvedDelayedCt
+           $ DelayedCt { dctSource = f
+                       , dctForall = ps
+                       , dctAsmps  = asmps0
+                       , dctGoals  = if dedupErrs then nubOrd us else us
+                       }
 
 
   asmps1 = concatMap pSplitAnd asmps0
   (asmps,gs) =
      let gs1 = [ g { goal = p } | g <- gs0, p <- pSplitAnd (goal g)
                                 , notElem p asmps1 ]
-     in case matchMaybe (improveProps True Map.empty asmps1) of
+     in case matchMaybe (improveProps True mempty asmps1) of
           Nothing -> (asmps1,gs1)
           Just (newSu,newAsmps) ->
              ( [ TVar x =#= t | (x,t) <- substToList newSu ]
@@ -308,9 +383,28 @@ cleanupError err =
 
 
 
-assumptionIntervals :: Ctxt -> [Prop] -> Ctxt
-assumptionIntervals as ps =
-  case computePropIntervals as ps of
-    NoChange -> as
-    InvalidInterval {} -> as -- XXX: say something
-    NewIntervals bs -> Map.union bs as
+buildSolverCtxt :: [Prop] -> Ctxt
+buildSolverCtxt ps0 =
+  let ps = saturateProps mempty ps0
+      ivals = assumptionIntervals mempty (Set.toList ps)
+   in SolverCtxt
+      { intervals = ivals
+      , saturatedAsmps = ps
+      }
+
+ where
+ saturateProps gs [] = gs
+ saturateProps gs (p:ps)
+   | Set.member p gs = saturateProps gs ps
+   | Just (n,_) <- pIsLiteral p =
+       let gs' = Set.fromList [p, pFin n] <> gs
+        in saturateProps gs' ps
+   | otherwise =
+        let gs' = Set.singleton p <> superclassSet p <> gs
+         in saturateProps gs' ps
+
+ assumptionIntervals as ps =
+   case computePropIntervals as ps of
+     NoChange -> as
+     InvalidInterval {} -> as -- XXX: say something
+     NewIntervals bs -> Map.union bs as
