@@ -10,11 +10,13 @@
 
 {-# LANGUAGE PatternGuards #-}
 {-# LANGUAGE BangPatterns #-}
+{-# LANGUAGE Safe #-}
 
 module Cryptol.TypeCheck.Solver.Numeric.Interval where
 
 import Cryptol.TypeCheck.AST
 import Cryptol.TypeCheck.Solver.InfNat
+import Cryptol.TypeCheck.PP(NameMap,ppWithNames)
 import Cryptol.Utils.PP hiding (int)
 
 import           Data.Map ( Map )
@@ -108,11 +110,11 @@ propInterval varInts prop = catMaybes
        x  <- tIsVar ty
        return (x,iAnyFin)
 
-  , do (l,r) <- pIsEq prop
+  , do (l,r) <- pIsEqual prop
        x     <- tIsVar l
        return (x,typeInterval varInts r)
 
-  , do (l,r) <- pIsEq prop
+  , do (l,r) <- pIsEqual prop
        x     <- tIsVar r
        return (x,typeInterval varInts l)
 
@@ -138,6 +140,14 @@ propInterval varInts prop = catMaybes
 
        return (x, Interval { iLower = Nat 0, iUpper = ub })
 
+    , do (e,_) <- pIsValidFloat prop
+         x <- tIsVar e
+         pure (x, iAnyFin)
+
+    , do (_,p) <- pIsValidFloat prop
+         x <- tIsVar p
+         pure (x, iAnyFin)
+
   ]
 
 
@@ -155,6 +165,14 @@ ppIntervals  = vcat . map ppr . Map.toList
   where
   ppr (var,i) = pp var <.> char ':' <+> ppInterval i
 
+ppIntervalsWithNames :: NameMap -> Map TVar Interval -> Doc
+ppIntervalsWithNames nms = vcat . map ppr . Map.toList
+  where
+  ppr :: (TVar,Interval) -> Doc
+  ppr (var,i) = ppWithNames nms var <.> char ':' <+> ppInterval i
+
+
+
 ppInterval :: Interval -> Doc
 ppInterval x = brackets (hsep [ ppr (iLower x)
                               , text ".."
