@@ -5,18 +5,20 @@
 -- Maintainer  :  cryptol@galois.com
 -- Stability   :  provisional
 -- Portability :  portable
-{-# LANGUAGE PatternGuards, Safe #-}
+{-# LANGUAGE PatternGuards #-}
 module Cryptol.TypeCheck.Solver.Selector (tryHasGoal) where
 
+import Cryptol.Parser.Position(Range)
 import Cryptol.TypeCheck.AST
 import Cryptol.TypeCheck.InferTypes
-import Cryptol.TypeCheck.Monad( InferM, unify, newGoals, lookupNewtype
+import Cryptol.TypeCheck.Monad( InferM, unify, newGoals
                               , newType, applySubst, solveHasGoal
-                              , newParamName
+                              , newLocalName
                               )
 import Cryptol.TypeCheck.Subst (listParamSubst, apSubst)
-import Cryptol.Utils.Ident (Ident, packIdent)
+import Cryptol.Utils.Ident (Ident, packIdent,Namespace(..))
 import Cryptol.Utils.Panic(panic)
+import Cryptol.Utils.RecordMap
 
 import Control.Monad(forM,guard)
 
@@ -26,7 +28,7 @@ recordType labels =
   do fields <- forM labels $ \l ->
         do t <- newType (TypeOfRecordField l) KType
            return (l,t)
-     return (TRec fields)
+     return (TRec (recordFromFields fields))
 
 tupleType :: Int -> InferM Type
 tupleType n =
@@ -40,8 +42,8 @@ listType n =
      return (tSeq (tNum n) elems)
 
 
-improveSelector :: Selector -> Type -> InferM Bool
-improveSelector sel outerT =
+improveSelector :: Maybe Range -> Selector -> Type -> InferM Bool
+improveSelector rng sel outerT =
   case sel of
     RecordSel _ mb -> cvt recordType mb
     TupleSel  _ mb -> cvt tupleType  mb
@@ -49,7 +51,8 @@ improveSelector sel outerT =
   where
   cvt _ Nothing   = return False
   cvt f (Just a)  = do ty <- f a
-                       newGoals CtExactType =<< unify ty outerT
+                       ps <- unify (WithSource outerT (selSrc sel) rng) ty
+                       newGoals CtExactType ps
                        newT <- applySubst outerT
                        return (newT /= outerT)
 
@@ -64,21 +67,18 @@ solveSelector sel outerT =
 
     (RecordSel l _, ty) ->
       case ty of
-        TRec fs  -> return (lookup l fs)
+        TRec fs -> return (lookupField l fs)
+        TNewtype nt ts ->
+          case lookupField l (ntFields nt) of
+            Nothing -> return Nothing
+            Just t ->
+              do let su = listParamSubst (zip (ntParams nt) ts)
+                 newGoals (CtPartialTypeFun (ntName nt))
+                   $ apSubst su $ ntConstraints nt
+                 return $ Just $ apSubst su t
+
         TCon (TC TCSeq) [len,el] -> liftSeq len el
         TCon (TC TCFun) [t1,t2]  -> liftFun t1 t2
-        TCon (TC (TCNewtype (UserTC x _))) ts ->
-          do mb <- lookupNewtype x
-             case mb of
-               Nothing -> return Nothing
-               Just nt ->
-                 case lookup l (ntFields nt) of
-                   Nothing -> return Nothing
-                   Just t  ->
-                     do let su = listParamSubst (zip (ntParams nt) ts)
-                        newGoals (CtPartialTypeFun x)
-                          $ apSubst su $ ntConstraints nt
-                        return $ Just $ apSubst su t
         _ -> return Nothing
 
     (TupleSel n _, ty) ->
@@ -118,13 +118,15 @@ solveSelector sel outerT =
 tryHasGoal :: HasGoal -> InferM (Bool, Bool) -- ^ changes, solved
 tryHasGoal has
   | TCon (PC (PHas sel)) [ th, ft ] <- goal (hasGoal has) =
-    do imped     <- improveSelector sel th
+    do let rng = Just (goalRange (hasGoal has))
+       imped     <- improveSelector rng sel th
        outerT    <- tNoUser `fmap` applySubst th
        mbInnerT  <- solveSelector sel outerT
        case mbInnerT of
          Nothing -> return (imped, False)
          Just innerT ->
-           do newGoals CtExactType =<< unify innerT ft
+           do newGoals CtExactType =<<
+                                  unify (WithSource innerT (selSrc sel) rng) ft
               oT <- applySubst outerT
               iT <- applySubst innerT
               sln <- mkSelSln sel oT iT
@@ -157,16 +159,16 @@ mkSelSln s outerT innerT =
       | RecordSel {} <- s -> liftFun t1 t2
 
     _ -> return HasGoalSln { hasDoSelect = \e -> ESel e s
-                           , hasDoSet    = \e v -> ESet e s v }
+                           , hasDoSet    = \e v -> ESet outerT e s v }
 
   where
   -- Has s a t => Has s ([n]a) ([n]t)
   -- xs.s             ~~> [ x.s           | x <- xs ]
   -- { xs | s = ys }  ~~> [ { x | s = y } | x <- xs | y <- ys ]
   liftSeq len el =
-    do x1 <- newParamName (packIdent "x")
-       x2 <- newParamName (packIdent "x")
-       y2 <- newParamName (packIdent "y")
+    do x1 <- newLocalName NSValue (packIdent "x")
+       x2 <- newLocalName NSValue (packIdent "x")
+       y2 <- newLocalName NSValue (packIdent "y")
        case tNoUser innerT of
          TCon _ [_,eli] ->
            do d <- mkSelSln s el eli
@@ -184,12 +186,12 @@ mkSelSln s outerT innerT =
 
          _ -> panic "mkSelSln" [ "Unexpected inner seq type.", show innerT ]
 
-  -- Has s b t => Has s (a -> b)
+  -- Has s b t => Has s (a -> b) (a -> t)
   -- f.s            ~~> \x -> (f x).s
   -- { f | s = g }  ~~> \x -> { f x | s = g x }
   liftFun t1 t2 =
-    do x1 <- newParamName (packIdent "x")
-       x2 <- newParamName (packIdent "x")
+    do x1 <- newLocalName NSValue (packIdent "x")
+       x2 <- newLocalName NSValue (packIdent "x")
        case tNoUser innerT of
          TCon _ [_,inT] ->
            do d <- mkSelSln s t2 inT
