@@ -15,13 +15,14 @@ import Data.Maybe(listToMaybe)
 
 
 
--- | All ways to split a type in the form: `a + t1`, where `a` is a variable.
+-- | All ways to split a type in the form: @a + t1@, where @a@ is a variable.
 splitVarSummands :: Type -> [(TVar,Type)]
 splitVarSummands ty0 = [ (x,t1) | (x,t1) <- go ty0, tNum (0::Int) /= t1 ]
   where
   go ty = case ty of
             TVar x      -> return (x, tNum (0::Int))
             TRec {}     -> []
+            TNewtype{}  -> []
             TUser _ _ t -> go t
             TCon (TF TCAdd) [t1,t2] ->
               do (a,yes) <- go t1
@@ -32,12 +33,12 @@ splitVarSummands ty0 = [ (x,t1) | (x,t1) <- go ty0, tNum (0::Int) /= t1 ]
             TCon _ _    -> [] -- XXX: we could do some distributivity etc
 
 
--- | Check if we can express a type in the form: `a + t1`.
+-- | Check if we can express a type in the form: @a + t1@.
 splitVarSummand :: TVar -> Type -> Maybe Type
 splitVarSummand a ty = listToMaybe [ t | (x,t) <- splitVarSummands ty, x == a ]
 
-{- | Check if we can express a type in the form: `k + t1`,
-where `k` is a constant > 0.
+{- | Check if we can express a type in the form: @k + t1@,
+where @k@ is a constant > 0.
 This assumes that the type has been simplified already,
 so that constants are floated to the left. -}
 splitConstSummand :: Type -> Maybe (Integer, Type)
@@ -45,6 +46,7 @@ splitConstSummand ty =
   case ty of
     TVar {}     -> Nothing
     TRec {}     -> Nothing
+    TNewtype{}  -> Nothing
     TUser _ _ t -> splitConstSummand t
     TCon (TF TCAdd) [t1,t2] ->
       do (k,t1') <- splitConstSummand t1
@@ -54,8 +56,8 @@ splitConstSummand ty =
     TCon (TC (TCNum k)) [] -> guard (k > 0) >> return (k, tNum (0::Int))
     TCon {}     -> Nothing
 
-{- | Check if we can express a type in the form: `k * t1`,
-where `k` is a constant > 1
+{- | Check if we can express a type in the form: @k * t1@,
+where @k@ is a constant > 1.
 This assumes that the type has been simplified already,
 so that constants are floated to the left. -}
 splitConstFactor :: Type -> Maybe (Integer, Type)
@@ -63,6 +65,7 @@ splitConstFactor ty =
   case ty of
     TVar {}     -> Nothing
     TRec {}     -> Nothing
+    TNewtype{}  -> Nothing
     TUser _ _ t -> splitConstFactor t
     TCon (TF TCMul) [t1,t2] ->
       do (k,t1') <- splitConstFactor t1
