@@ -2,6 +2,11 @@
 {-# Language FlexibleInstances, FlexibleContexts #-}
 {-# Language PatternGuards #-}
 {-# Language OverloadedStrings #-}
+{-| This module contains types related to typechecking and the output of the
+typechecker.  In particular, it should contain the types needed by
+interface files (see 'Crytpol.ModuleSystem.Interface'), which are (kind of)
+the output of the typechker.
+-}
 module Cryptol.TypeCheck.Type
   ( module Cryptol.TypeCheck.Type
   , module Cryptol.TypeCheck.TCon
@@ -11,27 +16,120 @@ module Cryptol.TypeCheck.Type
 import GHC.Generics (Generic)
 import Control.DeepSeq
 
+import           Data.Map(Map)
+import qualified Data.Map as Map
 import qualified Data.IntMap as IntMap
+import           Data.Maybe (fromMaybe)
 import           Data.Set (Set)
 import qualified Data.Set as Set
-import           Data.List(sortBy)
-import           Data.Ord(comparing)
+import           Data.Text (Text)
 
 import Cryptol.Parser.Selector
-import Cryptol.Parser.Fixity
-import Cryptol.Parser.Position(Range,emptyRange)
+import Cryptol.Parser.Position(Located,thing,Range,emptyRange)
+import Cryptol.Parser.AST(ImpName(..))
 import Cryptol.ModuleSystem.Name
-import Cryptol.Utils.Ident (Ident)
+import Cryptol.Utils.Ident (Ident, isInfixIdent, exprModName, ogModule, ModName)
 import Cryptol.TypeCheck.TCon
 import Cryptol.TypeCheck.PP
 import Cryptol.TypeCheck.Solver.InfNat
+import Cryptol.Utils.Fixity
 import Cryptol.Utils.Panic(panic)
+import Cryptol.Utils.RecordMap
 import Prelude
 
 infix  4 =#=, >==
 infixr 5 `tFun`
 
 
+--------------------------------------------------------------------------------
+-- Module parameters
+
+type FunctorParams = Map Ident ModParam
+
+-- | Compute the names from all functor parameters
+allParamNames :: FunctorParams -> ModParamNames
+allParamNames mps =
+  ModParamNames
+    { mpnTypes       = Map.unions (map mpnTypes ps)
+    , mpnConstraints = concatMap mpnConstraints ps
+    , mpnFuns        = Map.unions (map mpnFuns ps)
+    , mpnTySyn       = Map.unions (map mpnTySyn ps)
+    , mpnDoc         = Nothing
+    }
+  where
+  ps = map mpParameters (Map.elems mps)
+
+
+-- | A module parameter.  Corresponds to a "signature import".
+-- A single module parameter can bring multiple things in scope.
+data ModParam = ModParam
+  { mpName        :: Ident
+    -- ^ The name of a functor parameter.
+
+  , mpQual        :: !(Maybe ModName)
+    -- ^ This is the qualifier for the parameter.  We use it to
+    -- derive parameter names when doing `_` imports.
+
+  , mpIface       :: ImpName Name
+    -- ^ The interface corresponding to this parameter.
+    -- This is thing in `import interface`
+
+  , mpParameters  :: ModParamNames
+    {- ^ These are the actual parameters, not the ones in the interface
+      For example if the same interface is used for multiple parameters
+      the `ifmpParameters` would all be different. -}
+  } deriving (Show, Generic, NFData)
+
+-- | Information about the names brought in through an "interface import".
+-- This is also used to keep information about.
+data ModParamNames = ModParamNames
+  { mpnTypes       :: Map Name ModTParam
+    -- ^ Type parameters
+
+  , mpnTySyn      :: !(Map Name TySyn)
+    -- ^ Type synonyms
+
+  , mpnConstraints :: [Located Prop]
+    -- ^ Constraints on param. types
+
+
+  , mpnFuns        :: Map.Map Name ModVParam
+    -- ^ Value parameters
+
+  , mpnDoc         :: !(Maybe Text)
+    -- ^ Documentation about the interface.
+  } deriving (Show, Generic, NFData)
+
+-- | A type parameter of a module.
+data ModTParam = ModTParam
+  { mtpName   :: Name
+  , mtpKind   :: Kind
+  , mtpDoc    :: Maybe Text
+  } deriving (Show,Generic,NFData)
+
+
+-- | This is how module parameters appear in actual types.
+mtpParam :: ModTParam -> TParam
+mtpParam mtp = TParam { tpUnique = nameUnique (mtpName mtp)
+                      , tpKind   = mtpKind mtp
+                      , tpFlav   = TPModParam (mtpName mtp)
+                      , tpInfo   = desc
+                      }
+  where desc = TVarInfo { tvarDesc   = TVFromModParam (mtpName mtp)
+                        , tvarSource = nameLoc (mtpName mtp)
+                        }
+
+-- | A value parameter of a module.
+data ModVParam = ModVParam
+  { mvpName   :: Name
+  , mvpType   :: Schema
+  , mvpDoc    :: Maybe Text
+  , mvpFixity :: Maybe Fixity       -- XXX: This should be in the name?
+  } deriving (Show,Generic,NFData)
+--------------------------------------------------------------------------------
+
+
+
 
 -- | The types of polymorphic values.
 data Schema = Forall { sVars :: [TParam], sProps :: [Prop], sType :: Type }
@@ -48,7 +146,12 @@ data TParam = TParam { tpUnique :: !Int       -- ^ Parameter identifier
               deriving (Generic, NFData, Show)
 
 data TPFlavor = TPModParam Name
-              | TPOther (Maybe Name)
+              | TPUnifyVar
+              | TPSchemaParam Name
+              | TPTySynParam Name
+              | TPPropSynParam Name
+              | TPNewtypeParam Name
+              | TPPrimParam Name
               deriving (Generic, NFData, Show)
 
 tMono :: Type -> Schema
@@ -61,18 +164,17 @@ isMono s =
     _              -> Nothing
 
 
-
 schemaParam :: Name -> TPFlavor
-schemaParam x = TPOther (Just x)
+schemaParam = TPSchemaParam
 
 tySynParam :: Name -> TPFlavor
-tySynParam x = TPOther (Just x)
+tySynParam = TPTySynParam
 
 propSynParam :: Name -> TPFlavor
-propSynParam x = TPOther (Just x)
+propSynParam = TPPropSynParam
 
 newtypeParam :: Name -> TPFlavor
-newtypeParam x = TPOther (Just x)
+newtypeParam = TPNewtypeParam
 
 modTyParam :: Name -> TPFlavor
 modTyParam = TPModParam
@@ -81,8 +183,13 @@ modTyParam = TPModParam
 tpfName :: TPFlavor -> Maybe Name
 tpfName f =
   case f of
-    TPModParam x -> Just x
-    TPOther x -> x
+    TPUnifyVar       -> Nothing
+    TPModParam x     -> Just x
+    TPSchemaParam x  -> Just x
+    TPTySynParam x   -> Just x
+    TPPropSynParam x -> Just x
+    TPNewtypeParam x -> Just x
+    TPPrimParam x    -> Just x
 
 tpName :: TParam -> Maybe Name
 tpName = tpfName . tpFlav
@@ -99,20 +206,23 @@ data Type   = TCon !TCon ![Type]
               {- ^ This is just a type annotation, for a type that
               was written as a type synonym.  It is useful so that we
               can use it to report nicer errors.
-              Example: `TUser T ts t` is really just the type `t` that
-              was written as `T ts` by the user. -}
+              Example: @TUser T ts t@ is really just the type @t@ that
+              was written as @T ts@ by the user. -}
 
-            | TRec ![(Ident,Type)]
+            | TRec !(RecordMap Ident Type)
               -- ^ Record type
 
+            | TNewtype !Newtype ![Type]
+              -- ^ A newtype
+
               deriving (Show, Generic, NFData)
 
 
 
 -- | Type variables.
 data TVar   = TVFree !Int Kind (Set TParam) TVarInfo
-              -- ^ Unique, kind, ids of bound type variables that are in scope
-              -- The last field gives us some infor for nicer warnings/errors.
+              -- ^ Unique, kind, ids of bound type variables that are in scope.
+              -- The last field gives us some info for nicer warnings/errors.
 
 
             | TVBound {-# UNPACK #-} !TParam
@@ -124,13 +234,18 @@ tvInfo tv =
     TVFree _ _ _ d -> d
     TVBound tp     -> tpInfo tp
 
-data TVarInfo = TVarInfo { tvarSource :: Range -- ^ Source code that gave rise
-                         , tvarDesc   :: TVarSource -- ^ Description
+tvUnique :: TVar -> Int
+tvUnique (TVFree u _ _ _) = u
+tvUnique (TVBound TParam { tpUnique = u }) = u
+
+data TVarInfo = TVarInfo { tvarSource :: !Range -- ^ Source code that gave rise
+                         , tvarDesc   :: !TypeSource -- ^ Description
                          }
               deriving (Show, Generic, NFData)
 
 
-data TVarSource = TVFromModParam Name     -- ^ Name of module parameter
+-- | Explains how this type came to be, for better error messages.
+data TypeSource = TVFromModParam Name     -- ^ Name of module parameter
                 | TVFromSignature Name    -- ^ A variable in a signature
                 | TypeWildCard
                 | TypeOfRecordField Ident
@@ -141,13 +256,26 @@ data TVarSource = TVFromModParam Name     -- ^ Name of module parameter
                 | TypeParamInstPos   {-Fun-}Name {-Pos (from 1)-}Int
                 | DefinitionOf Name
                 | LenOfCompGen
-                | TypeOfArg (Maybe Int)
+                | TypeOfArg ArgDescr
                 | TypeOfRes
+                | FunApp
+                | TypeOfIfCondExpr
+                | TypeFromUserAnnotation
+                | GeneratorOfListComp
                 | TypeErrorPlaceHolder
                   deriving (Show, Generic, NFData)
 
+data ArgDescr = ArgDescr
+  { argDescrFun    :: Maybe Name
+  , argDescrNumber :: Maybe Int
+  }
+  deriving (Show,Generic,NFData)
+
+noArgDescr :: ArgDescr
+noArgDescr = ArgDescr { argDescrFun = Nothing, argDescrNumber = Nothing }
+
 -- | Get the names of something that is related to the tvar.
-tvSourceName :: TVarSource -> Maybe Name
+tvSourceName :: TypeSource -> Maybe Name
 tvSourceName tvs =
   case tvs of
     TVFromModParam x -> Just x
@@ -155,9 +283,19 @@ tvSourceName tvs =
     TypeParamInstNamed x _ -> Just x
     TypeParamInstPos x _ -> Just x
     DefinitionOf x -> Just x
+    TypeOfArg x -> argDescrFun x
     _ -> Nothing
 
--- | The type is supposed to be of kind `KProp`
+
+-- | A type annotated with information on how it came about.
+data TypeWithSource = WithSource
+  { twsType   :: Type
+  , twsSource :: TypeSource
+  , twsRange  :: !(Maybe Range)
+  }
+
+
+-- | The type is supposed to be of kind 'KProp'.
 type Prop   = Type
 
 
@@ -169,7 +307,7 @@ data TySyn  = TySyn { tsName        :: Name       -- ^ Name
                     , tsParams      :: [TParam]   -- ^ Parameters
                     , tsConstraints :: [Prop]     -- ^ Ensure body is OK
                     , tsDef         :: Type       -- ^ Definition
-                    , tsDoc         :: !(Maybe String) -- ^ Documentation
+                    , tsDoc         :: !(Maybe Text) -- ^ Documentation
                     }
               deriving (Show, Generic, NFData)
 
@@ -181,18 +319,26 @@ data TySyn  = TySyn { tsName        :: Name       -- ^ Name
 data Newtype  = Newtype { ntName   :: Name
                         , ntParams :: [TParam]
                         , ntConstraints :: [Prop]
-                        , ntFields :: [(Ident,Type)]
-                        , ntDoc :: Maybe String
+                        , ntConName :: !Name
+                        , ntFields :: RecordMap Ident Type
+                        , ntDoc :: Maybe Text
                         } deriving (Show, Generic, NFData)
 
 
+instance Eq Newtype where
+  x == y = ntName x == ntName y
+
+instance Ord Newtype where
+  compare x y = compare (ntName x) (ntName y)
+
+
 -- | Information about an abstract type.
 data AbstractType = AbstractType
   { atName    :: Name
   , atKind    :: Kind
   , atCtrs    :: ([TParam], [Prop])
   , atFixitiy :: Maybe Fixity
-  , atDoc     :: Maybe String
+  , atDoc     :: Maybe Text
   } deriving (Show, Generic, NFData)
 
 
@@ -200,6 +346,8 @@ data AbstractType = AbstractType
 
 --------------------------------------------------------------------------------
 
+instance HasKind AbstractType where
+  kindOf at = foldr (:->) (atKind at) (map kindOf (fst (atCtrs at)))
 
 instance HasKind TVar where
   kindOf (TVFree  _ k _ _) = k
@@ -212,6 +360,7 @@ instance HasKind Type where
       TCon c ts   -> quickApply (kindOf c) ts
       TUser _ _ t -> kindOf t
       TRec {}     -> KType
+      TNewtype{}  -> KType
 
 instance HasKind TySyn where
   kindOf ts = foldr (:->) (kindOf (tsDef ts)) (map kindOf (tsParams ts))
@@ -234,16 +383,15 @@ kindResult k         = k
 
 --------------------------------------------------------------------------------
 
--- Syntactic equality, ignoring type synonyms and record order
+-- | Syntactic equality, ignoring type synonyms and record order.
 instance Eq Type where
   TUser _ _ x == y        = x == y
   x == TUser _ _ y        = y == x
 
   TCon x xs == TCon y ys  = x == y && xs == ys
   TVar x    == TVar y     = x == y
-
-  TRec xs   == TRec ys    = norm xs == norm ys
-    where norm = sortBy (comparing fst)
+  TRec xs   == TRec ys    = xs == ys
+  TNewtype ntx xs == TNewtype nty ys = ntx == nty && xs == ys
 
   _         == _          = False
 
@@ -261,10 +409,11 @@ instance Ord Type where
       (TCon {}, _)            -> LT
       (_,TCon {})             -> GT
 
-      (TRec xs, TRec ys)      -> compare (norm xs) (norm ys)
-        where norm = sortBy (comparing fst)
-
+      (TRec xs, TRec ys)      -> compare xs ys
+      (TRec{}, _)             -> LT
+      (_, TRec{})             -> GT
 
+      (TNewtype x xs, TNewtype y ys) -> compare (x,xs) (y,ys)
 
 instance Eq TParam where
   x == y = tpUnique x == tpUnique y
@@ -279,11 +428,36 @@ tpVar p = TVBound p
 type SType  = Type
 
 
+--------------------------------------------------------------------
+-- Superclass
+
+-- | Compute the set of all @Prop@s that are implied by the
+--   given prop via superclass constraints.
+superclassSet :: Prop -> Set Prop
+
+superclassSet (TCon (PC PPrime) [n]) =
+  Set.fromList [ pFin n, n >== tTwo ]
+
+superclassSet (TCon (PC p0) [t]) = go p0
+  where
+  super p = Set.insert (TCon (PC p) [t]) (go p)
+
+  go PRing      = super PZero
+  go PLogic     = super PZero
+  go PField     = super PRing
+  go PIntegral  = super PRing
+  go PRound     = super PField <> super PCmp
+  go PCmp       = super PEq
+  go PSignedCmp = super PEq
+  go _ = mempty
+
+superclassSet _ = mempty
+
 
 newtypeConType :: Newtype -> Schema
 newtypeConType nt =
   Forall as (ntConstraints nt)
-    $ TRec (ntFields nt) `tFun` TCon (newtypeTyCon nt) (map (TVar . tpVar) as)
+    $ TRec (ntFields nt) `tFun` TNewtype nt (map (TVar . tpVar) as)
   where
   as = ntParams nt
 
@@ -291,7 +465,15 @@ newtypeConType nt =
 abstractTypeTC :: AbstractType -> TCon
 abstractTypeTC at =
   case builtInType (atName at) of
-    Just tcon -> tcon
+    Just tcon
+      | kindOf tcon == atKind at -> tcon
+      | otherwise ->
+        panic "abstractTypeTC"
+          [ "Mismatch between built-in and declared type."
+          , "Name: " ++ pretty (atName at)
+          , "Declared: " ++ pretty (atKind at)
+          , "Built-in: " ++ pretty (kindOf tcon)
+          ]
     _         -> TC $ TCAbstract $ UserTC (atName at) (atKind at)
 
 instance Eq TVar where
@@ -318,11 +500,20 @@ isBoundTV (TVBound {})  = True
 isBoundTV _             = False
 
 
-tIsError :: Type -> Maybe TCErrorMessage
+tIsError :: Type -> Maybe Type
 tIsError ty = case tNoUser ty of
-                TCon (TError _ x) _ -> Just x
+                TCon (TError _) [t] -> Just t
+                TCon (TError _) _   -> panic "tIsError" ["Malformed error"]
                 _                   -> Nothing
 
+tHasErrors :: Type -> Bool
+tHasErrors ty =
+  case tNoUser ty of
+    TCon (TError _) _   -> True
+    TCon _ ts           -> any tHasErrors ts
+    TRec mp             -> any tHasErrors mp
+    _                   -> False
+
 tIsNat' :: Type -> Maybe Nat'
 tIsNat' ty =
   case tNoUser ty of
@@ -367,12 +558,24 @@ tIsIntMod ty = case tNoUser ty of
                  TCon (TC TCIntMod) [n] -> Just n
                  _                      -> Nothing
 
+tIsRational :: Type -> Bool
+tIsRational ty =
+  case tNoUser ty of
+    TCon (TC TCRational) [] -> True
+    _                       -> False
+
+tIsFloat :: Type -> Maybe (Type, Type)
+tIsFloat ty =
+  case tNoUser ty of
+    TCon (TC TCFloat) [e, p] -> Just (e, p)
+    _                        -> Nothing
+
 tIsTuple :: Type -> Maybe [Type]
 tIsTuple ty = case tNoUser ty of
                 TCon (TC (TCTuple _)) ts -> Just ts
                 _                        -> Nothing
 
-tIsRec :: Type -> Maybe [(Ident, Type)]
+tIsRec :: Type -> Maybe (RecordMap Ident Type)
 tIsRec ty = case tNoUser ty of
               TRec fs -> Just fs
               _       -> Nothing
@@ -395,15 +598,20 @@ pIsFin ty = case tNoUser ty of
               TCon (PC PFin) [t1] -> Just t1
               _                   -> Nothing
 
+pIsPrime :: Prop -> Maybe Type
+pIsPrime ty = case tNoUser ty of
+                TCon (PC PPrime) [t1] -> Just t1
+                _                     -> Nothing
+
 pIsGeq :: Prop -> Maybe (Type,Type)
 pIsGeq ty = case tNoUser ty of
               TCon (PC PGeq) [t1,t2] -> Just (t1,t2)
               _                      -> Nothing
 
-pIsEq :: Prop -> Maybe (Type,Type)
-pIsEq ty = case tNoUser ty of
-             TCon (PC PEqual) [t1,t2] -> Just (t1,t2)
-             _                        -> Nothing
+pIsEqual :: Prop -> Maybe (Type,Type)
+pIsEqual ty = case tNoUser ty of
+                TCon (PC PEqual) [t1,t2] -> Just (t1,t2)
+                _                        -> Nothing
 
 pIsZero :: Prop -> Maybe Type
 pIsZero ty = case tNoUser ty of
@@ -415,11 +623,31 @@ pIsLogic ty = case tNoUser ty of
                 TCon (PC PLogic) [t1] -> Just t1
                 _                     -> Nothing
 
-pIsArith :: Prop -> Maybe Type
-pIsArith ty = case tNoUser ty of
-                TCon (PC PArith) [t1] -> Just t1
+pIsRing :: Prop -> Maybe Type
+pIsRing ty = case tNoUser ty of
+                TCon (PC PRing) [t1] -> Just t1
+                _                    -> Nothing
+
+pIsField :: Prop -> Maybe Type
+pIsField ty = case tNoUser ty of
+                TCon (PC PField) [t1] -> Just t1
                 _                     -> Nothing
 
+pIsIntegral :: Prop -> Maybe Type
+pIsIntegral ty = case tNoUser ty of
+                   TCon (PC PIntegral) [t1] -> Just t1
+                   _                        -> Nothing
+
+pIsRound :: Prop -> Maybe Type
+pIsRound ty = case tNoUser ty of
+                     TCon (PC PRound) [t1] -> Just t1
+                     _                     -> Nothing
+
+pIsEq :: Prop -> Maybe Type
+pIsEq ty = case tNoUser ty of
+             TCon (PC PEq) [t1] -> Just t1
+             _                  -> Nothing
+
 pIsCmp :: Prop -> Maybe Type
 pIsCmp ty = case tNoUser ty of
               TCon (PC PCmp) [t1] -> Just t1
@@ -435,6 +663,17 @@ pIsLiteral ty = case tNoUser ty of
                   TCon (PC PLiteral) [t1, t2] -> Just (t1, t2)
                   _                           -> Nothing
 
+pIsLiteralLessThan :: Prop -> Maybe (Type, Type)
+pIsLiteralLessThan ty =
+  case tNoUser ty of
+    TCon (PC PLiteralLessThan) [t1, t2] -> Just (t1, t2)
+    _                                   -> Nothing
+
+pIsFLiteral :: Prop -> Maybe (Type,Type,Type,Type)
+pIsFLiteral ty = case tNoUser ty of
+                   TCon (PC PFLiteral) [t1,t2,t3,t4] -> Just (t1,t2,t3,t4)
+                   _                                 -> Nothing
+
 pIsTrue :: Prop -> Bool
 pIsTrue ty  = case tNoUser ty of
                 TCon (PC PTrue) _ -> True
@@ -445,6 +684,11 @@ pIsWidth ty = case tNoUser ty of
                 TCon (TF TCWidth) [t1] -> Just t1
                 _                      -> Nothing
 
+pIsValidFloat :: Prop -> Maybe (Type,Type)
+pIsValidFloat ty = case tNoUser ty of
+                     TCon (PC PValidFloat) [a,b] -> Just (a,b)
+                     _                           -> Nothing
+
 --------------------------------------------------------------------------------
 
 
@@ -471,15 +715,27 @@ tNat' n'  = case n' of
 tAbstract :: UserTC -> [Type] -> Type
 tAbstract u ts = TCon (TC (TCAbstract u)) ts
 
+tNewtype :: Newtype -> [Type] -> Type
+tNewtype nt ts = TNewtype nt ts
+
 tBit     :: Type
 tBit      = TCon (TC TCBit) []
 
 tInteger :: Type
 tInteger  = TCon (TC TCInteger) []
 
+tRational :: Type
+tRational  = TCon (TC TCRational) []
+
+tFloat   :: Type -> Type -> Type
+tFloat e p = TCon (TC TCFloat) [ e, p ]
+
 tIntMod :: Type -> Type
 tIntMod n = TCon (TC TCIntMod) [n]
 
+tArray :: Type -> Type -> Type
+tArray a b = TCon (TC TCArray) [a, b]
+
 tWord    :: Type -> Type
 tWord a   = tSeq a tBit
 
@@ -492,22 +748,18 @@ tChar = tWord (tNum (8 :: Int))
 tString :: Int -> Type
 tString len = tSeq (tNum len) tChar
 
-tRec     :: [(Ident,Type)] -> Type
+tRec     :: RecordMap Ident Type -> Type
 tRec      = TRec
 
 tTuple   :: [Type] -> Type
 tTuple ts = TCon (TC (TCTuple (length ts))) ts
 
-newtypeTyCon :: Newtype -> TCon
-newtypeTyCon nt = TC $ TCNewtype $ UserTC (ntName nt) (kindOf nt)
-
-
 -- | Make a function type.
 tFun     :: Type -> Type -> Type
 tFun a b  = TCon (TC TCFun) [a,b]
 
 -- | Eliminate outermost type synonyms.
-tNoUser  :: Type -> Type
+tNoUser :: Type -> Type
 tNoUser t = case t of
               TUser _ _ a -> tNoUser a
               _           -> t
@@ -516,13 +768,11 @@ tNoUser t = case t of
 --------------------------------------------------------------------------------
 -- Construction of type functions
 
--- | Make a malformed numeric type.
-tBadNumber :: TCErrorMessage -> Type
-tBadNumber = tError KNum
-
--- | Make an error value of the given type.
-tError :: Kind -> TCErrorMessage -> Type
-tError k msg = TCon (TError k msg) []
+-- | Make an error value of the given type to replace
+-- the given malformed type (the argument to the error function)
+tError :: Type -> Type
+tError t = TCon (TError (k :-> k)) [t]
+  where k = kindOf t
 
 tf1 :: TFun -> Type -> Type
 tf1 f x = TCon (TF f) [x]
@@ -581,8 +831,20 @@ pZero t = TCon (PC PZero) [t]
 pLogic :: Type -> Prop
 pLogic t = TCon (PC PLogic) [t]
 
-pArith :: Type -> Prop
-pArith t = TCon (PC PArith) [t]
+pRing :: Type -> Prop
+pRing t = TCon (PC PRing) [t]
+
+pIntegral :: Type -> Prop
+pIntegral t = TCon (PC PIntegral) [t]
+
+pField :: Type -> Prop
+pField t = TCon (PC PField) [t]
+
+pRound :: Type -> Prop
+pRound t = TCon (PC PRound) [t]
+
+pEq :: Type -> Prop
+pEq t = TCon (PC PEq) [t]
 
 pCmp :: Type -> Prop
 pCmp t = TCon (PC PCmp) [t]
@@ -593,11 +855,14 @@ pSignedCmp t = TCon (PC PSignedCmp) [t]
 pLiteral :: Type -> Type -> Prop
 pLiteral x y = TCon (PC PLiteral) [x, y]
 
+pLiteralLessThan :: Type -> Type -> Prop
+pLiteralLessThan x y = TCon (PC PLiteralLessThan) [x, y]
+
 -- | Make a greater-than-or-equal-to constraint.
 (>==) :: Type -> Type -> Prop
 x >== y = TCon (PC PGeq) [x,y]
 
--- | A `Has` constraint, used for tuple and record selection.
+-- | A @Has@ constraint, used for tuple and record selection.
 pHas :: Selector -> Type -> Type -> Prop
 pHas l ty fi = TCon (PC (PHas l)) [ty,fi]
 
@@ -629,18 +894,77 @@ pFin :: Type -> Prop
 pFin ty =
   case tNoUser ty of
     TCon (TC (TCNum _)) _ -> pTrue
-    TCon (TC TCInf)     _ -> pError (TCErrorMessage "`inf` is not finite.")
-    _                     -> TCon (PC PFin) [ty]
+    TCon (TC TCInf)     _ -> tError prop -- XXX: should we be doing this here??
+    _                     -> prop
+  where
+  prop = TCon (PC PFin) [ty]
+
+pValidFloat :: Type -> Type -> Type
+pValidFloat e p = TCon (PC PValidFloat) [e,p]
+
+pPrime :: Type -> Prop
+pPrime ty =
+  case tNoUser ty of
+    TCon (TC TCInf) _ -> tError prop
+    _ -> prop
+  where
+  prop = TCon (PC PPrime) [ty]
+
+-- Negation --------------------------------------------------------------------
+
+{-| `pNegNumeric` negates a simple (i.e., not And, not prime, etc) prop
+over numeric type vars.  The result is a conjunction of properties.  -}
+pNegNumeric :: Prop -> [Prop]
+pNegNumeric prop =
+  case tNoUser prop of
+    TCon tcon tys ->
+      case tcon of
+        PC pc ->
+          case pc of
+
+            -- not (x == y)  <=>  x /= y
+            PEqual -> [TCon (PC PNeq) tys]
+
+            -- not (x /= y)  <=>  x == y
+            PNeq -> [TCon (PC PEqual) tys]
+
+            -- not (x >= y)  <=>  x /= y and y >= x
+            PGeq -> [TCon (PC PNeq) tys, TCon (PC PGeq) (reverse tys)]
+
+            -- not (fin x)  <=>  x == Inf
+            PFin | [ty] <- tys -> [ty =#= tInf]
+                 | otherwise -> bad
+
+            -- not True  <=>  0 == 1
+            PTrue -> [TCon (PC PEqual) [tZero, tOne]]
+
+            _ -> bad
+
+        TError _ki -> [prop] -- propogates `TError`
+
+        TC _tc -> bad
+        TF _tf -> bad
+
+    _ -> bad
+
+  where
+  bad = panic "pNegNumeric"
+          [ "Unexpeceted numeric constraint:"
+          , pretty prop
+          ]
 
--- | Make a malformed property.
-pError :: TCErrorMessage -> Prop
-pError msg = TCon (TError KProp msg) []
 
 --------------------------------------------------------------------------------
 
 noFreeVariables :: FVS t => t -> Bool
 noFreeVariables = all (not . isFreeTV) . Set.toList . fvs
 
+freeParams :: FVS t => t -> Set TParam
+freeParams x = Set.unions (map params (Set.toList (fvs x)))
+  where
+    params (TVFree _ _ tps _) = tps
+    params (TVBound tp) = Set.singleton tp
+
 class FVS t where
   fvs :: t -> Set TVar
 
@@ -652,7 +976,36 @@ instance FVS Type where
         TCon _ ts   -> fvs ts
         TVar x      -> Set.singleton x
         TUser _ _ t -> go t
-        TRec fs     -> fvs (map snd fs)
+        TRec fs     -> fvs (recordElements fs)
+        TNewtype _nt ts -> fvs ts
+
+
+-- | Find the abstract types mentioned in a type.
+class FreeAbstract t where
+  freeAbstract :: t -> Set UserTC
+
+instance FreeAbstract a => FreeAbstract [a] where
+  freeAbstract = Set.unions . map freeAbstract
+
+instance (FreeAbstract a, FreeAbstract b) => FreeAbstract (a,b) where
+  freeAbstract (a,b) = Set.union (freeAbstract a) (freeAbstract b)
+
+instance FreeAbstract TCon where
+  freeAbstract tc =
+    case tc of
+      TC (TCAbstract ut) -> Set.singleton ut
+      _                  -> Set.empty
+
+instance FreeAbstract Type where
+  freeAbstract ty =
+    case ty of
+      TCon tc ts      -> freeAbstract (tc,ts)
+      TVar {}         -> Set.empty
+      TUser _ _ t     -> freeAbstract t
+      TRec fs         -> freeAbstract (recordElements fs)
+      TNewtype _nt ts -> freeAbstract ts
+
+
 
 instance FVS a => FVS (Maybe a) where
   fvs Nothing  = Set.empty
@@ -669,10 +1022,6 @@ instance FVS Schema where
       Set.difference (Set.union (fvs ps) (fvs t)) bound
     where bound = Set.fromList (map tpVar as)
 
-
-
-
-
 -- Pretty Printing -------------------------------------------------------------
 
 instance PP TParam where
@@ -705,6 +1054,17 @@ ppNewtypeShort nt =
   ps  = ntParams nt
   nm = addTNames ps emptyNameMap
 
+ppNewtypeFull :: Newtype -> Doc
+ppNewtypeFull nt =
+  text "newtype" <+> pp (ntName nt) <+> hsep (map (ppWithNamesPrec nm 9) ps)
+  $$ nest 2 (cs $$ ("=" <+> pp (ntConName nt) $$ nest 2 fs))
+  where
+  ps = ntParams nt
+  nm = addTNames ps emptyNameMap
+  fs = vcat [ pp f <.> ":" <+> pp t | (f,t) <- canonicalFields (ntFields nt) ]
+  cs = vcat (map pp (ntConstraints nt))
+
+
 
 instance PP Schema where
   ppPrec = ppWithNamesPrec IntMap.empty
@@ -712,17 +1072,17 @@ instance PP Schema where
 instance PP (WithNames Schema) where
   ppPrec _ (WithNames s ns)
     | null (sVars s) && null (sProps s) = body
-    | otherwise = hang (vars <+> props) 2 body
+    | otherwise = nest 2 (sep (vars ++ props ++ [body]))
     where
     body = ppWithNames ns1 (sType s)
 
     vars = case sVars s of
-      [] -> empty
-      vs -> braces $ commaSep $ map (ppWithNames ns1) vs
+      [] -> []
+      vs -> [nest 1 (braces (commaSepFill (map (ppWithNames ns1) vs)))]
 
     props = case sProps s of
-      [] -> empty
-      ps -> parens (commaSep (map (ppWithNames ns1) ps)) <+> text "=>"
+      [] -> []
+      ps -> [nest 1 (parens (commaSepFill (map (ppWithNames ns1) ps))) <+> text "=>"]
 
     ns1 = addTNames (sVars s) ns
 
@@ -731,17 +1091,20 @@ instance PP TySyn where
 
 instance PP (WithNames TySyn) where
   ppPrec _ (WithNames ts ns) =
-    text "type" <+> ctr <+> lhs <+> char '=' <+> ppWithNames ns1 (tsDef ts)
+    nest 2 $ sep
+      [ fsep ([text "type"] ++ ctr ++ lhs ++ [char '='])
+      , ppWithNames ns1 (tsDef ts)
+      ]
     where ns1 = addTNames (tsParams ts) ns
           ctr = case kindResult (kindOf ts) of
-                  KProp -> text "constraint"
-                  _     -> empty
+                  KProp -> [text "constraint"]
+                  _     -> []
           n = tsName ts
           lhs = case (nameFixity n, tsParams ts) of
                   (Just _, [x, y]) ->
-                    ppWithNames ns1 x <+> pp (nameIdent n) <+> ppWithNames ns1 y
+                    [ppWithNames ns1 x, pp (nameIdent n), ppWithNames ns1 y]
                   (_, ps) ->
-                    pp n <+> sep (map (ppWithNames ns1) ps)
+                    [pp n] ++ map (ppWithNames ns1) ps
 
 instance PP Newtype where
   ppPrec = ppWithNamesPrec IntMap.empty
@@ -751,19 +1114,38 @@ instance PP (WithNames Newtype) where
 
 
 
-
+-- | The precedence levels used by this pretty-printing instance
+-- correspond with parser non-terminals as follows:
+--
+--   * 0-1: @type@
+--
+--   * 2: @infix_type@
+--
+--   * 3: @app_type@
+--
+--   * 4: @dimensions atype@
+--
+--   * 5: @atype@
 instance PP (WithNames Type) where
   ppPrec prec ty0@(WithNames ty nmMap) =
     case ty of
       TVar a  -> ppWithNames nmMap a
-      TRec fs -> braces $ fsep $ punctuate comma
-                    [ pp l <+> text ":" <+> go 0 t | (l,t) <- fs ]
+      TNewtype nt ts -> optParens (prec > 3) $ fsep (pp (ntName nt) : map (go 5) ts)
+      TRec fs -> ppRecord
+                    [ pp l <+> text ":" <+> go 0 t | (l,t) <- displayFields fs ]
 
       _ | Just tinf <- isTInfix ty0 -> optParens (prec > 2)
                                      $ ppInfix 2 isTInfix tinf
 
-      TUser c [] _ -> pp c
-      TUser c ts _ -> optParens (prec > 3) $ pp c <+> fsep (map (go 4) ts)
+      TUser c ts t ->
+        withNameDisp $ \disp ->
+        case asOrigName c of
+          Just og | NotInScope <- getNameFormat og disp ->
+              go prec t -- unfold type synonym if not in scope
+          _ ->
+            case ts of
+              [] -> pp c
+              _ -> optParens (prec > 3) $ fsep (pp c : map (go 5) ts)
 
       TCon (TC tc) ts ->
         case (tc,ts) of
@@ -771,37 +1153,45 @@ instance PP (WithNames Type) where
           (TCInf,   [])       -> text "inf"
           (TCBit,   [])       -> text "Bit"
           (TCInteger, [])     -> text "Integer"
-          (TCIntMod, [n])     -> optParens (prec > 3) $ text "Z" <+> go 4 n
+          (TCRational, [])    -> text "Rational"
+
+          (TCIntMod, [n])     -> optParens (prec > 3) $ text "Z" <+> go 5 n
 
           (TCSeq,   [t1,TCon (TC TCBit) []]) -> brackets (go 0 t1)
-          (TCSeq,   [t1,t2])  -> optParens (prec > 3)
-                              $ brackets (go 0 t1) <.> go 3 t2
+          (TCSeq,   [t1,t2])  -> optParens (prec > 4)
+                              $ brackets (go 0 t1) <.> go 4 t2
 
           (TCFun,   [t1,t2])  -> optParens (prec > 1)
                               $ go 2 t1 <+> text "->" <+> go 1 t2
 
-          (TCTuple _, fs)     -> parens $ fsep $ punctuate comma $ map (go 0) fs
+          (TCTuple _, fs)     -> ppTuple $ map (go 0) fs
 
-          (_, _)              -> pp tc <+> fsep (map (go 4) ts)
+          (_, _)              -> optParens (prec > 3) $ fsep (pp tc : (map (go 5) ts))
 
       TCon (PC pc) ts ->
         case (pc,ts) of
           (PEqual, [t1,t2])   -> go 0 t1 <+> text "==" <+> go 0 t2
           (PNeq ,  [t1,t2])   -> go 0 t1 <+> text "!=" <+> go 0 t2
           (PGeq,  [t1,t2])    -> go 0 t1 <+> text ">=" <+> go 0 t2
-          (PFin,  [t1])       -> text "fin" <+> (go 4 t1)
+          (PFin,  [t1])       -> optParens (prec > 3) $ text "fin" <+> (go 5 t1)
+          (PPrime,  [t1])     -> optParens (prec > 3) $ text "prime" <+> (go 5 t1)
           (PHas x, [t1,t2])   -> ppSelector x <+> text "of"
                                <+> go 0 t1 <+> text "is" <+> go 0 t2
+          (PAnd, [t1,t2])     -> nest 1 (parens (commaSepFill (map (go 0) (t1 : pSplitAnd t2))))
 
-          (PArith, [t1])      -> pp pc <+> go 4 t1
-          (PCmp, [t1])        -> pp pc <+> go 4 t1
-          (PSignedCmp, [t1])  -> pp pc <+> go 4 t1
-          (PLiteral, [t1,t2]) -> pp pc <+> go 4 t1 <+> go 4 t2
+          (PRing, [t1])       -> pp pc <+> go 5 t1
+          (PField, [t1])      -> pp pc <+> go 5 t1
+          (PIntegral, [t1])   -> pp pc <+> go 5 t1
+          (PRound, [t1])      -> pp pc <+> go 5 t1
 
-          (_, _)              -> pp pc <+> fsep (map (go 4) ts)
+          (PCmp, [t1])        -> pp pc <+> go 5 t1
+          (PSignedCmp, [t1])  -> pp pc <+> go 5 t1
+          (PLiteral, [t1,t2]) -> pp pc <+> go 5 t1 <+> go 5 t2
+          (PLiteralLessThan, [t1,t2]) -> pp pc <+> go 5 t1 <+> go 5 t2
 
-      TCon f ts -> optParens (prec > 3)
-                $ pp f <+> fsep (map (go 4) ts)
+          (_, _)              -> optParens (prec > 3) $ fsep (pp pc : map (go 5) ts)
+
+      TCon f ts -> optParens (prec > 3) $ fsep (pp f : map (go 5) ts)
 
     where
     go p t = ppWithNamesPrec nmMap p t
@@ -809,18 +1199,15 @@ instance PP (WithNames Type) where
     isTInfix (WithNames (TCon tc [ieLeft',ieRight']) _) =
       do let ieLeft  = WithNames ieLeft' nmMap
              ieRight = WithNames ieRight' nmMap
-         (ieOp,fi) <- infixPrimTy tc
-         let ieAssoc = fAssoc fi
-             iePrec  = fLevel fi
+         (ieOp, ieFixity) <- infixPrimTy tc
          return Infix { .. }
 
-    isTInfix (WithNames (TUser n [ieLeft',ieRight'] _) _) =
-      do let ieLeft  = WithNames ieLeft' nmMap
-             ieRight = WithNames ieRight' nmMap
-         fi <- nameFixity n
-         let ieAssoc = fAssoc fi
-             iePrec  = fLevel fi
-             ieOp    = nameIdent n
+    isTInfix (WithNames (TUser n [ieLeft',ieRight'] _) _)
+      | isInfixIdent (nameIdent n) =
+      do let ieLeft   = WithNames ieLeft' nmMap
+             ieRight  = WithNames ieRight' nmMap
+             ieFixity = fromMaybe defaultFixity (nameFixity n)
+             ieOp     = nameIdent n
          return Infix { .. }
 
     isTInfix _ = Nothing
@@ -829,19 +1216,30 @@ instance PP (WithNames Type) where
 
 instance PP (WithNames TVar) where
 
-  ppPrec _ (WithNames (TVBound x) mp) =
-    case IntMap.lookup (tpUnique x) mp of
-      Just a  -> text a
-      Nothing ->
-        case tpFlav x of
-          TPModParam n     -> ppPrefixName n
-          TPOther (Just n) -> pp n <.> "`" <.> int (tpUnique x)
-          _  -> pickTVarName (tpKind x) (tvarDesc (tpInfo x)) (tpUnique x)
+  ppPrec _ (WithNames tv mp) =
+    case tv of
+      TVBound {} -> nmTxt
+      TVFree {} -> "?" <.> nmTxt
+    where
+    nmTxt
+      | Just a <- IntMap.lookup (tvUnique tv) mp = text a
+      | otherwise =
+          case tv of
+            TVBound x ->
+              let declNm n = pp n <.> "`" <.> int (tpUnique x) in
+              case tpFlav x of
+                TPModParam n     -> ppPrefixName n
+                TPUnifyVar       -> pickTVarName (tpKind x) (tvarDesc (tpInfo x)) (tpUnique x)
+                TPSchemaParam n  -> declNm n
+                TPTySynParam n   -> declNm n
+                TPPropSynParam n -> declNm n
+                TPNewtypeParam n -> declNm n
+                TPPrimParam n    -> declNm n
 
-  ppPrec _ (WithNames (TVFree x k _ d) _) =
-    char '?' <.> pickTVarName k (tvarDesc d) x
+            TVFree x k _ d -> pickTVarName k (tvarDesc d) x
 
-pickTVarName :: Kind -> TVarSource -> Int -> Doc
+
+pickTVarName :: Kind -> TypeSource -> Int -> Doc
 pickTVarName k src uni =
   text $
   case src of
@@ -856,12 +1254,20 @@ pickTVarName k src uni =
     LenOfSeq               -> mk "n"
     TypeParamInstNamed _ i -> using i
     TypeParamInstPos f n   -> mk (sh f ++ "_" ++ show n)
-    DefinitionOf x         -> using x
+    DefinitionOf x ->
+      case nameInfo x of
+        GlobalName SystemName og
+          | ogModule og == TopModule exprModName -> mk "it"
+        _ -> using x
     LenOfCompGen           -> mk "n"
-    TypeOfArg mb           -> mk (case mb of
+    GeneratorOfListComp    -> "seq"
+    TypeOfIfCondExpr       -> "b"
+    TypeOfArg ad           -> mk (case argDescrNumber ad of
                                     Nothing -> "arg"
                                     Just n  -> "arg_" ++ show n)
     TypeOfRes              -> "res"
+    FunApp                 -> "fun"
+    TypeFromUserAnnotation -> "user"
     TypeErrorPlaceHolder   -> "err"
   where
   sh a      = show (pp a)
@@ -875,12 +1281,21 @@ instance PP Type where
   ppPrec n t = ppWithNamesPrec IntMap.empty n t
 
 instance PP TVarInfo where
-  ppPrec _ tvinfo = pp (tvarDesc tvinfo) <+> loc
+  ppPrec _ tvinfo = hsep $ [pp (tvarDesc tvinfo)] ++ loc
     where
-    loc = if rng == emptyRange then empty else "at" <+> pp rng
+    loc = if rng == emptyRange then [] else ["at" <+> pp rng]
     rng = tvarSource tvinfo
 
-instance PP TVarSource where
+instance PP ArgDescr where
+  ppPrec _ ad = hsep ([which, "argument"] ++ ofFun)
+        where
+        which = maybe "function" ordinal (argDescrNumber ad)
+        ofFun = case argDescrFun ad of
+                  Nothing -> []
+                  Just f  -> ["of" <+> pp f]
+
+
+instance PP TypeSource where
   ppPrec _ tvsrc =
     case tvsrc of
       TVFromModParam m    -> "module parameter" <+> pp m
@@ -896,11 +1311,31 @@ instance PP TVarSource where
                                                       quotes (pp f)
       DefinitionOf x      -> "the type of" <+> quotes (pp x)
       LenOfCompGen        -> "length of comprehension generator"
-      TypeOfArg mb ->
-        case mb of
-          Nothing -> "type of function argument"
-          Just n -> "type of" <+> ordinal n <+> "function argument"
+      TypeOfArg ad        -> "type of" <+> pp ad
       TypeOfRes             -> "type of function result"
+      TypeOfIfCondExpr      -> "type of `if` condition"
+      TypeFromUserAnnotation -> "user annotation"
+      GeneratorOfListComp    -> "generator in a list comprehension"
+      FunApp                -> "function call"
       TypeErrorPlaceHolder  -> "type error place-holder"
 
+instance PP ModParamNames where
+  ppPrec _ ps =
+    let tps = Map.elems (mpnTypes ps)
+    in
+    vcat $ map pp tps ++
+          if null (mpnConstraints ps) then [] else
+            [ "type constraint" <+>
+                parens (commaSep (map (pp . thing) (mpnConstraints ps)))
+            ] ++
+           [ pp t | t <- Map.elems (mpnTySyn ps) ] ++
+           map pp (Map.elems (mpnFuns ps))
+
+instance PP ModTParam where
+  ppPrec _ p =
+    "type" <+> pp (mtpName p) <+> ":" <+> pp (mtpKind p)
+
+instance PP ModVParam where
+  ppPrec _ p = pp (mvpName p) <+> ":" <+> pp (mvpType p)
+
 
